/*
 * Exercises for get{addr,name}info()
 *
 * Francis.Dupont@enst-bretagne.fr, November 2000
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>

char *
v4str(struct sockaddr_in *sin)
{
	static char buf[128];
	char *p;

	bzero(buf, sizeof(buf));

	if (sin->sin_len != sizeof(*sin)) {
		snprintf(buf, sizeof(buf), "IPv4-length-%d", sin->sin_len);
		return buf;
	}

	if (inet_ntop(AF_INET, &sin->sin_addr, buf, 64) == NULL) 
		return "bad IPv4";

	p = buf + strlen(buf);
	if (sin->sin_port)
		snprintf(p, 16, " port %d", sin->sin_port);

	return buf;
}

char *
v6str(struct sockaddr_in6 *sin)
{
	static char buf[128];
	char *p;

	bzero(buf, sizeof(buf));

	if (sin->sin6_len != sizeof(*sin)) {
		snprintf(buf, sizeof(buf), "IPv6-length-%d", sin->sin6_len);
		return buf;
	}

	if (inet_ntop(AF_INET6, &sin->sin6_addr, buf, 64) == NULL)
		return "bad IPv6";

	p = buf + strlen(buf);
	if (sin->sin6_port)
		snprintf(p, 16, " port %d", sin->sin6_port);

	p = buf + strlen(buf);
	if (sin->sin6_scope_id)
		snprintf(p, 16, " scid %d", sin->sin6_scope_id);

	return buf;
}

char *
addr2str(struct sockaddr *sa)
{
	static char buf[64];

	switch (sa->sa_family) {
	    case AF_INET:
		return v4str((struct sockaddr_in *)sa);

	    case AF_INET6:
		return v6str((struct sockaddr_in6 *)sa);

	    default:
		snprintf(buf, sizeof(buf),
			 "family-%d, length-%d",
			 sa->sa_family, sa->sa_len);
		return buf;
	}
}

void
str2addr(char *src, struct sockaddr *sa)
{
	if (sa == NULL)
		return;

	switch (sa->sa_family) {
	    case AF_INET:
		(void) inet_pton(AF_INET, src,
				 &((struct sockaddr_in *)sa)->sin_addr);
		break;

	    case AF_INET6:
		(void) inet_pton(AF_INET6, src,
				 &((struct sockaddr_in6 *)sa)->sin6_addr);
		break;
	}
}

void
print_ai(struct addrinfo *ai)
{
	printf("flags = 0x%x (%s%s%s%s%s%s)\n",
	       ai->ai_flags,
	       ai->ai_flags & AI_PASSIVE ? "P" : "",
	       ai->ai_flags & AI_CANONNAME ? "C" : "",
	       ai->ai_flags & AI_NUMERICHOST ? "N" : "",
	       ai->ai_flags & AI_ALL ? "A" : "",
	       ai->ai_flags & AI_ADDRCONFIG ? "D" : "",
	       ai->ai_flags & AI_V4MAPPED ? "4" : "");
	printf("family = %d (%s)\n",
	       ai->ai_family,
	       ai->ai_family == AF_INET ? "inet" :
	       ai->ai_family == AF_INET6 ? "inet6" :
	       ai->ai_family == AF_UNSPEC ? "unspec" : "unknown");
	printf("socktype = %d (%s)\n",
	       ai->ai_socktype,
	       ai->ai_socktype == SOCK_STREAM ? "stream" :
	       ai->ai_socktype == SOCK_DGRAM ? "datagram" :
	       ai->ai_socktype == SOCK_RAW ? "raw" : "unknown");
	printf("protocol = %d (%s)\n",
	       ai->ai_protocol,
	       ai->ai_protocol == IPPROTO_TCP ? "tcp" :
	       ai->ai_protocol == IPPROTO_UDP ? "udp" : "XXX");
	printf("address length = %d\n", ai->ai_addrlen);
	printf("canon name = %s\n",
	       ai->ai_canonname ? ai->ai_canonname : "NULL");
	printf("address = %s\n", ai->ai_addr ? addr2str(ai->ai_addr) : "NULL");
}

void
print_ais(char *what, struct addrinfo *ai)
{
	if (ai == NULL) {
		printf("%s: none\n\n", what);
		return;
	}

	printf("%s:\n", what);
	for (; ai != NULL; ai = ai->ai_next) {
		print_ai(ai);
		if (ai->ai_next)
			printf("next\n");
	}

	printf("\n");
}

void
do_getaddrinfo(char *node, char *serv, struct addrinfo *hints)
{
	struct addrinfo *r;
	int cc;

	printf("Nodename = %s\n", node ? node : "NULL");
	printf("Servname = %s\n", serv ? serv : "NULL");
	print_ais("Hints", hints);

	cc = getaddrinfo(node, serv, hints, &r);

	if (cc != 0) {
		printf("failure %d (%s)\n", cc, gai_strerror(cc));
		return;
	}

	print_ais("Result", r);
}

void
do_getnameinfo(struct sockaddr *sa, int flags)
{
	char host[NI_MAXHOST];
	char serv[NI_MAXSERV];
	int cc;

	if (sa == NULL) {
		printf("NO ADDRESS?\n");
		return;
	}

	printf("Address = %s\n", addr2str(sa));
	printf("Flags = 0x%x (%s%s%s%s%s%s%s)\n", flags,
	       flags & NI_NOFQDN ? "F" : "",
	       flags & NI_NUMERICHOST ? "H" : "",
	       flags & NI_NAMEREQD ? "N" : "",
	       flags & NI_NUMERICSERV ? "S" : "",
	       flags & NI_DGRAM ? "D" : "",
	       flags & NI_WITHSCOPEID ? "I" : "",
	       flags & NI_NUMERICSCOPE ? "#" : "");


	bzero(host, sizeof(host));
	bzero(serv, sizeof(serv));

	cc = getnameinfo(sa, sa->sa_len,
			 host, sizeof(host),
			 serv, sizeof(serv),
			 flags);

	if (cc != 0) {
		printf("failure %d (%s)\n", cc, gai_strerror(cc));
		return;
	}

	printf("Host = %s\n", host);
	printf("Serv = %s\n", serv);
}

int
main(int argc, char *argv[])
{
	extern char *optarg;
	extern int optind;
	struct addrinfo hint;
	struct sockaddr_in sin4;
	struct sockaddr_in6 sin6;
	struct sockaddr *sa;
	char *node, *serv;
	int ch, reverse, flags;

	bzero(&hint, sizeof(hint));
	bzero(&sin4, sizeof(sin4));
	bzero(&sin6, sizeof(sin6));
	sin4.sin_family = AF_INET;
	sin4.sin_len = sizeof(sin4);
	sin6.sin6_family = AF_INET6;
	sin6.sin6_len = sizeof(sin6);
	sa = (struct sockaddr *)0;
	node = serv = NULL;
	reverse = flags = 0;

	while ((ch = getopt(argc, argv, "Rn:s:f:t:p:PCNAD4a:RFHSI#")) != -1)
		switch (ch) {
		    case 'R':
			reverse = 1;
			printf("Will do address to name!\n");
			break;

		    case 'n':
			node = optarg;
			break;

		    case 's':
			if (reverse)
				sin6.sin6_scope_id = atoi(optarg);
			else
				serv = optarg;
			break;

		    case 'f':
			switch (*optarg) {
			    case '4':
				hint.ai_family = AF_INET;
				sa = (struct sockaddr *)&sin4;
				break;
			    case '6':
				hint.ai_family = AF_INET6;
				sa = (struct sockaddr *)&sin6;
				break;
			}
			break;

		    case 't':
			switch (*optarg) {
			    case 's':
				hint.ai_socktype = SOCK_STREAM;
				break;
			    case 'd':
				hint.ai_socktype = SOCK_DGRAM;
				break;
			    case 'r':
				hint.ai_socktype = SOCK_RAW;
				break;
			}
			break;

		    case 'p':
			if (reverse) {
				int port = atoi(optarg);

				sin4.sin_port = htons(port);
				sin6.sin6_port = htons(port);
			} else {
				switch (*optarg) {
				    case 't':
					hint.ai_protocol = IPPROTO_TCP;
					break;
				    case 'u':
					hint.ai_protocol = IPPROTO_UDP;
					break;
				}
			}
			break;

		    case 'P':
			hint.ai_flags |= AI_PASSIVE;
			printf("set passive in addrinfo\n");
			break;

		    case 'C':
			hint.ai_flags |= AI_CANONNAME;
			printf("set canonname in addrinfo\n");
			break;

		    case 'N':
			if (reverse) {
				flags |= NI_NAMEREQD;
				printf("set namereqd\n");
			} else {
				hint.ai_flags |= AI_NUMERICHOST;
				printf("set numerichost in addrinfo\n");
			}
			break;

		    case 'A':
			hint.ai_flags |= AI_ALL;
			printf("set all in addrinfo\n");
			break;

		    case 'D':
			if (reverse) {
				flags |= NI_DGRAM;
				printf("set dgram\n");
			} else {
				hint.ai_flags |= AI_ADDRCONFIG;
				printf("set addrconfig in addrinfo\n");
			}
			break;

		    case '4':
			hint.ai_flags |= AI_V4MAPPED;
			printf("set v4mapped in addrinfo\n");
			break;

		    case 'a':
			str2addr(optarg, sa);
			break;

		    case 'F':
			flags |= NI_NOFQDN;
			printf("set noFQDN\n");
			break;

		    case 'H':
			flags |= NI_NUMERICHOST;
			printf("set numerichost\n");
			break;

		    case 'S':
			flags |= NI_NUMERICSERV;
			printf("set numericserv\n");
			break;

		    case 'I':
			flags |= NI_WITHSCOPEID;
			printf("set withscopeid\n");
			break;

		    case '#':
			flags |= NI_NUMERICSCOPE;
			printf("set numericscope\n");
			break;

		    case '?':
		    default:
			if (reverse)
				printf("R -f family -a address -p port -s scid -FHNSDI#\n");
			else
				printf("-n nodename -s servname -f family -t socktype -p protocol -PCNAD4\n");
			exit(1);
		}

	if (reverse)
		do_getnameinfo(sa, flags);
	else
		do_getaddrinfo(node, serv, &hint);

	exit(0);
}
