/*
 * Test TUG (IPv6 over IPv6) BPF cache
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#ifndef if_type
#include <net/if_var.h>
#endif
#include <net/radix.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <netinet/in_pcb.h>
#define KERNEL
#define _KERNEL
#include <netinet/if_tug6.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <kvm.h>
#include <limits.h>
#include <nlist.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

int s;
struct ifreq req = { "tug0" };
struct nlist nl[] = { { "_tugif" }, { "" } };

int main __P((int, char *[]));
void getsize __P((void));
void setsize __P((int));
void dump __P((void));

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int ch, size;

	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}
	while ((ch = getopt(argc, argv, "i:")) != -1)
		switch (ch) {
		    case 'i':
			(void) strncpy(req.ifr_name, optarg, IFNAMSIZ);
			break;

		    default:
		    usage:
			fprintf(stderr,
				"usage: tugcache [-i if] [command...]\n"
				"\tcommand: dump\n"
				"\tcommand: get\n"
				"\tcommand: set size\n");
			exit(1);
		}
	argc -= optind;
	argv += optind;
	if (argc == 0)
		goto usage;
	if (strncasecmp(argv[0], "dump", 4) == 0)
		dump();
	if (strncasecmp(argv[0], "get", 4) == 0)
		getsize();
	if (strncasecmp(argv[0], "set", 4) == 0) {
		if (argc == 1) {
			fprintf(stderr, "tugcache: size is missing\n");
			goto usage;
		}
		setsize(atoi(argv[1]));
	}
	goto usage;
}

void
getsize()
{
	req.ifr_csize = 0;
	if (ioctl(s, SIOCGTUGCSIZE, &req) < 0) {
		perror("ioctl SIOCGTUGCSIZE");
		exit(1);
	}
	printf("get: %.*s cache size is %d\n",
	       IFNAMSIZ, req.ifr_name, req.ifr_csize);
	exit(0);
}

void
setsize(size)
	int size;
{
	req.ifr_csize = size;
	if (ioctl(s, SIOCSTUGCSIZE, &req) < 0) {
		perror("ioctl SIOCSTUGCSIZE");
		exit(1);
	}
	printf("set: %.*s cache size is %d\n",
	       IFNAMSIZ, req.ifr_name, req.ifr_csize);
	exit(0);
}

void
dump()
{
	kvm_t *kvmd;
	char buf[_POSIX2_LINE_MAX], name[IFNAMSIZ];
	struct tug_softc tug;
	struct tug_cache tc;
	struct tug_centry te;
	struct timeval now;
	u_long addr;
	int i, cc;

	kvmd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, buf);
	if (kvmd == NULL) {
		fprintf(stderr, "dump: kvm_open: %s\n", buf);
		exit(1);
	}
	if (kvm_nlist(kvmd, nl) < 0) {
		fprintf(stderr, "dump: kvm_nlist: %s\n", kvm_geterr(kvmd));
		exit(1);
	}
	if (nl[0].n_type == 0) {
		fprintf(stderr, "dump: no namelist\n");
		exit(1);
	}
	if (nl[0].n_value == 0) {
		fprintf(stderr, "dump: symbol %s not found\n", nl[0].n_name);
		exit(1);
	}
	cc = kvm_read(kvmd, nl[0].n_value, &tug, sizeof(tug));
	if (cc != sizeof(tug)) {
		fprintf(stderr, "dump: can't get tug[0]\n", kvm_geterr(kvmd));
		exit(1);
	}
	(void) gettimeofday(&now, NULL);
	bcopy("tug%d\0\0\0\0\0\0\0\0\0\0\0", buf, IFNAMSIZ);

	for (;;) {
		cc = kvm_read(kvmd, (u_long)tug.tug_if.if_name,
			      name, IFNAMSIZ);
		if (cc != IFNAMSIZ) {
			fprintf(stderr,
				"dump: can't get if_name\n",
				kvm_geterr(kvmd));
			exit(1);
		}
		if (strcmp(name, "tug") != 0) {
			fprintf(stderr, "dump: can't find %.*s\n",
				IFNAMSIZ, req.ifr_name);
			exit(1);
		}
#ifndef __NetBSD__
		snprintf(name, IFNAMSIZ, buf, tug.tug_if.if_unit);
#endif
		if (strncmp(name, req.ifr_name, IFNAMSIZ) == 0)
			break;
		cc = kvm_read(kvmd, (u_long)tug.tug_if.if_link.tqe_next,
			      &tug, sizeof(tug));
		if (cc != sizeof(tug)) {
			fprintf(stderr,
				"dump: can't get next tug\n",
				kvm_geterr(kvmd));
			exit(1);
		}
	}
	addr = (u_long)tug.tug_bpfc;
	if (addr == 0) {
		printf("%.*s has no cache\n", IFNAMSIZ, req.ifr_name);
		exit(0);
	}
	cc = kvm_read(kvmd, addr, &tc, sizeof(tc));
	if (cc != sizeof(tc)) {
		fprintf(stderr, "dump: can't get cache\n", kvm_geterr(kvmd));
		exit(1);
	}
	printf("%.*s cache:\n", IFNAMSIZ, req.ifr_name);
	printf("\tmax=%u used=%u\n", tc.tc_nbmax, tc.tc_nbused);
	addr += 2 * sizeof(u_int);
	for (i = 0; i < tc.tc_nbmax; i++) {
		cc = kvm_read(kvmd, addr, &te, sizeof(te));
		if (cc != sizeof(te)) {
			fprintf(stderr, "dump: can't get entry[%d]\n",
				i, kvm_geterr(kvmd));
			exit(1);
		}
		te.te_last.tv_sec -= now.tv_sec;
		te.te_last.tv_usec -= now.tv_usec;
		while (te.te_last.tv_usec < 0) {
			te.te_last.tv_sec -= 1;
			te.te_last.tv_usec += 1000000;
		}
		printf("\taddr=%s last=%ld.%06ld flags=%x count=%u\n",
		       inet_ntop(AF_INET6, &te.te_src, buf, sizeof(buf)),
		       te.te_last.tv_sec, te.te_last.tv_usec,
		       te.te_flags, te.te_cnt);
		addr += sizeof(te);
	}
	exit(0);
}
