/*
 * autococonf6 - set IPv6 link local addresses, initialize lo0, routes ...
 */

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <err.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#ifndef if_type
#include <net/if_var.h>
#endif
#include <net/route.h>
#ifdef DEBUG
#ifdef __FreeBSD__
#include <net/ethernet.h>
#endif
#endif

#include <netinet/in_systm.h>
#define INET6
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in6_var.h>
#include <netinet/if_ether.h>

#include <arpa/inet.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sysexits.h>

#if defined(DEBUG) && !defined(NO_IID_FROM_KERNEL)
#error DEBUG need NO_IID_FROM_KERNEL defined
#endif

#define SA(x)		((struct sockaddr *)(x))
#define S4(x)		((struct sockaddr_in *)(x))
#define S6(x)		((struct sockaddr_in6 *)(x))
#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n)	((x) += ROUNDUP(SA(n)->sa_len))

#define Calloc(p, t)	do { \
		if ((p = (struct t *)calloc(1, sizeof(struct t))) == NULL) \
			errx(EX_CANTCREAT, "out of memory"); \
	} while(0);

struct interface {
	TAILQ_ENTRY(interface) if_list;	/* list of interfaces */
	struct sockaddr_dl *addrl;	/* link address */
	struct sockaddr_in6 *addr6;	/* IPv6 address */
	struct sockaddr_in *addr4;	/* IPv4 address */
	u_int plen4;			/* IPv4 prefix len */
	u_int plen6;			/* IPv6 prefix len */
#define PREF_UNKNOWN	(-1)
#define PREF_HOST6	128
#define PREF_HOST4	32
	u_int flags;			/* copy of kernel flags */
	u_int kind;			/* kind */
	u_int unit;			/* rank in kind */
	u_int mode;			/* copy of kernel flags */
#define	AC_SETLL	0x1
#define	AC_HAS4		0x2
#define	AC_HAS6		0x4
#define	AC_HASLL6	0x8
#define	AC_IEEEOK	(AC_SETLL|AC_HASLL6)
#define	AC_SETSIT	0x10
	char name[IFNAMSIZ];		/* name */
};
TAILQ_HEAD(, interface) interfaces;
#define FORALLIFP(ifp)	\
	for(ifp = interfaces.tqh_first; ifp; ifp = ifp->if_list.tqe_next)
struct interface *main_if_kern, *main_if;
int main_if_index = 0;

#define IFK_IEEE	1		/* Ethernet, FDDI, ... */
#define IFK_LOOP	2		/* Loopback */
#define IFK_SIT		3		/* IPv6 over IPv4 */
#define IFK_P2P		4		/* point to point */
#define IFK_OTHER	5		/* others */

int ifk_ieee4 = 0;
int ifk_ieee6 = 0;
int ifk_loop = 0;
int ifk_sit = 0;
int ifk_p2p = 0;
int ifk_other = 0;

int mib[] = { CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0 };

struct {
	struct rt_msghdr m_rtm;
        char m_space[512];
} m_rtmsg;
#define rtm	m_rtmsg.m_rtm
int rtm_seq = 0;
int pid;

struct if_laddrreq iflr;

char hbuf[64];
#define	INET6_STOP(x)	inet_ntop(AF_INET6, &(x)->sin6_addr, hbuf, sizeof hbuf)

int r, u;			/* Socket file descriptors */

/* Common variables */
int verbose = 0;		/* extra printfs */
int install_sit = 0;		/* install SIT interface */
char *mname = NULL;		/* main interface name */
int llprefixlen = PREF_HOST6-64; /* Official LL prefix length in octets */
char **interflist;
int nbinterflist;
int useiflist;
int doall;
int donotsit;
int donotdefroute;
int donotlo;
int donotmult;
int sit_active;
struct interface *if_defv4, *if_defv6;

/* dad parameter */
extern int nd_waitbefore_dad;

int main(int, char *[]);
void init(void);
int nb_anycast(void);
struct interface *find_main(char *);
void setup_anycast(char *, struct in6_addr *);
void setup_ieee(struct interface *);
void setup_sit(struct interface *);
void setup_loop(struct interface *);
void do_route(int, struct interface *, u_int, int,
	      struct sockaddr *, struct sockaddr *, struct sockaddr *);
#define add_route(ifp, f, p, d, g, m)	\
	do_route(RTM_ADD, (ifp), (f), (p), (d), (g), (m))
#define del_route(ifp, f, p, d, g, m)	\
	do_route(RTM_DELETE, (ifp), (f), (p), (d), (g), (m))
void plen_to_mask(void *, u_int, u_int);
#define	plen_to_mask6(x, l)	plen_to_mask((x), (l), PREF_HOST6)
#define	plen_to_mask4(x, l)	plen_to_mask((x), (l), PREF_HOST4)
u_int mask_to_plen(void *, u_int);
#define mask6_to_plen(x)	mask_to_plen((x), PREF_HOST6)
#define mask4_to_plen(x)	mask_to_plen((x), PREF_HOST4)
void usage(void);
void find_routes(void);
struct interface *find_route(struct sockaddr *, int);
int dad_ipv6(int, u_int, struct if_laddrreq *);
int dad_add_multi(int, int);

/*
   - main if = arg or if of ipv4 default route, or first ieee if
   - set ieee up interfaces without LL addr
   - set lo0
   - set sit0 .. sitn if asked
   - set default route :: through main if if exists
 */

void
usage()
{
        fprintf(stderr,
		"Usage: autoconf6 [-ais6FMORv] [-m main_if] [interface ...]\n");
        exit(2);
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	register struct interface *ifp;
	char ch;
	extern int optind;
	extern char *optarg;

	pid = getpid();
	nd_waitbefore_dad = 1;	/*wait before first dad */

	while ((ch = getopt(argc, argv, "ais6FMORvm:")) != EOF)
		switch (ch) {
		case 'a':
			doall = 1;
			break;
		case 'i':
			useiflist = 1;
			break;
		case 's':
			install_sit = 1;
			break;
		case '6':
			donotsit = 1;
			break;
		case 'F':
			nd_waitbefore_dad = 0;
			break;
		case 'M':
			donotmult = 1;
			break;
		case 'O':
			donotlo = 1;
			break;
		case 'R':
			donotdefroute = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		case 'm':
			mname = optarg;
			break;
		default:
			usage();
		}

	argc -= optind;
	argv += optind;

	if ((doall && useiflist) || ((argc != 0) && !useiflist))
		usage ();
	interflist = argv;
	nbinterflist = argc;

	srandom((int)gethostid());
	setlinebuf(stdout);

	/* Open route "r" socket */
	if ((r = socket(PF_ROUTE, SOCK_RAW, 0)) < 0)
		err(EX_CANTCREAT, "socket (ROUTE)");

	/* Open UDP/IPv6 "u" socket */
	if ((u = socket(PF_INET6, SOCK_DGRAM, 0)) < 0)
		err(EX_CANTCREAT, "socket (UDP)");

	/* Get interfaces */
	init();

	/* Get existing routes */
	find_routes();

	(void) shutdown(r, 0);
	/*
	 * Find the main IEEE LAN interface and put it first with unit 0.
	 */
	main_if = find_main(mname);
	if (verbose)
		printf("default IEEE interface is %s\n",
		       main_if ? main_if->name : "none");
	if (main_if && main_if->unit != 0) {
		FORALLIFP(ifp)
			if ((ifp->kind == IFK_IEEE) && (ifp->unit == 0))
				break;
		ifp->unit = main_if->unit;
		main_if->unit = 0;
	}
	if (main_if && interfaces.tqh_first != main_if) {
		TAILQ_REMOVE(&interfaces, main_if, if_list);
		TAILQ_INSERT_HEAD(&interfaces, main_if, if_list);
	}

	/*
	 * Set the main interface
	 */
	if (mname && main_if && main_if->addrl) {
		main_if_index = main_if->addrl->sdl_index;
		(void) main_interface(&main_if_index);
		if (verbose)
			printf("Set main interface to %s\n", main_if->name);
	}

	/*
	 * Put link-local address on LAN IEEE interfaces.
	 */
	FORALLIFP(ifp)
		if (ifp->mode & AC_SETLL)
			setup_ieee(ifp);

	/*
	 * Setup the LOOPBACK interface.
	 */
	if (!donotlo)
	    FORALLIFP(ifp)
		if ((ifp->kind == IFK_LOOP) && (ifp->unit == 0))
			setup_loop(ifp);

	/*
	 * Don't setup other interfaces.
	 */

	/*
	 * In all the case fe80:: is an anycast address.
	 * declare it if there is not yet any anycast
	 */
	if (nb_anycast() == 0)
		setup_anycast(main_if ? main_if->name : "lo0", NULL);

	/*
	 * add a direct default route on the main interface
	 * (every thing is on-link).
	 */
	if (!donotdefroute && !if_defv6 && main_if && main_if->addrl) {
		struct sockaddr_in6 dest, mask;

		bzero(&dest, sizeof dest);
		dest.sin6_family = AF_INET6;
		dest.sin6_len = sizeof dest;

		bzero(&mask, sizeof mask);
		mask.sin6_family = AF_UNSPEC;
		mask.sin6_len = sizeof mask;

		add_route(main_if, RTF_CLONING, 0,
			  SA(&dest), SA(main_if->addrl), SA(&mask));
	}

	/*
	 * Setup the SIT interface, Put secondary addresses.
	 */
	if (install_sit)
		FORALLIFP(ifp)
			if (ifp->mode & AC_SETSIT)
				setup_sit(ifp);

	exit(EX_OK);
}

/*
 * Get interfaces with their addresses.
 */
void
init()
{
	register struct if_msghdr *ifm;
	register struct interface *ifp;
	struct sockaddr_dl *sdl;
	struct sockaddr *sa, *addr, *mask;
	int i;
	char *buf, *cp, *ap;
	unsigned bufsize;

	main_if_index = main_interface(NULL);
	if (main_if_index < 0)
		errx(EX_OSERR, "get main_interface failed");
	TAILQ_INIT(&interfaces);
	ifp = NULL;

	if (sysctl(mib, 6, NULL, &bufsize, NULL, 0) < 0)
		err(EX_OSERR, "sysctl0");
	buf = (char *)malloc(bufsize);
	if (buf == NULL)
		errx(EX_CANTCREAT, "out of memory");
	if (sysctl(mib, 6, buf, &bufsize, NULL, 0) < 0)
		 err(EX_OSERR, "sysctl");

	for (cp = buf; cp < buf + bufsize; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;

		if (ifm->ifm_type == RTM_IFINFO) {
			ifp = NULL;
			if (ifm->ifm_addrs != RTA_IFP) {
				if (verbose)
					warnx("bad ifinfo (ifp#%d)",
					      ifm->ifm_index);
				continue;
			}
			ap = cp + sizeof(struct if_msghdr);
			sdl = (struct sockaddr_dl *)ap;
			if (sdl->sdl_family != AF_LINK) {
				if (verbose)
					warnx("bad sdl address");
				continue;
			}
			Calloc(ifp, interface);
			TAILQ_INSERT_TAIL(&interfaces, ifp, if_list);
			ifp->flags = ifm->ifm_flags;
			ifp->plen6 = PREF_UNKNOWN;
			ifp->plen4 = PREF_UNKNOWN;

			ifp->addrl = (struct sockaddr_dl *)malloc(sdl->sdl_len);
			if (ifp->addrl == NULL)
				errx(EX_CANTCREAT, "out of memory\n");
			bcopy(sdl, ifp->addrl, sdl->sdl_len);
			bcopy(sdl->sdl_data, ifp->name, sdl->sdl_nlen);

			if (sdl->sdl_index == main_if_index)
				main_if_kern = ifp;
			if (sscanf(ifp->name, "sit%u", &ifp->unit) == 1) {
				ifp->kind = IFK_SIT;
				ifk_sit++;
			} else if ((ifm->ifm_flags & IFF_BROADCAST) &&
				   (ifm->ifm_flags & IFF_MULTICAST) &&
				   (ifm->ifm_data.ifi_addrlen == 6)) {
				ifp->kind = IFK_IEEE;
			} else if ((ifm->ifm_flags & IFF_LOOPBACK)) {
				ifp->kind = IFK_LOOP;
				ifp->unit = ifk_loop++;
			} else if ((ifm->ifm_flags & IFF_POINTOPOINT)) {
				ifp->kind = IFK_P2P;
				ifp->unit = ifk_p2p++;
			} else {
				ifp->kind = IFK_OTHER;
				ifp->unit = ifk_other;
			}

			if (verbose)
				printf("got interface %s\n", ifp->name);
			continue;
		}
		if (ifp == NULL)
			continue;
		if (ifm->ifm_type != RTM_NEWADDR) {
			if (verbose)
				warnx("bad message type (%d)", ifm->ifm_type);
			continue;
		}
		if ((ifm->ifm_addrs & RTA_IFA) == 0) {
			if (verbose)
				warnx("no ifa for %s", ifp->name);
			continue;
		}
		mask = NULL;
		addr = NULL;
		ap = cp + sizeof(struct ifa_msghdr);
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifm->ifm_addrs & (1 << i)) == 0)
				continue;
			sa = SA(ap);
			ADVANCE(ap, sa);
			switch (i) {
			    case RTAX_NETMASK:
				mask = sa;
				break;

			    case RTAX_IFA:
				switch (sa->sa_family) {
				    case AF_INET:
					ifp->mode |= AC_HAS4;
					addr = sa;
					break;
				    case AF_INET6:
					ifp->mode |= AC_HAS6;
					if (IN6_IS_ADDR_LINKLOCAL(&S6(sa)->sin6_addr))
						addr = sa;
				}
				break;
			}
		}
		if (addr == NULL)
			continue;
		switch (addr->sa_family) {
		    case AF_INET:
			if (ifp->addr4) {
				if (verbose)
					warnx("multi IPv4 addresses?");
					continue;
			}
			Calloc(ifp->addr4, sockaddr_in);
			bcopy(addr, ifp->addr4, sizeof(*ifp->addr4));
			if (mask) {
			/* is the V4 mask a prefix ? */
				u_int32_t m;

				ifp->plen4 = mask4_to_plen(
						&S4(mask)->sin_addr);
				plen_to_mask4(&m, ifp->plen4);
				if (m != S4(mask)->sin_addr.s_addr)
					ifp->plen4 = PREF_UNKNOWN;
			} else
				ifp->plen4 = PREF_HOST4;
			break;

		    case AF_INET6:
			if (ifp->addr6) {
				if (verbose)
					warnx("multi LL IPv6 addresses?");
					continue;
			}
			ifp->mode |= AC_HASLL6;
			Calloc(ifp->addr6, sockaddr_in6);
			bcopy(addr, ifp->addr6, sizeof(*ifp->addr6));
			if (mask)
				ifp->plen6 = mask6_to_plen(
						&S6(mask)->sin6_addr);
			else
				ifp->plen6 = PREF_HOST6;
			break;
		}
	}
	free(buf);

#ifdef DEBUG
	/* to test duplicate addresses, use extra args
	 * le0=xxx, means that xxx is the link layer addr of if le0 */
	if (useiflist) {
		char **p = interflist, *q;
		struct ether_addr *e;

		for (i = nbinterflist; i > 0; p++, i--) {
			if (!(q = strchr(*p, '=')))
				continue;
			*q = 0;
			FORALLIFP(ifp)
				if (!strcmp(ifp->name, *p))
					break;
			if (!ifp || ifp->kind != IFK_IEEE ||
			    !ifp->addrl || ifp->addrl->sdl_alen != 6 ||
			    (e = ether_aton(q+1)) == 0)
				continue;
			bcopy(e, LLADDR(ifp->addrl), 6);
		}
	}
#endif
	FORALLIFP(ifp) {
		if (ifp->kind == IFK_IEEE) {
			if (ifp->mode & AC_HAS6)
				donotdefroute = 1;
			if (ifp->mode & AC_HASLL6) {
				ifk_ieee6++;
			} else {
				i = 0;
				if (useiflist) {
					char **p = interflist;

					for (i = nbinterflist; i > 0; i--)
						if (!strcmp(ifp->name, *p++))
							break;
				} else if (doall || (ifp->flags & IFF_UP))
					i = 1;
				if (i != 0) {
					ifp->mode |= AC_SETLL;
					ifk_ieee6++;
				}
			}
			if (ifp->plen4 != PREF_UNKNOWN) {
				ifp->mode |= AC_SETSIT;
				ifk_ieee4++;
			}
		}
		if (ifp->kind == IFK_SIT && (ifp->mode & AC_HAS6))
			sit_active++;
	}
	if (sit_active && !donotsit)
		install_sit = 1;
	if (donotsit)
		install_sit = 0;
	if (install_sit) {
		if (ifk_sit == 0)
			warnx("no SIT interface?");
		else  if (ifk_ieee4 > (ifk_sit - sit_active))
			warnx("not enough SIT interfaces?");
	}

	if (verbose) {
		if (ifk_loop > 1)
			printf("extra local interfaces (%d)\n", ifk_loop - 1);
		if (ifk_ieee6 == 0)
			printf("no IEEE LAN interface?\n");
	}
}

int
nb_anycast()
{
#define ANYCASTSZ	(sizeof(struct in6_addr) + 2 * sizeof(u_int))
	static int mib[] = { CTL_NET, PF_INET6, IPPROTO_IP, IP6CTL_ANYCASTS};
	size_t size = 0;

	if (sysctl(mib, sizeof mib/sizeof(int), NULL, &size, NULL, 0) < 0) {
		if (verbose)
			warn("get anycasts");
		return (0);
	}
	return((size + ANYCASTSZ - 1) / ANYCASTSZ);
}

/*
 * Determine the main LAN IEEE interface.
 * (given by user, IPv4 default interface, first one)
 */
struct interface *
find_main(name)
	char *name;
{
	register struct interface *ifp;

#define TRY(x, y)	if ((x)->kind == IFK_IEEE && ((x)->mode & (y))) \
				return (x)
#define TRY2(x, y)	if ((x)) TRY(x, y)

	/* user has given a name */
	if (name) {
		if (strcasecmp(name, "none") == 0)
			return(NULL);
		FORALLIFP(ifp)
			if (strcasecmp(ifp->name, name) == 0)
				break;
		if (ifp == 0)
			warnx("can't find interface %s", name);
		else {
			TRY(ifp, AC_IEEEOK);
			warnx("interface %s hasn't the good type (%d/%x)?",
				ifp->name, ifp->kind, ifp->mode);
		}
	}
	/* inherit ? */
	TRY2(main_if_kern, AC_IEEEOK);
	/* use default V6 route */
	TRY2(if_defv6, AC_IEEEOK);
	/* use default V4 route */
	TRY2(if_defv4, AC_IEEEOK);

	/* use first of already known ieees */
	FORALLIFP(ifp)
		TRY(ifp, AC_HASLL6);
	/* use first of proposed ieee */
	FORALLIFP(ifp)
		TRY(ifp, AC_SETLL);
	/* failure! */
	return(NULL);
#undef TRY
#undef TRY2
}

void
find_routes()
{
	struct sockaddr_in6 sin6;

	bzero(&sin6, sizeof sin6);
	/* get IPv4, v6 default routes */
	sin6.sin6_len = sizeof(struct sockaddr_in);
	sin6.sin6_family = AF_INET;
	if_defv4 = find_route (SA(&sin6), 0);
	sin6.sin6_len = sizeof(struct sockaddr_in6);
	sin6.sin6_family = AF_INET6;
	if_defv6 = find_route (SA(&sin6), 0);
}

/*
 * Determine the interface for a given destination
 * return ifp or NULL (doesn't exist)
 * prefix : prefix len in bits; < 0: host; -1 : return only exact match
 */
struct interface *
find_route(dest, prefix)
	struct sockaddr *dest;
	int prefix;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr *sa;
	struct sockaddr_dl *sdl;
	register struct interface *ifp;
	int len, i;
	char *p;

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_GET;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	if (dest->sa_family == AF_INET6)
		rtm.rtm_index = S6(dest)->sin6_scope_id;

	sa = SA(cp);
	bcopy(dest, sa, dest->sa_len);
	ADVANCE(cp, sa);

	if (prefix < 0) {
		rtm.rtm_flags = RTF_UP|RTF_GATEWAY|RTF_HOST;
		rtm.rtm_addrs = RTA_DST|RTA_IFP;
	} else {
		rtm.rtm_flags = RTF_UP|RTF_GATEWAY;
		rtm.rtm_addrs = RTA_DST|RTA_IFP|RTA_NETMASK;
		sa = SA(cp);
		sa->sa_len = dest->sa_len;
		sa->sa_family = AF_UNSPEC;
		if (dest->sa_family == AF_INET6)
			plen_to_mask6(&S6(sa)->sin6_addr, prefix);
		else
			plen_to_mask4(&S4(sa)->sin_addr, prefix);
		ADVANCE(cp, sa);
	}
	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_len = sizeof(*sa);
	sdl->sdl_family = AF_LINK;
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose && errno != ESRCH)
			warn("write (get route message)");
		goto failed;
	}

	do {
		if (verbose)
			bzero(&m_rtmsg, sizeof(m_rtmsg));
		len = read(r, (char *)&m_rtmsg, sizeof(m_rtmsg));
	} while (len > 0 && (rtm.rtm_seq != rtm_seq ||  rtm.rtm_pid != pid));
	if (len < 0) {
		if (verbose)
			warn("read (get route message)");
		goto failed;
	}

	if (rtm.rtm_version != RTM_VERSION || rtm.rtm_msglen != len) {
		if (verbose)
			warnx("bad route message");
		goto failed;
	}
	if (rtm.rtm_errno) {
		if (verbose && rtm.rtm_errno != ESRCH)
			warnx("get route failed (%s)",
			      strerror(rtm.rtm_errno));
		goto failed;
	}
	if ((rtm.rtm_addrs & RTA_IFP) == 0 ||
	    (rtm.rtm_flags & RTF_UP) == 0 ||
	    ((prefix == -1 && (rtm.rtm_flags & RTF_HOST) == 0))) {
		if (verbose)
			warnx("no valid route");
		goto failed;
	}
	cp = m_rtmsg.m_space;
	for (i = 0; i < RTAX_MAX; i++) {
		if ((rtm.rtm_addrs & (1 << i)) == 0)
			continue;
		sdl = (struct sockaddr_dl *)cp;
		if (i == RTAX_IFP)
			break;
		ADVANCE(cp, sdl);
	}
	if (sdl->sdl_family != AF_LINK || sdl->sdl_nlen == 0) {
		if (verbose)
			warnx("unusable IFP");
		goto failed;
	}
	p = sdl->sdl_data;
	p[sdl->sdl_nlen] = '\0';

	FORALLIFP(ifp)
		if (strcasecmp(ifp->name, p) == 0)
			break;
	return(ifp);

    failed:
	return(NULL);
}

/*
 * Setup an anycast address (default is fe80::, a link-local router.
 */
void
setup_anycast(name, addr)
	char *name;
	struct in6_addr *addr;
{
	struct sockaddr_in6 *sin;

	bzero(&iflr, sizeof(iflr));
	strncpy(iflr.iflr_name, name, IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&iflr.addr;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	iflr.flags = IP6ANY_ANYCAST;
	if (addr)
		bcopy(addr, &sin->sin6_addr, sizeof(struct in6_addr));
	else {
		sin->sin6_addr.s6_addr[0] = 0xfe;
		sin->sin6_addr.s6_addr[1] = 0x80;
	}

	if (ioctl(u, SIOCADDANY, &iflr) < 0) {
		err(EX_OSERR, "ioctl (SIOCADDANY)");
	}

	if (verbose)
		printf("anycast %s on %s\n", INET6_STOP(sin), name);
}

/*
 * Configure an IEEE LAN interface.
 */
void
setup_ieee(ifp)
	register struct interface *ifp;
{
	if (verbose)
		printf("setup IEEE interface %s\n", ifp->name);

	if (verbose)
		printf("add all multicast route on %s\n", ifp->name);
	if (dad_add_multi(ifp->addrl->sdl_index, 0))
		warn("add route multi");

	Calloc(ifp->addr6, sockaddr_in6);
	ifp->addr6->sin6_len = sizeof(struct sockaddr_in6);
	ifp->addr6->sin6_family = AF_INET6;
	ifp->addr6->sin6_addr.s6_addr[0] = 0xfe;
	ifp->addr6->sin6_addr.s6_addr[1] = 0x80;
#ifdef NO_IID_FROM_KERNEL
	bcopy(LLADDR(ifp->addrl), &ifp->addr6->sin6_addr.s6_addr[8], 3);
	ifp->addr6->sin6_addr.s6_addr[8] ^= 0x02;
	ifp->addr6->sin6_addr.s6_addr[11] = 0xff;
	ifp->addr6->sin6_addr.s6_addr[12] = 0xfe;
	bcopy(LLADDR(ifp->addrl) + 3, &ifp->addr6->sin6_addr.s6_addr[13], 3);
#endif

	ifp->plen6 = llprefixlen;

	bzero(&iflr, sizeof(iflr));
	strncpy(iflr.iflr_name, ifp->name, IFNAMSIZ);
	bcopy(ifp->addr6, &iflr.addr, sizeof(struct sockaddr_in6));
	iflr.prefixlen = ifp->plen6;
#ifndef NO_IID_FROM_KERNEL
	iflr.flags = IFLR_PREFIX;
#endif

#ifdef NO_IID_FROM_KERNEL
	if (verbose)
		printf("config %s %s/%d\n",
		       ifp->name,
		       INET6_STOP(ifp->addr6),
		       ifp->plen6);
#endif

	if (dad_ipv6(u, ifp->addrl->sdl_index, &iflr) <= 0)
		/* XXX - What to do if ifp is the main interface ? */
		warnx("can't setup IEEE interface %s", ifp->name);
#ifndef NO_IID_FROM_KERNEL
	bcopy(&iflr.addr, ifp->addr6, sizeof(struct sockaddr_in6));
#endif
#ifndef NO_IID_FROM_KERNEL
	if (verbose)
		printf("config %s %s/%d\n",
		       ifp->name,
		       INET6_STOP(ifp->addr6),
		       ifp->plen6);
#endif
}

/*
 * Configure the loopback interface.
 */
void
setup_loop(ifp)
	register struct interface *ifp;
{
	if (ifp->unit > 0 || (ifp->mode & AC_HAS6))
		return;

	if (verbose)
		printf("setup loopback interface %s\n", ifp->name);

	if (verbose)
		printf("add all multicast route on %s\n", ifp->name);
	if (dad_add_multi(ifp->addrl->sdl_index, 0))
		warn("add route multi");

	Calloc(ifp->addr6, sockaddr_in6);
	ifp->addr6->sin6_len = sizeof(struct sockaddr_in6);
	ifp->addr6->sin6_family = AF_INET6;
	ifp->addr6->sin6_addr.s6_addr[15] = 1;
	ifp->plen6 = PREF_HOST6;

	bzero(&iflr, sizeof(iflr));
	snprintf(iflr.iflr_name, IFNAMSIZ + 1, "lo%d", ifp->unit);
	bcopy(ifp->addr6, &iflr.addr, sizeof(struct sockaddr_in6));
	iflr.prefixlen = ifp->plen6;

	if (ioctl(u, SIOCALIFADDR, &iflr) < 0) {
		warn("ioctl(SIOCALIFADDR) interface %s", ifp->name);
		return;
	}

	if (verbose)
		printf("config %s %s/%d\n",
		       ifp->name,
		       INET6_STOP(ifp->addr6),
		       ifp->plen6);
}

/*
 * Put IPv4-compatible addresses on SIT interfaces.
 */
void
setup_sit(ifpi)
	register struct interface *ifpi;
{
	struct interface *ifps = NULL, *ifp;

	FORALLIFP(ifp)
		if ((ifp->kind == IFK_SIT) && !(ifp->mode & AC_HAS6)) {
			if ((ifps == NULL) || (ifps->unit > ifp->unit))
				ifps = ifp;
		}

	if (ifps == NULL)
		return;
	/* not enough sit interfaces, do only sit0 */
	if ((ifps->unit != 0) && (ifk_ieee4 > (ifk_sit - sit_active)))
		return;

	if (verbose)
		printf("setup2 SIT interface %s for %s\n",
		       ifps->name, ifpi->name);

	ifps->mode |= AC_HAS6;

	Calloc(ifps->addr6, sockaddr_in6);
	ifps->addr6->sin6_family = AF_INET6;
	ifps->addr6->sin6_len = sizeof(struct sockaddr_in6);
	bcopy(&ifpi->addr4->sin_addr,
	      &ifps->addr6->sin6_addr.s6_addr[
			sizeof(struct in6_addr)-sizeof(struct in_addr)],
	      sizeof(struct in_addr));

	ifps->plen6 = PREF_HOST6 - PREF_HOST4;
	if (ifps->unit != 0)
		ifps->plen6 += ifpi->plen4;

	bzero(&iflr, sizeof(iflr));
	strncpy(iflr.iflr_name, ifps->name, IFNAMSIZ + 1);
	bcopy(ifps->addr6, &iflr.addr, sizeof(struct sockaddr_in6));
	iflr.prefixlen = ifps->plen6;

	if (ioctl(u, SIOCALIFADDR, &iflr) < 0) {
		fprintf(stderr, "can't setup interface %s\n", ifps->name);
		return;
	}
	if (verbose)
		printf("config %s %s/%d\n",
		       ifps->name,
		       INET6_STOP(ifps->addr6),
		       ifps->plen6);
}

/*
 * Add/delete a direct route for dst on an interface.
 */
void
do_route(cmd, ifp, flags, prefix, dest, gateway, mask)
	int cmd;
	register struct interface *ifp;
	u_int flags;
	int prefix;	/* in bits */
	struct sockaddr *dest, *gateway, *mask;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr *sa;
	int len;

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = cmd;
	rtm.rtm_flags = RTF_UP|RTF_STATIC|flags;
	rtm.rtm_addrs = RTA_DST|RTA_GATEWAY|RTA_NETMASK;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = ifp->addrl->sdl_index;

	sa = SA(cp);
	bcopy(dest, cp, dest->sa_len);
	ADVANCE(cp, sa);
	sa = SA(cp);
	bcopy(gateway, cp, gateway->sa_len);
	ADVANCE(cp, sa);
	sa = SA(cp);
	bcopy(mask, cp, mask->sa_len);
	ADVANCE(cp, sa);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	if (write(r, (char *)&m_rtmsg, len) < len)
		if (verbose)
			warn("write (route message)");
	if (verbose) {
		printf("add route flags=%x\n", rtm.rtm_flags);
		printf("\tdest=%s\n", INET6_STOP(S6(dest)));
		if (gateway->sa_family == AF_INET6)
			printf("\tgateway=%s\n", INET6_STOP(S6(gateway)));
		else if (gateway->sa_family == AF_LINK)
			printf("\tgateway=%s#%d\n", ifp->name, rtm.rtm_index);
		printf("\tprefix len=%d bits\n", prefix);
	}
}

void
plen_to_mask(a, pref, len)
void *a;
u_int pref, len;
{
	register u_int8_t *p = ( u_int8_t *)a;

	if (pref > len)		/* should not happen */
		abort();
	for (; pref >= 8; pref -=8, len -=8)
		*p++ = 0xff;
	if (pref) {
		len -= 8;
		*p++ = (0xff00 >> pref);
	}
	for (; len >= 8; len -=8)
		*p++ = 0;
}

u_int
mask_to_plen(a, len)
void *a;
u_int len;
{
	register u_int8_t *p = ( u_int8_t *)a;
	u_int pref = 0, i;
	
	for (; len >= 8 && *p == 0xff; len -=8, pref +=8)
		p++;
	if (len) {
		i = *p;
		if (len < 8)
			i &= (0xff00 >> len);
		for (; i & 0x80; i >>=1)
			pref++;
	}
	return pref;
}
