/*
 * Dual Stack Transition Mechanism Daemon
 *
 * Francis.Dupont@inria.fr, August 2000
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/sysctl.h>
#include <sys/time.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#ifndef if_type
#include <net/if_var.h>
#endif
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_pcb.h>
#include <netinet/ip6.h>
#include <netinet/if_bti6.h>
#include <netinet/if_dti6.h>
#include <arpa/inet.h>

#include <netdb.h>
#include <errno.h>
#include <kvm.h>
#include <nlist.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "getopt.h"

int active = 0;
int s;
int r;
int seq = 1;
int verbose = 0;
u_int idx;
u_int keep = 20 * 60;
u_int extend = 5 * 60;
pid_t pid;
int timo;
char *ifname = "dti0";
int mib[] = { CTL_NET, PF_INET6, IPPROTO_IPIP, DTI6CTL_MAGIC };
struct in_addr tep4, myaddr, mask;
struct in6_addr tep6;
struct {
	struct rt_msghdr m_rtm;
	char m_space[512];
} m_rtmsg0, m_rtmsg;
#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n)	(x += ROUNDUP((n)->sa_len))

struct rt_addrinfo info;
#define rtidst		((struct sockaddr_in *)info.rti_info[RTAX_DST])
#define rtiifp		((struct sockaddr_dl *)info.rti_info[RTAX_IFP])

struct nlist nl[] = {
	{ "_rt_tables" },
	{ "" },
};
kvm_t *kvmd;
char *kfile;
char *mfile;

void parseargs(int, char *const *);
void expired(int);
int getmagic(void);
int setmagic(int);
int getrtmsg();
int defroute(u_int);
int setdti();
int resetdti();
int getdefrc();

int
main(argc, argv)
	int argc;
	char *const *argv;
{
	pid = getpid();
	if (!getmagic()) {
		fprintf(stderr, "dstmd: sysctl get failed\n");
		exit(1);
	}

	parseargs(argc, argv);

	idx = if_nametoindex(ifname);
	if (verbose)
		printf("dstmd: dti0 index: %d\n", (int)idx);

	if ((s = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket/inet");
		exit(1);
	}

	if ((r = socket(PF_ROUTE, SOCK_RAW, AF_INET)) < 0) {
		perror("socket/route");
		exit(1);
	}

	(void) signal(SIGALRM, expired);
	(void) signal(SIGTERM, expired);
	(void) siginterrupt(SIGALRM, 1);

	for (;;) {
		if (!getrtmsg())
			continue;
		if (!setdti())
			continue;
		if (!defroute(RTM_ADD)) {
			(void)defroute(RTM_DELETE);
			if (defroute(RTM_ADD))
				goto recovered;
			fprintf(stderr, "dstmd: can't set default route\n");
			if (!resetdti()) {
				fprintf(stderr, "dstmd: too many errors\n");
				exit(1);
			}
			continue;
		}
	    recovered:
		if (!setmagic(0)) {
			fprintf(stderr, "dstmd: sysctl set failed\n");
			exit(1);
		}

		active = 1;
		(void)alarm(keep);
		while (active)
			(void)getrtmsg();
	}
}

void
parseargs(argc, argv)
	int argc;
	char *const *argv;
{
	static struct option long_options[] =
	{
		{ "addr", 1, NULL, 0 },
		{ "extend", 1, NULL, 0 },
		{ "if", 1, NULL, 0 },
		{ "keep", 1, NULL, 0 },
		{ "mask", 1, NULL, 0 },
		{ "tep4", 1, NULL, 0 },
		{ "tep6", 1, NULL, 0 },
		{ "timeout", 1, NULL, 0 },
		{ "verbose", 0, NULL, 0 },
		{ NULL, 0, NULL, 0 }
	};
	u_int got = 0;
	int ch, option_index = 0;

	while (1) {
		ch = getopt_long(argc, argv, "vh",
				 long_options, &option_index);
		if (ch == -1)
			break;
		if (ch == 0) switch (option_index) {
		    case 0:
			if (!inet_aton(optarg, &myaddr)) {
				fprintf(stderr,
					"dstmd: bad addr %s\n",
					optarg);
				exit(1);
			}
			got |= 1;
			break;

		    case 1:
			extend = atoi(optarg);
			break;

		    case 2:
			ifname = optarg;
			break;

		    case 3:
			keep = atoi(optarg);
			break;

		    case 4:
			if (!inet_aton(optarg, &mask)) {
				fprintf(stderr,
					"dstmd: bad mask %s\n",
					optarg);
				exit(1);
			}
			got |= 2;
			break;

		    case 5:
			if (!inet_aton(optarg, &tep4)) {
				fprintf(stderr,
					"dstmd: bad tep4 %s\n",
					optarg);
				exit(1);
			}
			got |= 4;
			break;

		    case 6:
			if (inet_pton(AF_INET6, optarg, &tep6) != 1) {
				fprintf(stderr,
					"dstmd: bad tep6 %s\n",
					optarg);
				exit(1);
			}
			got |= 8;
			break;

		    case 7:
			timo = atoi(optarg);
			break;

		    case 8:
			verbose++;
			break;

		    default:
			fprintf(stderr, "dstmd: can't parse arguments\n");
			exit(1);
		}
		if (ch != 0) switch (ch) {
		    case 'v':
			verbose++;
			break;

		    case 'h':
		    case '?':
			goto usage;

		    default:
			fprintf(stderr, "dstmd: can't parse arguments\n");
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc) {
		fprintf(stderr, "dstmd: %d extra arguments\n", argc);
		goto usage;
	}
	if (got == 15)
		return;

	if ((got & 1) == 0)
		fprintf(stderr, "dstmd: addr is needed\n");
	if ((got & 2) == 0)
		fprintf(stderr, "dstmd: mask is needed\n");
	if ((got & 4) == 0)
		fprintf(stderr, "dstmd: tep4 is needed\n");
	if ((got & 8) == 0)
		fprintf(stderr, "dstmd: tep6 is needed\n");

    usage:
	fprintf(stderr,
		"usage: dstmd %s%s%s%s%s%s%s%s%s",
		"\t[-v|--verbose] (increase verbosity)\n",
		"\t--addr <my IPv4 address>\n",
		"\t--mask <IPv4 subnet mask>\n",
		"\t--tep4 <TEP IPv4 address>\n",
		"\t--tep6 <TEP IPv6 address>\n",
		"\t[--if <interface name>] (default: dti0)\n",
		"\t[--timeout <tsleep timeout>] (default: dti_magic)\n",
		"\t[--keep <delay> (default: 20 minutes)\n",
		"\t[--extend <delay> (default: 5 minutes)\n");
	exit(1);
}

void
expired(sig)
	int sig;
{
	int refcount;

	/* signal is not SIGALRM: we'll quit */
	if (sig != SIGALRM)
		(void)alarm(0);

	if (verbose)
		printf("dstmd: kept signal: %d\n", sig);

	/* see if we should extend */
	if ((sig == SIGALRM) && (extend != 0)) {
		refcount = getdefrc();
		if (refcount > 0) {
			if (verbose)
				printf("dstmd: extend\n");
			(void)alarm(extend);
			return;
		}
		if ((refcount < 0) && verbose)
			fprintf(stderr,
				"dstmd: can't get default "
				"route reference count\n");
		else if ((refcount == 0) && verbose)
			printf("dstmd: unused default route\n");
	}			

	if (!setmagic(timo)) {
		fprintf(stderr, "dstmd: sysctl set failed\n");
		exit(1);
	}
	if (!defroute(RTM_DELETE))
		fprintf(stderr, "dstmd: can't delete default route\n");
	if (!resetdti())
		fprintf(stderr, "dstmd: can't reset dti\n");

	if (sig == SIGALRM)
		active = 0;
	else {
		if (verbose)
			printf("dstmd: finish on signal\n");
		exit(0);
	}
}

int
getmagic()
{
	int size = sizeof(int);

	if (sysctl(mib, 4, &timo, &size, NULL, 0) < 0) {
		perror("sysctl/get");
		exit(1);
	}
	if (size != sizeof(int)) {
		fprintf(stderr, "dstmd: sysctl bad size\n");
		exit(1);
	}
	if (verbose)
		printf("dstmd: sysctl get: %d\n", timo);
	return(1);
}

int
setmagic(val)
	int val;
{
	if (sysctl(mib, 4, NULL, NULL, &val, sizeof(int)) < 0) {
		perror("sysctl/set");
		return(0);
	}
	if (verbose)
		printf("dstmd: sysctl set: %d\n", val);
	return(1);
}

int
getrtmsg()
{
	register struct rt_msghdr *rtm;
	register char *cp;
	register int i, cc;
	struct sockaddr *sa;

	bzero(&m_rtmsg0, sizeof(m_rtmsg0));

	cc = read(r, (char *)&m_rtmsg0, sizeof(m_rtmsg0));
	if (cc < 0) {
		if (errno == EINTR)
			return(0);
		perror("read");
		exit(1);
	}
	if (cc == 0) {
		fprintf(stderr, "dstmd: read returns 0?\n");
		exit(1);
	}
	rtm = &m_rtmsg0.m_rtm;
	cp = m_rtmsg0.m_space;

	if (rtm->rtm_version != RTM_VERSION) {
		fprintf(stderr,
			"dstmd: bad version (%d)\n",
			rtm->rtm_version);
		return(0);
	}
	if (rtm->rtm_type != RTM_RESOLVE)
		return(0);
	if (rtm->rtm_pid != 0)
		return(0);
	if (rtm->rtm_msglen > cc)
		cc = rtm->rtm_msglen;	/* ??? */
	bzero(&info, sizeof(info));
	info.rti_addrs = rtm->rtm_addrs;
	if ((info.rti_addrs & RTA_DST) == 0) {
		fprintf(stderr, "dstmd: no destination\n");
		return(0);
	}
	if ((info.rti_addrs & RTA_IFP) == 0)
		return(0);
	for (i = 0; i < RTAX_MAX; i++) {
		if ((info.rti_addrs & (1 << i)) == 0)
			continue;
		info.rti_info[i] = sa = (struct sockaddr *)cp;
		ADVANCE(cp, sa);
	}
	if (cp > (char *)(&m_rtmsg0 + 1)) {
		fprintf(stderr, "dstmd: read overflow\n");
		return(0);
	}
	if ((rtiifp->sdl_family != AF_LINK) ||
	    (rtiifp->sdl_index != idx))
		return(0);
	if (rtidst->sin_family != AF_INET) {
		fprintf(stderr, "dstmd: no IP destination\n");
		return(0);
	}

	if (verbose)
		printf("dstmd: addr=%s\n", inet_ntoa(rtidst->sin_addr));

	return(1);
}

int
defroute(cmd)
	u_int cmd;
{
	register struct rt_msghdr *rtm;
	register char *cp;
	register int cc;
	struct sockaddr_in *sin;

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm = &m_rtmsg.m_rtm;
	cp = m_rtmsg.m_space;

	rtm->rtm_version = RTM_VERSION;
	rtm->rtm_type = cmd;
	rtm->rtm_flags = RTF_UP | RTF_GATEWAY | RTF_STATIC;
	rtm->rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
	rtm->rtm_pid = pid;
	rtm->rtm_seq = seq++;

	sin = (struct sockaddr_in *)cp;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	ADVANCE(cp, (struct sockaddr *)sin);

	sin = (struct sockaddr_in *)cp;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	sin->sin_addr = tep4;
	ADVANCE(cp, (struct sockaddr *)sin);

	sin = (struct sockaddr_in *)cp;
	sin->sin_family = AF_UNSPEC;
	sin->sin_len = sizeof(*sin);
	ADVANCE(cp, (struct sockaddr *)sin);

	if (cp > (char *)(&m_rtmsg + 1)) {
		fprintf(stderr, "dstmd: write overflow\n");
		return(0);
	}
	rtm->rtm_msglen = cc = cp - (char *)rtm;

	if (write(r, (char *)&m_rtmsg, cc) < 0) {
		perror("write/route");
		exit(1);
	}
	if (verbose)
		printf("dstmd: %s default route via %s\n",
		       cmd == RTM_ADD ? "add" : "delete",
		       inet_ntoa(tep4));

	return(1);
}

int
setdti()
{
	struct rt_msghdr *rtm;
	char *cp;
	int cc;
	struct ifaliasreq ifra;
	struct sockaddr_in *sin;
	struct sockaddr_in6 *sin6;

	bzero(&ifra, sizeof(ifra));
	strncpy(ifra.ifra_name, ifname, IFNAMSIZ);
	sin = (struct sockaddr_in *)&ifra.ifra_addr;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	sin->sin_addr = myaddr;
	sin = (struct sockaddr_in *)&ifra.ifra_mask;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	sin->sin_addr = mask;
	if (ioctl(s, SIOCAIFADDR, (char *)&ifra) < 0) {
		perror("ioctl/SIOCAIFADDR");
		return(0);
	}
	if (verbose)
		printf("dstmd: set interface %s with %s\n",
		       ifname, inet_ntoa(myaddr));

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm = &m_rtmsg.m_rtm;
	cp = m_rtmsg.m_space;

	rtm->rtm_version = RTM_VERSION;
	rtm->rtm_type = RTM_ADD;
	rtm->rtm_flags = RTF_UP | RTF_HOST | RTF_STATIC;
	rtm->rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFA;
	rtm->rtm_pid = pid;
	rtm->rtm_seq = seq++;

	sin = (struct sockaddr_in *)cp;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	sin->sin_addr = tep4;
	ADVANCE(cp, (struct sockaddr *)sin);

	sin6 = (struct sockaddr_in6 *)cp;
	sin6->sin6_family = AF_INET6;
	sin6->sin6_len = sizeof(*sin6);
	bcopy(&tep6, &sin6->sin6_addr, sizeof(tep6));
	ADVANCE(cp, (struct sockaddr *)sin6);

	sin = (struct sockaddr_in *)cp;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	sin->sin_addr = myaddr;
	ADVANCE(cp, (struct sockaddr *)sin);

	if (cp > (char *)(&m_rtmsg + 1)) {
		fprintf(stderr, "dstmd: write overflow\n");
		return(0);
	}
	rtm->rtm_msglen = cc = cp - (char *)rtm;

	if (write(r, (char *)&m_rtmsg, cc) < 0) {
		perror("write/route");
		exit(1);
	}
	if (verbose)
		printf("dstmd: add route for %s\n",
		       inet_ntoa(tep4));

	return(1);
}

int
resetdti()
{
	struct rt_msghdr *rtm;
	char *cp;
	int cc;
	struct ifreq ifr;
	struct sockaddr_in *sin;

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm = &m_rtmsg.m_rtm;
	cp = m_rtmsg.m_space;

	rtm->rtm_version = RTM_VERSION;
	rtm->rtm_type = RTM_DELETE;
	rtm->rtm_flags = RTF_UP | RTF_HOST | RTF_LLINFO | RTF_STATIC;
	rtm->rtm_addrs = RTA_DST;
	rtm->rtm_pid = pid;
	rtm->rtm_seq = seq++;

	sin = (struct sockaddr_in *)cp;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	sin->sin_addr = tep4;
	ADVANCE(cp, (struct sockaddr *)sin);

	if (cp > (char *)(&m_rtmsg + 1)) {
		fprintf(stderr, "dstmd: write overflow\n");
		return(0);
	}
	rtm->rtm_msglen = cc = cp - (char *)rtm;

	if (write(r, (char *)&m_rtmsg, cc) < 0) {
		perror("write/route");
		exit(1);
	}
	if (verbose)
		printf("dstmd: delete route for %s\n",
		       inet_ntoa(tep4));

	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
	if (ioctl(s, SIOCGIFFLAGS, (char *)&ifr) < 0) {
		perror("ioctl/SIOCGIFFLAGS");
		return(0);
	}
	ifr.ifr_flags &= ~IFF_UP;
	if (ioctl(s, SIOCSIFFLAGS, (char *)&ifr) < 0) {
		perror("ioctl/SIOCSIFFLAGS");
		return(0);
	}
	if (verbose)
		printf("dstmd: interface %s down\n", ifname);

	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
	sin = (struct sockaddr_in *)&ifr.ifr_addr;
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);
	sin->sin_addr = myaddr;
	if (ioctl(s, SIOCDIFADDR, (char *)&ifr) < 0) {
		perror("ioctl/SIOCDIFADDR");
		return(0);
	}
	if (verbose)
		printf("dstmd: interface %s delete %s\n",
		       ifname, inet_ntoa(myaddr));

	return(1);
}

int
getdefrc()
{
	struct radix_node_head head, *rnh, *rt_tables[AF_INET + 1];
	struct radix_node rnode, *rn;
	struct rtentry rtentry;
	struct sockaddr_in sin;
	int cc;

	if (kvmd == 0)
		kvmd = kvm_openfiles(kfile, mfile, NULL, O_RDONLY, NULL);
	if (kvmd == 0) {
		fprintf(stderr, "dstmd: kvm not available\n");
		return(-1);
	}
	if (kvm_nlist(kvmd, nl) < 0) {
		fprintf(stderr, "dstmd: kvm_nlist: %s\n", kvm_geterr(kvmd));
	    geterr:
		(void)kvm_close(kvmd);
		kvmd = 0;
		return(-1);
	}
	if (nl[0].n_type == 0) {
		fprintf(stderr, "dstmd: no namelist\n");
		goto geterr;
	}

	cc = kvm_read(kvmd, nl[0].n_value, &rt_tables, sizeof(rt_tables));
	if (cc != sizeof(rt_tables)) {
		fprintf(stderr,
			"dstmd: get rt_tables: %s\n",
			kvm_geterr(kvmd));
		goto geterr;
	}

	rnh = rt_tables[AF_INET];
	cc = kvm_read(kvmd, (u_long)rnh, &head, sizeof(head));
	if (cc != sizeof(head)) {
		fprintf(stderr, "dstmd: get rnh: %s\n", kvm_geterr(kvmd));
		goto geterr;
	}

	rn = head.rnh_treetop;
	for (;;) {
		cc = kvm_read(kvmd, (u_long)rn, &rnode, sizeof(rnode));
		if (cc != sizeof(rnode)) {
			fprintf(stderr, "dstmd: get rnode(%lx): %s\n",
				(u_long)rn, kvm_geterr(kvmd));
			goto geterr;
		}

		if (rnode.rn_b < 0) {
			if (rnode.rn_dupedkey) {
				rn = rnode.rn_dupedkey;
				continue;
			}
			break;
		} else {
			rn = rnode.rn_l;
			continue;
		}
	}

	if ((rnode.rn_b >= 0) || (rnode.rn_flags & RNF_ROOT)) {
		fprintf(stderr,
			"dstmd: getdefrc ends on non-terminal(%lx)\n",
			(u_long)rn);
		goto geterr;
	}

	cc =  kvm_read(kvmd, (u_long)rn, &rtentry, sizeof(rtentry));
	if (cc != sizeof(rtentry)) {
		fprintf(stderr, "dstmd: get rtentry(%lx): %s\n",
			(u_long)rn, kvm_geterr(kvmd));
		goto geterr;
	}
	if (((rtentry.rt_flags & RTF_UP) == 0) ||
	    ((rtentry.rt_flags & RTF_GATEWAY) == 0) ||
	    ((rtentry.rt_flags & RTF_HOST) != 0)) {
		fprintf(stderr, "dstmd: bad rtentry flags\n");
		goto geterr;
	}

	cc = kvm_read(kvmd, (u_long)rnode.rn_key, &sin, sizeof(sin));
	if (cc != sizeof(sin)) {
		fprintf(stderr, "dstmd: get rt_key: %s\n", kvm_geterr(kvmd));
		goto geterr;
	}
	if ((sin.sin_family != AF_INET) ||
	    (sin.sin_len != sizeof(sin)) ||
	    (sin.sin_addr.s_addr != INADDR_ANY)) {
		fprintf(stderr, "dstmd: bad rt_key\n");
		goto geterr;
	}

	cc = kvm_read(kvmd, (u_long)rnode.rn_mask, &sin, sizeof(sin));
	if (cc != sizeof(sin)) {
		fprintf(stderr, "dstmd: get rt_mask: %s\n", kvm_geterr(kvmd));
		goto geterr;
	}
	if ((sin.sin_family != AF_UNSPEC) ||
	    (sin.sin_addr.s_addr != INADDR_ANY)) {
		fprintf(stderr, "dstmd: bad rt_mask\n");
		goto geterr;
	}

	cc = kvm_read(kvmd, (u_long)rtentry.rt_gateway, &sin, sizeof(sin));
	if (cc != sizeof(sin)) {
		fprintf(stderr,
			"dstmd: get rt_gateway: %s\n",
			kvm_geterr(kvmd));
		goto geterr;
	}
	if ((sin.sin_family != AF_INET) ||
	    (sin.sin_len != sizeof(sin)) ||
	    (sin.sin_addr.s_addr != tep4.s_addr)) {
		fprintf(stderr, "dstmd: bad rt_gateway\n");
		goto geterr;
	}

	return((int)rtentry.rt_refcnt);
}
