/* <> and not " " : use the one in search path, not in current dir */
#include <ndpd.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

/*
 * Shared between ndpd-host and ndpd-router
 * Use #ifdef NDPD_ROUTER when needed
 */

static void logdate(FILE *);

void
add_timeout(tp)
	register struct timeout *tp;
{
	register struct timeout *tp2;

	tp->status = 1;
	if (timeouts.tqh_first == NULL) {
		TAILQ_INSERT_HEAD(&timeouts, tp, tm_list);
		return;
	}
	for (tp2 = timeouts.tqh_first; tp2; tp2 = tp2->tm_list.tqe_next)
		if (!AFTER(tp->tv, tp2->tv))
			break;

	if (tp2 == NULL)
		TAILQ_INSERT_TAIL(&timeouts, tp, tm_list);
	else
		TAILQ_INSERT_BEFORE(tp2, tp, tm_list);
}

void
set_timeout(tp, sec, usec, func, arg)
	register struct timeout *tp;
	int sec, usec;
	func_to func;
	void * arg;
{
	tp->tv.tv_sec = now.tv_sec + sec;
	tp->tv.tv_usec = now.tv_usec + usec;
	NORMTIME(tp->tv);
	tp->func = func;
	tp->arg = arg;
	add_timeout(tp);
}

struct timeout *
new_timeout(sec, usec, func, arg)
	int sec, usec;
	func_to func;
	void * arg;
{
	register struct timeout *tp;

	if (Calloc(tp, timeout) == NULL) {
		logerr("out of memory");
		return 0;
	}
	set_timeout(tp, sec, usec, func, arg);
	return tp;
}

void
kill_timeout(tp)
	register struct timeout *tp;
{
	if (tp->status > 0) {
		TAILQ_REMOVE(&timeouts, tp, tm_list);
		tp->status = 0;
	}
	if (tp->status >= 0 && tp->refcnt == 0)
		free(tp);
}

void
run_timeouts()
{
	register struct timeout *tp;

	for(;;) {
		(void) gettimeofday(&now, NULL);
		tp = timeouts.tqh_first;
		if (tp == 0 || !AFTER(now, tp->tv))
			return;
		TAILQ_REMOVE(&timeouts, tp, tm_list);
		tp->status = -1;
		(*(tp->func)) (tp->arg);
		if (tp->status < 0) {	/* not reused */
			if (tp->refcnt == 0)
				free(tp);
			else
				tp->status = 0;
		}
	}
}

void
dump_timeout(tp)
	register struct timeout *tp;
{
	struct timeval delta;
	int sign = 0;

	logdebug("timeout (%lx):", (u_long)tp);
	if (AFTER(tp->tv, now)) {
		sign = +1;
		delta.tv_sec = tp->tv.tv_sec - now.tv_sec - 1;
		delta.tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
		NORMTIME(delta);
	} else {
		sign = -1;
		delta.tv_sec =  now.tv_sec - tp->tv.tv_sec - 1;
		delta.tv_usec = now.tv_usec + 1000000 - tp->tv.tv_usec;
		NORMTIME(delta);
	}
	logdebug("\texpire=%+ld.%06lu", sign * delta.tv_sec, delta.tv_usec);
	logdebug("\tstatus=%d\trefcnt=%d", tp->status, tp->refcnt);
	logdebug("\tfunction=%p\targument=%p", tp->func, tp->arg);
}

void
apply_mask(addr, mask)
	register struct in6_addr *addr, *mask;
{
	register int i;

	for (i = 0; i < 16; i++)
		addr->s6_addr[i] &= mask->s6_addr[i];
}

#ifndef NDPD_ROUTER
void
mk_addr(pp)
	register struct prefix *pp;
{
	struct in6_addr mask;
	u_char ltoken = pp->ifp->ltoken;

	/*
	 * if flag on link, the mask is prfix-wide
	 * if not flag on link, the mask is "host address"
	 */
	if (pp->alen == 0) {
		if (pp->nflags & ND_OPT_PI_FLAG_ONLINK)
			pp->alen = pp->plen;
		else
			pp->alen = 128;
	}

	CLR_ADDR6(mask);
	mk_mask(&mask, pp->alen);
	COPY_ADDR6(pp->pref, pp->addr);
	apply_mask(&pp->addr, &mask);
	bcopy(&pp->ifp->llocal.s6_addr[16 - ltoken],
	      &pp->addr.s6_addr[16 - ltoken],
	      ltoken);
}

void
mk_peer(pp)
	register struct prefix *pp;
{
	struct interface *ifp = pp->ifp;
	struct router *rp;
	struct prefix *pp1;
	struct in6_addr *piid = NULL;
	struct if_laddrreq ifr;
	int lastchance = 0;
	u_char ltoken = ifp->ltoken;

	/* already known? */
	if (!IS_ANYADDR6(pp->peer) && !IS_LINKLADDR6(pp->peer))
		return;

	/* try a router */
	FORALLRP(rp, ifp)
		if (!IS_ANYADDR6(rp->addr.sin6_addr))
			break;
	if (rp != NULL) {
		piid = &rp->addr.sin6_addr;
		goto fake_piid;
	}

	/* try a prefix */
	FORALLPP(pp1, ifp)
		if (!IS_ANYADDR6(pp1->peer))
			break;
	if (pp1 != NULL) {
		piid = &pp1->peer;
		goto fake_piid;
	}
			
	/* try the hard way */
    retry:
	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, ifp->name, IFNAMSIZ);
	if (!lastchance) {
		struct sockaddr_in6 *sin;

		/* try link-local prefix first */
		sin = (struct sockaddr_in6 *)&ifr.addr;
		sin->sin6_len = sizeof(struct sockaddr_in6);
		sin->sin6_family = AF_INET6;
		sin->sin6_addr.s6_addr[0] = 0xfe;
		sin->sin6_addr.s6_addr[1] = 0x80;

		ifr.prefixlen = 10;
		ifr.flags = IFLR_PREFIX;
	}

	if (ioctl(u, SIOCGLIFADDR, (caddr_t)&ifr) < 0) {
		if (!lastchance) {
			lastchance++;
			goto retry;
		}
		logerr("ioctl (SIOCGLIFADDR) on interface %s", ifp->name);
		abort();
	}
	piid = &((struct sockaddr_in6 *)&ifr.dstaddr)->sin6_addr;
	if (verbose)
		logtrace("get peer %s on interface %s",
			 inet6pr(piid), ifp->name);

    fake_piid:
	if (verbose)
		logtrace("fake peer IID from %s for %s on %s length %d",
			 inet6pr(piid), inet6pr(&pp->pref),
			 ifp->name, ltoken);

	COPY_ADDR6(pp->addr, pp->peer);
	bcopy(&piid->s6_addr[16 - ltoken],
	      &pp->peer.s6_addr[16 - ltoken],
	      ltoken);
}

void
mk_local_peer(pp)
	register struct prefix *pp;
{
	struct interface *ifp = pp->ifp;
	struct in6_addr *peer = &pp->peer;
	struct router *rp;
	struct if_laddrreq ifr;
	int lastchance = 0;
	u_char ltoken = ifp->ltoken;

	/* already known? */
	if (!IS_ANYADDR6(pp->peer))
		return;

	/* try a router */
	FORALLRP(rp, ifp)
		if (!IS_ANYADDR6(rp->addr.sin6_addr))
			break;
	if (rp != NULL) {
		COPY_ADDR6(rp->addr.sin6_addr, *peer);
		goto make_llocal;
	}

    retry:
	/* try a prefix */
	if (lastchance) {
		FORALLPP(pp, ifp)
			if (!IS_ANYADDR6(pp->peer))
				break;
		if (pp != NULL) {
			COPY_ADDR6(pp->peer, *peer);
			goto make_llocal;
		}
	}
			
	/* try the hard way */
	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, ifp->name, IFNAMSIZ);
	if (!lastchance) {
		struct sockaddr_in6 *sin;

		/* try link-local prefix first */
		sin = (struct sockaddr_in6 *)&ifr.addr;
		sin->sin6_len = sizeof(struct sockaddr_in6);
		sin->sin6_family = AF_INET6;
		sin->sin6_addr.s6_addr[0] = 0xfe;
		sin->sin6_addr.s6_addr[1] = 0x80;

		ifr.prefixlen = 10;
		ifr.flags = IFLR_PREFIX;
	}

	if (ioctl(u, SIOCGLIFADDR, (caddr_t)&ifr) < 0) {
		if (!lastchance) {
			lastchance++;
			goto retry;
		}
		logerr("ioctl (SIOCGLIFADDR) on interface %s", ifp->name);
		abort();
	}
	COPY_ADDR6(((struct sockaddr_in6 *)&ifr.dstaddr)->sin6_addr, *peer);
	if (verbose)
		logtrace("get peer %s on interface %s",
			 inet6pr(peer), ifp->name);

    make_llocal:
	if (IS_LINKLADDR6(*peer))
		return;

	if (verbose)
		logtrace("make fake link-local from %s on %s length %d",
			 inet6pr(peer), ifp->name, ltoken);
	bzero(&peer->s6_addr[0], 16 - ltoken);
	peer->s6_addr[0] = 0xfe;
	peer->s6_addr[1] = 0x80;
}
#endif	/* !NDPD_ROUTER */

void
mk_pref(pp)
	register struct prefix *pp;
{
	mk_mask(&pp->pmask, pp->plen);
	apply_mask(&pp->pref, &pp->pmask);
}

void
mk_mask(addr, len)
	register struct in6_addr *addr;
	register int len;
{
	register u_char *cp;

	CLR_ADDR6(*addr);
	cp = (u_char *)addr;
	for (cp = (u_char *)addr; len > 7; len -= 8)
		*cp++ = 0xff;
	if (len)
		*cp = 0xff << (8 - len);
}

/* return 0 on error */
u_char
get_prefixlen(sin)
	struct sockaddr_in6 *sin;	/* family=AF_UNSPEC, may be truncated */
{
	register u_char *cp;
	register int byte, bit, len;

#define FBEGIN	((int)&(((struct sockaddr_in6 *)0)->sin6_addr))
#define FEND	(FBEGIN + sizeof(struct in6_addr))
	if (sin->sin6_len < FBEGIN ||
	    sin->sin6_len > sizeof(struct sockaddr_in6))
		return(0);

	cp = (u_char *)sin + FBEGIN;

	len = 0;
	for (byte = FBEGIN;
	     (byte < sin->sin6_len) && (byte < FEND);
	     byte++, cp++, len += 8)
		if (*cp != 0xff)
			break;
	if ((byte == sin->sin6_len) || (byte == FEND))
		return((u_char)len);

	for (bit = 7; bit > 0; bit--, len++)
		if (!(*cp & (1 << bit)))
			break;
	for (; bit > 0; bit--)
		if (*cp & (1 << bit))
			return(0);

	byte++; cp++;
	for (; (byte < sin->sin6_len) && (byte < FEND); byte++, cp++)
		if (*cp)
			return(0);
	return((u_char)len);
}

/* Internet checksum */
int
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	u_short odd_byte = 0;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&odd_byte) = *(u_char *)w;
		sum += odd_byte;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

static char p_buf[NPR6BUF][INET6_ADDRSTRLEN];
static int p_pos = 0;
#define PR_NXTBUF	p_buf[p_pos = (p_pos + 1) % NPR6BUF]
#define PR_CURBUF	p_buf[p_pos]

char *
mac2ascii(addr, len)
	register caddr_t addr;
	int len;
{
	char *c = PR_NXTBUF;

	if (len <= 0 || len == NOPHY)
		sprintf(c, "NOMAC");
	else {
		if (len >= MAXTOKEN)
			len = MAXTOKEN;
		while (len > 0) {
			sprintf(c, "%02x:", *((u_char *)addr));
			addr++;
			c += 3;
			len--;
		}
		c[-1] = '\0';
	}
	return PR_CURBUF;
}

const char *
inet6pr(addr)
	register struct in6_addr *addr;
{
	return inet_ntop(AF_INET6, addr, PR_NXTBUF, INET6_ADDRSTRLEN);
}

const char *
tvpr(tvp)
	register struct timeval *tvp;
{
	struct timeval delta;
	char *c = PR_NXTBUF;

	if (tvp == NULL)
		sprintf(c, "NOTV");
	else {
		delta.tv_sec = now.tv_sec - (tvp->tv_sec + 1);
		delta.tv_usec = (now.tv_usec + 1000000) - tvp->tv_usec;
		NORMTIME(delta);
		snprintf(c, INET6_ADDRSTRLEN,
			 "%ld.%06ld [%ld.%06ld]",
			 tvp->tv_sec, tvp->tv_usec,
			 delta.tv_sec, delta.tv_usec);
	}
	return PR_CURBUF;
}

const char *
lifepr(v)
	u_int32_t v;
{
	char *p = PR_NXTBUF;

	if (v == INFLIFE)
		strcpy(p, "infinity");
	else if (v >= 3600) {
		if (v >= 7*24*3600) {
			p += sprintf(p, "%dw", v/(7*24*3600));
			v = v % (7*24*3600);
		}
		if (v >= 24*3600) {
			p += sprintf(p, "%dd", v/(24*3600));
			v = v % (24*3600);
		}
		if (v)
			sprintf(p, "%dh%dm%ds", v/3600, (v%3600)/60, v%60);
	} else
		sprintf(p, "%d", v);
	return PR_CURBUF;
}

#ifdef NDPD_ROUTER

const char *
datelifepr(type, val)
	int type;
	union datelife *val;
{
	if (type) {	/* @CCYYMMDDhhmm[.SS] */
		strftime(PR_NXTBUF, INET6_ADDRSTRLEN,
			 "@%Y%m%d%H%M.%S", localtime(&val->date));
		return PR_CURBUF;
	}
	return lifepr(val->life);
}

#endif	/* NDPD_ROUTER only */

#undef PR_NXTBUF
#undef PR_CURBUF

static int logging = 0;

void
initlog()
{
	logging++;
	openlog("ndpd", LOG_PID | LOG_CONS, LOG_DAEMON);
}

static void
logdate(f)
	FILE *f;
{
	char *timestamp;

	if (!timestamps)
		return;
	(void)gettimeofday(&now, NULL);
	timestamp = ctime(&now.tv_sec);
	timestamp[19] = '\0';
	(void) fprintf(f, "%s.%06ld: ", timestamp + 11, now.tv_usec);
}

void
#if __STDC__
logerr(char *fmt, ...)
#else
logerr(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	va_list ap;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (logging)
		vsyslog(LOG_ERR, fmt, ap);
	else {
		logdate(stderr);
		(void) vfprintf(stderr, fmt, ap);
		(void) fprintf(stderr, "\n");
	}
}

void
#if __STDC__
logtrace(char *fmt, ...)
#else
logtrace(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	va_list ap;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (logging)
		vsyslog(LOG_INFO, fmt, ap);
	else {
		logdate(stdout);
		(void) vfprintf(stdout, fmt, ap);
		printf("\n");
	}
}

void
#if __STDC__
logdebug(char *fmt, ...)
#else
logdebug(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	va_list ap;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (logging)
		vsyslog(LOG_DEBUG, fmt, ap);
	else {
		logdate(stdout);
		(void) vfprintf(stdout, fmt, ap);
		printf("\n");
	}
}

void
logperror(str)
	char *str;
{
	if (logging)
		syslog(LOG_ERR, "%s: %m", str);
	else {
		logdate(stderr);
		(void) fprintf(stderr, "%s: %s\n", str, strerror(errno));
	}
}

#ifdef TRACE

#define RND(l)	{ if ((l) == 0) (l) = 1; (l) = ((l) + 3) / 4; }
void
DUMPS(p)
	u_int32_t *p;
{
	struct sockaddr *s = (struct sockaddr *)p;
	int l = s->sa_len;

	RND(l);
	if (l-- > 0)
		printf("%08x", (int)htonl(*p++));
	while (l-- > 0)
		printf(" %08x", (int)htonl(*p++));
	printf("\n");
}

static const char *rtan[] = {
	"DST", "GATEWAY", "NETMASK", "GENMASK", "IFP",
	"IFA", "AUTHOR", "BRD", "DOWNSTREAM",
};

void
DUMPSP(addrs, pa, len)
	int addrs, len;
	u_int32_t *pa;
{
	int b, l, i;
	u_int32_t *p;
	char vb[80];
#define P(t)	((struct t *)p)

	printf("addrs");
	for (b = 0, l = len, p = pa; b < RTAX_MAX; b++) {
		if ((addrs & (1 << b)) == 0)
			continue;
		addrs &= ~(1 << b);
		printf(" %s", rtan[b]);
		if (l <= 0)
			continue;
		switch (P(sockaddr)->sa_family) {
		case AF_INET:
			inet_ntop(AF_INET,
				  &P(sockaddr_in)->sin_addr,
				  vb, sizeof vb);
			printf("(%s)", vb);
			break;
		case AF_INET6:
			inet_ntop(AF_INET6,
				  &P(sockaddr_in6)->sin6_addr,
				  vb, sizeof vb);
			printf("(%s", vb);
			if (P(sockaddr_in6)->sin6_scope_id)
				printf("#%d", P(sockaddr_in6)->sin6_scope_id);
			printf(")");
			break;
		case AF_LINK:
			printf("(%s#%d)",
				link_ntoa(P(sockaddr_dl)),
				P(sockaddr_dl)->sdl_index);
		}
		i = P(sockaddr)->sa_len;
		RND(i);
		p += i;
		l -= 4 * i;
	}
	if (addrs)
		printf(" 0x%x(???)", addrs);
	printf("\n");

	for (l = len, p = pa; l > 0; ) {
		DUMPS(p);
		i = P(sockaddr)->sa_len;
		RND(i);
		p += i;
		l -= 4 * i;
	}
#undef P
}

void
DUMPRT(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct rt_msghdr *r = (struct rt_msghdr *)ad;
	u_int32_t *p;
	int len = l - sizeof(*r);

	printf("route(%s) ty=%d l=%d(%d) idx=%d fl=%x "
	       "add=%x pid=%d/%d err=%d us=%d mtx=%lx\n",
		nom, r->rtm_type, r->rtm_msglen, l, r->rtm_index, r->rtm_flags,
		r->rtm_addrs,
		r->rtm_pid, r->rtm_seq, r->rtm_errno,
		r->rtm_use, r->rtm_inits);
	p = (u_int32_t *)&(r->rtm_rmx);
	printf("%08x %08x %08x %08x %08x\n%08x %08x %08x %08x %08x\n",
	       *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++);
	p = (u_int32_t *)(ad + sizeof(*r));
	DUMPSP(r->rtm_addrs, p, len);
}

void
DUMPA(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct ifa_msghdr *r = (struct ifa_msghdr *)ad;
	u_int32_t *p;
	int len = l - sizeof(*r);

	printf("ifa(%s) ty=%d l=%d(%d) idx=%d fl=%x add=%x mtx=%x\n",
		nom, r->ifam_type, r->ifam_msglen, l, r->ifam_index,
		r->ifam_flags, r->ifam_addrs, r->ifam_metric);
	p = (u_int32_t *)(ad + sizeof(*r));
	DUMPSP(r->ifam_addrs, p, len);
}

void
DUMPI(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct if_msghdr *r = (struct if_msghdr *)ad;
	u_int32_t *p;
	int len = l - sizeof(*r);

	printf("if(%s) ty=%d l=%d(%d) idx=%d fl=%x add=%x\n",
		nom, r->ifm_type, r->ifm_msglen, l, r->ifm_index, r->ifm_flags,
		r->ifm_addrs);
	p = (u_int32_t *)&(r->ifm_data);
	printf("%08x %08x %08x %08x %08x\n%08x %08x %08x %08x %08x\n",
	       *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++);
	printf("%08x %08x %08x %08x %08x\n%08x %08x %08x %08x %08x\n",
	       *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++);
	p = (u_int32_t *)(ad + sizeof(*r));
	DUMPSP(r->ifm_addrs, p, len);
}

void
DUMPX(p, l)
	u_int32_t *p;
	int l;
{
	int i = 0;

	while (l > 0) {
		printf("%08x ", *p++);
		l--;
		i++;
		if (i == 8) {
			i = 0;
			printf("\n");
		}
	}
	if (i)
		printf("\n");
}

void
DUMP(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct rt_msghdr *r = (struct rt_msghdr *)ad;

	if (r->rtm_type == RTM_IFINFO)
		DUMPI(nom, ad, l);
	else if (r->rtm_type == RTM_NEWADDR || r->rtm_type == RTM_DELADDR)
		DUMPA(nom, ad, l);
	else DUMPRT(nom, ad, l);
}

#endif /* TRACE */

