/* Discovery for home agent */

#ifdef HOME_AGENT

#include "ndpd.h"
#include "mobility.h"

static struct in6_addr *ha_any_addr(struct prefix *);

/* Compute the home agent anycast address from a prefix */

static struct in6_addr *
ha_any_addr(pp)
	struct prefix *pp;
{
	static struct in6_addr addr;

	COPY_ADDR6(pp->pref, addr);
	addr.s6_addr[8] = 0xff;
	addr.s6_addr[8] ^= pp->addr.s6_addr[8] & 0x02;
	addr.s6_addr[9] = 0xff;
	addr.s6_addr[10] = 0xff;
	addr.s6_addr[11] = 0xff;
	addr.s6_addr[12] = 0xff;
	addr.s6_addr[13] = 0xff;
	addr.s6_addr[14] = 0xff;
	addr.s6_addr[15] = 0x7e;
	return(&addr);
}

/* Add an anycast home agent addresses */

void
add_anycast_agent(pp)
	struct prefix *pp;
{
	struct interface *ifp = pp->ifp;
	struct in6_addr *addr = ha_any_addr(pp);
	register struct anycast *ap;

	if (Calloc(ap, anycast) == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}

	TAILQ_INIT(&ap->presps);
	ap->prefix = pp;
	ap->ifp = ifp;
	ap->type = IP6ANY_ANYCAST;
	COPY_ADDR6(*addr, ap->addr);
	TAILQ_INSERT_TAIL(&pp->anycasts, ap, any_list);

	if (verbose)
		logtrace("add home agent anycast %s on %s",
			 inet6pr(addr), pp->ifp->name);

	add_anycast(ap);
	pp->hacnt = 0;
}

/* Delete an anycast home agent address */

void
del_anycast_agent(pp)
	struct prefix *pp;
{
	struct in6_addr *addr = ha_any_addr(pp);
	register struct anycast *ap;
	register struct hagent *hap;

	if (verbose)
		logtrace("delete home agent anycast %s on %s",
			 inet6pr(addr), pp->ifp->name);

	FORALLPANY(ap, pp)
		if (IN6_ARE_ADDR_EQUAL(&ap->addr, addr))
			break;
	if (ap)
		del_anycast(ap);
	else if (verbose)
		logerr("can't find the home agent anycast to delete");

	while ((hap = pp->hagents.tqh_first) != NULL)
		del_hagent(hap);
	pp->hacnt = -1;
}

/* Receive a possible home agent discovery request */

void
recv_ha_disc(ip, cp, careof, ap)
	struct ip6_hdr *ip;
	u_char *cp;
	struct in6_addr *careof;
	struct anycast *ap;
{
	struct prefix *pp = ap->prefix;
	struct in6_addr *tmp;
	struct mobile mo;
	u_char flags;
	u_short seq;

	/* Is destination an anycast home agent address? */

	if (pp == NULL)
		return;
	tmp = ha_any_addr(pp);
	if (!IN6_ARE_ADDR_EQUAL(&ap->addr, tmp))
		return;

	/* Fix care-of */

	GET8(flags, cp);
	cp++;
	GET16(seq, cp);
	cp += 4;

	if (verbose)
		logtrace("received home agent discovery from "
			 "(home) %s for (care-of) %s seq %u",
			 inet6pr(&ip->ip6_src), inet6pr(careof), seq);

	/* Fill a temporary mobile entry */

	bzero(&mo, sizeof(mo));

	COPY_ADDR6(pp->addr, mo.me);
	COPY_ADDR6(ip->ip6_src, mo.home);
	COPY_ADDR6(*careof, mo.careof);
	mo.seq = seq;

	send_ha_disc(&mo, pp);
}

/* Update a home agent entry */

void
update_hagent(pp, aip, addr)
	struct prefix *pp;
	struct nd_opt_agti *aip;
	struct in6_addr *addr;
{
	struct hagent *hap, *hap1;
	struct timeout *tp = NULL;

	if (verbose)
		logtrace("update home agent %s on %s",
			 inet6pr(addr), pp->ifp->name);

	/* find it */

	FORALLHA(hap, pp)
		if (IN6_ARE_ADDR_EQUAL(&hap->addr, addr))
			break;
	if (hap != NULL) {
		if ((hap->tp == NULL) ||
		    (hap->hapref != aip->nd_opt_hai_pref) ||
		    (hap->halife != aip->nd_opt_hai_life)) {
			if (verbose)
				logtrace("home agent %s has changed",
					 inet6pr(addr));
			del_hagent(hap);
		} else
			goto dotimeout;
	}

	/* lifetime zero is delete */

	if (aip->nd_opt_hai_life == 0) {
		if (verbose)
			logtrace("home agent %s has zero lifetime",
				 inet6pr(addr));
		return;
	}

	/* allocate a new home agent */

	if (Calloc(tp, timeout) == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}

	if (Calloc(hap, hagent) == NULL) {
		if (verbose)
			logerr("out of memory");
		free(tp);
		return;
	}

	COPY_ADDR6(*addr, hap->addr);
	hap->prefix = pp;
	hap->tp = tp;
	hap->hapref = aip->nd_opt_hai_pref;
	hap->halife = aip->nd_opt_hai_life;

	/* insert in the list */

	FORALLHA(hap1, pp)
		if ((hap1->hapref < hap->hapref) ||
		    ((hap1->hapref == hap->hapref) &&
		     (hap1->halife < hap->halife)))
			break;
	if (hap1) {
		TAILQ_INSERT_BEFORE(hap1, hap, ha_list);
		if (verbose)
			logtrace("insert home agent %s before %s",
				 inet6pr(addr), inet6pr(&hap1->addr));
	} else {
		TAILQ_INSERT_TAIL(&pp->hagents, hap, ha_list);
		if (verbose)
			logtrace("insert home agent %s at tail",
				 inet6pr(addr));
	}
	pp->hacnt++;

	/* remove after lifetime */

    dotimeout:

	if (tp == NULL) {
		hap->tp->refcnt++;
		kill_timeout(hap->tp);
	}

	set_timeout(hap->tp, (int)hap->halife, 0, (func_to) del_hagent, hap);

	if (tp == NULL)
		hap->tp->refcnt--;

	if (verbose)
		logtrace("home agent %s will be cleared in %d seconds",
			 (int)hap->halife);
}

/* Delete a home agent entry */

void
del_hagent(hap)
	struct hagent *hap;
{
	struct prefix *pp = hap->prefix;

	if (hap->tp) {
		kill_timeout(hap->tp);
		hap->tp = NULL;
	}
	TAILQ_REMOVE(&pp->hagents, hap, ha_list);
	free(hap);
	pp->hacnt--;
}
#endif
