/* Home registration processing */

#ifdef HOME_AGENT

#include "ndpd.h"
#include "mobility.h"

static void schedule_req(struct mobile *);
static int change_client(struct mobile *, struct in6_addr *);

/* Receive a home registration in a binding update */

void
recv_home_reg(ip, cp, careof)
	struct ip6_hdr *ip;
	u_char *cp;
	struct in6_addr *careof;
{
	register struct mobile *mop;
	u_char flags, plen, status = OPT6_BNDA_ACCEPT;
	u_short seq;
	u_int life;
	struct in6_addr *home, *me;
	struct timeout *tp;
	struct prefix *pp;

#define sendnack(x)	{ status = (x); goto sendnack; }

	(void)gettimeofday(&now, NULL);

	/* decode home registration */

	home = &ip->ip6_src;
	me = &ip->ip6_dst;
	GET8(flags, cp);
	GET8(plen, cp);
	GET16(seq, cp);
	GET32(life, cp);

	if (verbose) {
		logtrace("received home registration from "
			 "(home) %s for (care-of) %s",
			 inet6pr(home), inet6pr(careof));
		logtrace("flags=%x [%s] plen=%d seq=%u life=%u",
			 flags, bnd_upd_flg(flags),
			 plen, seq, life);
	}

	/* allocate a timeout */

	if (Calloc(tp, timeout) == NULL) {
		if (verbose)
			logerr("out of memory");
		/* simulate a packet loss ??? */
		return;
	}

	/* look for a mobile entry */

	FORALLMOB(mop)
		if (IN6_ARE_ADDR_EQUAL(home, &mop->home) &&
		    IN6_ARE_ADDR_EQUAL(me, &mop->me))
			break;

	if (mop) {
		/* got a mobile entry */

		if (verbose)
			logtrace("find mop %x", (u_long)mop);

		/* check sequence number */

		if (!SEQ_GT(seq, mop->seq)) {
			if (verbose)
				logerr("bad sequence (%d !GT %d)",
				       (int)seq, (int)(mop->seq));
			free(tp);
			return;
		}
		mop->seq = seq;
		mop->refresh = now.tv_sec;

		/* cleanup cases */

		if (IN6_ARE_ADDR_EQUAL(home, careof)) {
			if (verbose)
				logtrace("cleanup (same addresses)");
			if (mop->flags & OPT6_BNDU_HR)
				del_client(mop);
			else
				clean_bindings(mop);
			mop = NULL;
			sendnack(OPT6_BNDA_ACCEPT);
		} else if (life == 0) {
			if (verbose)
				logtrace("cleanup (life == 0)");
			if (mop->flags & OPT6_BNDU_HR)
				del_client(mop);
			else
				clean_bindings(mop);
			mop = NULL;
			sendnack(OPT6_BNDA_ACCEPT);
		}

		/* not a client */

		if ((mop->flags & OPT6_BNDU_HR) == 0) {
			clean_bindings(mop);
			goto alloc;
		}

		/* changed ? */

		if (!IN6_ARE_ADDR_EQUAL(careof, &mop->careof)) {
			status = change_client(mop, careof);
			if (status >= OPT6_BNDA_UNSPEC) {
				clean_bindings(mop);
				goto alloc;
			}
			change_binding(mop);
		}

		mop->flags = flags;
		mop->lifetime = life;
		mop->plen = plen;

		/* cancel any timeout */

		if (mop->tp) {
			kill_timeout(mop->tp);
			mop->tp = NULL;
		}

		/* go to common process */

		goto dotimeout;

	}

	/* got no mobile entry */

	/* cleanup cases */

	if (IN6_ARE_ADDR_EQUAL(home, careof)) {
		if (verbose)
			logtrace("cleanup (same addresses)");
		sendnack(OPT6_BNDA_ACCEPT);
	} else if (life == 0) {
		if (verbose)
			logtrace("cleanup (life == 0)");
		sendnack(OPT6_BNDA_ACCEPT);
	}

	/* new client */

    alloc:
	if (verbose)
		logtrace("allocate client");

	if (!is_router()) {
		logerr("can't forward: reject home registration");
		sendnack(OPT6_BNDA_UNSUP);
	}

	if (Calloc(mop, mobile) == NULL) {
		if (verbose)
			logerr("out of memory");
		sendnack(OPT6_BNDA_RES);
	}

	/* fill in */

	COPY_ADDR6(*home, mop->home);
	COPY_ADDR6(*careof, mop->careof);
	COPY_ADDR6(*me, mop->me);
	mop->flags = flags;
	mop->refresh = now.tv_sec;
	mop->lifetime = life;
	mop->seq = seq;
	mop->plen = plen;

	/* locate it */

	pp = loc_client(mop);
	if (pp == NULL) {
		if (verbose)
			logtrace("no prefix for the client %s/%d",
				 inet6pr(home), plen);
		sendnack(OPT6_BNDA_SUBNET);
	}
	mop->pp = pp;

	/* get security */

	if (!get_security(mop))
		sendnack(OPT6_BNDA_ADMIN);

	/* set tunnel */

	status = add_tunnel(mop);
	if (status >= OPT6_BNDA_UNSPEC) {
		del_security(mop);
		goto sendnack;
	}

	/* set binding */

	add_binding(mop);

	/* done */

	TAILQ_INSERT_TAIL(&mobiles, mop, mb_list);

   dotimeout:

	/* request/timeout (mop->tp is always NULL here) */

	if (mop->lifetime != INFLIFE) {
		/* schedule a binding request */

		set_timeout(tp, (2 * mop->lifetime) / 3, 0,
			    (func_to) schedule_req, mop);
		mop->tp = tp;
	} else
		free(tp);

	/* send an acknowledge (should be wanted) */

	if (flags & OPT6_BNDU_ACK)
		send_bnd_ack(mop, status, mop->lifetime / 2);

	return;

    sendnack:

	free(tp);

	if (flags & OPT6_BNDU_ACK) {
		if (mop == NULL) {
			struct mobile mo;

			bzero(&mo, sizeof(mo));

			COPY_ADDR6(*me, mo.me);
			COPY_ADDR6(*home, mo.home);
			COPY_ADDR6(*careof, mo.careof);
			mo.flags = flags;
			mo.seq = seq;

			send_bnd_ack(&mo, status, 0);
		} else
			send_bnd_ack(mop, status, 0);
	}

	switch (status) {
	    case OPT6_BNDA_ACCEPT:
	    case OPT6_BNDA_SUBNET:
	    case OPT6_BNDA_ADMIN:
		if (mop)
			free(mop);
	}
}

/* Schedule a binding request */

static void
schedule_req(mop)
	struct mobile *mop;
{
	(void)gettimeofday(&now, NULL);

	/* cancel timeout */

	if (mop->tp == NULL) {
		if (verbose)
			logerr("schedule without timeout?");
		return;
	}

	mop->tp->refcnt++;	/* keep timeout structure */
	kill_timeout(mop->tp);

	/* send the binding request */

	send_bnd_req(mop);

	/* schedule final clean up */

	mop->tp->refcnt--;
	set_timeout(mop->tp, mop->lifetime / 3, 0, (func_to) del_client, mop);
}

/* Delete a client */

void
del_client(mop)
	struct mobile *mop;
{
	/* cancel timeout */

	if (mop->tp) {
		kill_timeout(mop->tp);
		mop->tp = NULL;
	}

	/* delete binding */

	del_binding(mop);

	/* delete tunnel */

	del_tunnel(mop);

	/* delete security */

	del_security(mop);

	/* free... */

	TAILQ_REMOVE(&mobiles, mop, mb_list);

	if (verbose)
		logtrace("delete client for %s", inet6pr(&mop->home));

	free(mop);
}

/* Change a client */

static int
change_client(mop, newca)
	struct mobile *mop;
	struct in6_addr *newca;
{
	int cc;

	del_tunnel(mop);
	COPY_ADDR6(*newca, mop->careof);
	cc = add_tunnel(mop);
	return(cc);
}

#endif
