/* Manage tunnels to mobiles */

#ifdef HOME_AGENT

#include "ndpd.h"
#include "mobility.h"

#define TNNM	"tug%d"

static void del_troute(struct mobile *);

/* Init tunnel list */

void
init_tunnels()
{
	struct tunnel *tup;
	struct ifreq ifr;
	int unit, d, cc;

	/* UDP/IPv6 socket for interface ioctls */
	if ((d = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		logperror("socket (home-agent)");
		return;
	}

	for (unit = 0;; unit++) {
		bzero(&ifr, sizeof(ifr));
		snprintf(ifr.ifr_name, IFNAMSIZ, TNNM, unit);

		cc = ioctl(d, SIOCGIFFLAGS, &ifr);

		if (cc < 0) {
			if (errno != ENXIO)
				logperror("ioctl (SIOCGIFFLAGS)");
			(void) close(d);
			return;
		}

		if (Calloc(tup, tunnel) == NULL) {
			if (verbose)
				logerr("out of memory");
			continue;
		}
		strncpy(tup->name, ifr.ifr_name, IFNAMSIZ);
		if (unit == 0)
			tup->state = TUN_RES;
		else if ((ifr.ifr_flags & (IFF_UP | IFF_RUNNING)) != 0)
			tup->state = TUN_USED;
		else
			tup->state = TUN_FREE;
		if (tup->state > TUN_FREE)
			TAILQ_INSERT_TAIL(&tunnels, tup, tn_list);
		else
			TAILQ_INSERT_HEAD(&tunnels, tup, tn_list);
	}
}

/* Add a tunnel */

int
add_tunnel(mop)
	struct mobile *mop;
{
	struct tunnel *tup;
	struct if_laddrreq ifr;
	struct anycast *ap;
	struct sockaddr_in6 *sin;
	int d, retried = 2;

	tup = tunnels.tqh_first;
	if ((tup == NULL) || (tup->state != TUN_FREE)) {
		if (verbose)
			logtrace("no free tunnels");
		return(OPT6_BNDA_RES);
	}

	/* allocate proxy entry */
	if (Calloc(ap, anycast) == NULL) {
		if (verbose)
			logerr("out of memory");
		return(OPT6_BNDA_RES);
	}
	TAILQ_INIT(&ap->presps);
	ap->prefix = mop->pp;
	ap->ifp = mop->pp->ifp;
	COPY_ADDR6(mop->home, ap->addr);
	if (mop->flags & OPT6_BNDU_RT)
		ap->type = IP6ANY_ROUTER_PROXY;
	else
		ap->type = IP6ANY_HOST_PROXY;
	ap->ntosend = MAX_ADVERT_REXMIT;
	mop->ap = ap;

	/* UDP/IPv6 socket for interface ioctls */
	if ((d = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		logperror("socket (home-agent)");
		mop->ap = NULL;
		free(ap);
		return(OPT6_BNDA_UNSPEC);
	}

	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, tup->name, IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(mop->careof, sin->sin6_addr);

	if (ioctl(d, SIOCSTUGLADDR, &ifr) < 0) {
		logperror("ioctl (SIOCSTUGLADDR)");
		mop->ap = NULL;
		free(ap);
		(void) close(d);
		return(OPT6_BNDA_UNSPEC);
	}

    retry:

	del_troute(mop);

	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, tup->name, IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(mop->me, sin->sin6_addr);
	sin = (struct sockaddr_in6 *)&ifr.dstaddr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(mop->home, sin->sin6_addr);
	ifr.prefixlen = 128;

	if (ioctl(d, SIOCALIFADDR, &ifr) < 0) {
		logperror("ioctl (SIOCALIFADDR)");
		if (--retried >= 0)
			goto retry;
		mop->ap = NULL;
		free(ap);
		(void) close(d);
		return(OPT6_BNDA_UNSPEC);
	}
	(void) close(d);
	tup->state = TUN_USED;
	TAILQ_REMOVE(&tunnels, tup, tn_list);
	TAILQ_INSERT_TAIL(&tunnels, tup, tn_list);
	mop->tup = tup;

	/* finish proxy stuff */
	TAILQ_INSERT_TAIL(&mop->pp->anycasts, ap, any_list);
	add_anycast(ap);

	return(OPT6_BNDA_ACCEPT);
}

/* Delete a tunnel */

void
del_tunnel(mop)
	struct mobile *mop;
{
	struct tunnel *tup = mop->tup;
	struct ifreq ifr;
	struct if_laddrreq ifr6;
	struct sockaddr_in6 *sin;
	int d;

	/* do proxy stuff */
	if (mop->ap)
		del_anycast(mop->ap);
	mop->ap = NULL;

	if ((d = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		logperror("socket (home-agent)");
		return;
	}

	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, tup->name, IFNAMSIZ);

	if (ioctl(d, SIOCGIFFLAGS, &ifr) < 0) {
		logperror("ioctl (SIOCGIFFLAGS)");
		(void) close(d);
		return;
	}

	ifr.ifr_flags &= ~IFF_UP;

	if (ioctl(d, SIOCSIFFLAGS, &ifr) < 0) {
		logperror("ioctl (SIOCSIFFLAGS)");
		(void) close(d);
		return;
	}

	bzero(&ifr6, sizeof(ifr6));
	strncpy(ifr6.iflr_name, tup->name, IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&ifr6.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(mop->me, sin->sin6_addr);

	if (ioctl(d, SIOCDLIFADDR, &ifr6) < 0) {
		logperror("ioctl (SIOCDLIFADDR)");
		(void) close(d);
		return;
	}

	bzero(&ifr6, sizeof(ifr6));
	strncpy(ifr6.iflr_name, tup->name, IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&ifr6.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(in6addr_any, sin->sin6_addr);

	if (ioctl(d, SIOCSTUGLADDR, &ifr6) < 0) {
		logperror("ioctl (SIOCSTUGLADDR)");
		(void) close(d);
		return;
	}
	(void) close(d);

	del_troute(mop);

	tup->state = TUN_FREE;
	TAILQ_REMOVE(&tunnels, tup, tn_list);
	TAILQ_INSERT_HEAD(&tunnels, tup, tn_list);
	mop->tup = NULL;
}

/* Delete a cloned route to the mobile */

static void
del_troute(mop)
	struct mobile *mop;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_flags = RTF_UP | RTF_HOST;
	rtm.rtm_addrs = RTA_DST;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(mop->home, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("del_troute (mobility)");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose && (errno != ESRCH))
			logperror("write (del_troute in mobility)");
		return;
	} else if (verbose)
		logtrace("delete route to mobile %s", inet6pr(&mop->home));
}

#endif
