/* Initializations */

#include "ndpd.h"
#include "mobility.h"

int m, b = -1;			/* Socket file descriptors */

#ifdef MOBILE
u_int bnd_lifetime = BND_LIFETIME;
u_int reg_lifetime = REG_LIFETIME;

int multi = 0;			/* not in multi-homed mode by default */
int mscope = SCP_LOCAL;		/* enforce nothing by default */
#endif

/* Initialize mobility support */

void
mobility_init()
{
	if ((m = socket(PF_INET6, SOCK_RAW, IPPROTO_DSTOPTS)) < 0) {
		logperror("socket (mobility)");
		exit(2);
	}
	(void) shutdown(m, 1);

	TAILQ_INIT(&mobiles);

#ifdef MOBILE
	if (!mobile)
		return;
	TAILQ_INIT(&correspondents);
	read_mobile(0);
#endif

#ifdef HOME_AGENT
	TAILQ_INIT(&tunnels);
	init_tunnels();
#endif
}

/* called when the default route is here */

void
mobile_booted()
{
#ifdef MOBILE
	if (mobile)
		read_mobile(1);
#endif
}

/* called by finish() */

void
mobility_cleanup()
{
	struct mobile *mop;
#ifdef MOBILE
	struct corr *cop;
#endif

	FORALLMOB(mop)
#ifdef HOME_AGENT
		if (mop->flags & OPT6_BNDU_HR)
			del_client(mop);
		else
#endif
		clean_bindings(mop);

#ifdef MOBILE
	FORALLCOR(cop) {
		mobile_move(cop, &cop->home);
		del_corr(cop);
	}
#endif
}
		

/* called when an important prefix was zapped */

void
mobile_zappref(pp)
	struct prefix *pp;
{
#ifdef MOBILE
	if (!mobile)
		return;
	if (multi)
		multi_move(pp->ifp, pp);
	else if (pp->mflags & MOB_ISCA)
		minor_move(pp->ifp, pp);
#endif
}

/* called when an important prefix was changed */

void
mobile_chgpref(ifp)
	struct interface *ifp;
{
#ifdef MOBILE
	if (!mobile)
		return;
	if (multi)
		multi_move(ifp, NULL);
	else if (PRIMHA && (PRIMHA->state != HAST_UNKNOWN))
		minor_move(ifp, NULL);
#endif
}

/* called when a home prefix is advertised */

#ifdef MOB_ISHA

void
mobile_chgha(hap, pp)
	struct homeaddr *hap;
	struct prefix *pp;
{
	int nomove = 0, outgone = 0, nochange = 0;

	if (!mobile)
		return;

	if (pp == NULL) {
		if (verbose)
			logtrace("no home prefix for %s", inet6pr(&hap->addr));
		return;
	}

	/* for debugging: should *never* happen! */
	if (!IN6_ARE_ADDR_EQUAL(&hap->addr, &pp->addr)) {
		logerr("home prefix mismatch (%s != %s)",
		       inet6pr(&hap->addr), inet6pr(&pp->addr));
		abort();
	}
	if (hap != pp->hap) {
		logerr("home prefix mismatch (%lx != %lx)",
		       (u_long)hap, (u_long)(pp->hap));
		abort();
	}

	if (verbose)
		logtrace("mobile_chgha for %s/%d %s",
			 inet6pr(&hap->addr), pp->plen,
			 hap->state == HAST_IN ? "at home" : "in visit");

	/* special case for unknown state */

	if (hap->state == HAST_UNKNOWN) {

		if ((pp->nflags & ND_OPT_PI_FLAG_AUTO) &&
		    (pp->deplife != 0) && (pp->invlife != 0)) {
			nomove = 1;
			if (pp->deplife < DEFPREFLIFE)
				pp->deplife = DEFPREFLIFE;
			goto comein;
		}

		if (pp->nflags & ND_OPT_PI_FLAG_ONLINK) {
			if (pp->plen != 128)
				pp->mflags |= MOB_HASFL;
		} else
			pp->mflags &= MOB_HASFL;

		pp->nflags |= ND_OPT_PI_FLAG_AUTO;
		pp->invlife = INFLIFE;

		return;
	}

	/* mobile was in visit, still there? */

	if (hap->state == HAST_OUT) {

		if (((pp->nflags & ND_OPT_PI_FLAG_AUTO) == 0) ||
		    (pp->deplife == 0) || (pp->invlife == 0))
			goto invisit;

	    comein:
		/* come back to home! */
		hap->state = HAST_IN;
		pp->invlife = INFLIFE;
		pp->mflags &= ~MOB_ISFA;

		/* reordering */
		if (pp != HAPP(hap)) {
			TAILQ_REMOVE(&hap->homeprefs, pp, mob_link);
			TAILQ_INSERT_HEAD(&hap->homeprefs, pp, mob_link);
		}
#ifdef MOBILE
		if (!nomove)
			major_move(pp, 1);
#endif
		return;
	}

	/* mobile was at home, still here? */

	outgone = 1;

	if (((pp->nflags & ND_OPT_PI_FLAG_AUTO) == 0) ||
	    (pp->deplife == 0) || (pp->invlife == 0)) {

	    invisit:

		/* don't worry about fakes! */
		if (pp->mflags & MOB_ISFA) {
			outgone = 0;
			nomove = nochange = 1;
		}

		if (outgone && (IS_UP(pp->ifp) || !is_dyn(pp->ifp))) {
			mobile = 0;
			del_prefix(pp, 0);
		}
		pp->nflags = pp->pflags = ND_OPT_PI_FLAG_AUTO;
		pp->invlife = INFLIFE;
		pp->deplife = INFLIFE;
		if (pp->plen != 128) {
			COPY_ADDR6(pp->addr, pp->pref);
			mk_mask(&pp->pmask, 128);
			pp->alen = pp->plen = 128;
			pp->mflags |= MOB_ISFA;
		}
		if (outgone && (IS_UP(pp->ifp) || !is_dyn(pp->ifp))) {
			add_prefix(pp);		
			if (pp == FIRST(pp->ifp))
				prompt(pp->ifp);
			mobile = 1;
		}
		if (!nochange)
			hap->state = HAST_OUT;

#ifdef MOBILE
		if (outgone && !nomove)
			major_move(pp, 1);
#endif
		return;
	}

	
	/* mobile still at home */

	pp->invlife = INFLIFE;
	pp->mflags &= ~MOB_ISFA;
}

int
mobile_chgrt(pp, cmd, why)
	struct prefix *pp;
	int cmd, why;
{
#ifdef MOBILE
	if (!mobile)
		return (0);

	/*
	 * Sanity
	 */

	if ((cmd != RTM_ADD) && (cmd != RTM_DELETE)) {
		logerr("mobile_chgrt: unknown cmd %d on %s/%d because %d",
		       cmd, inet6pr(&pp->addr), pp->plen, why);
		return (1);
	}
	if ((why != ND_OPT_PI_FLAG_AUTO) && (why != ND_OPT_PI_FLAG_ONLINK)) {
		logerr("mobile_chgrt: %s on %s/%d because %d?",
		       cmd == RTM_ADD ? "add" : "delete",
		       inet6pr(&pp->addr), pp->plen, why);
		return (1);
	}

	/*
	 * Restriction of new prefixes
	 * (can return because one is being called again for the route)
	 */

	if ((mscope & SCP_PREFIX) &&
	    (cmd == RTM_ADD) &&
	    (why == ND_OPT_PI_FLAG_AUTO)) {
		restrict(pp, 0);
		return (0);
	}


	/*
	 * Protection for candidate care-of prefixes
	 */

	if ((pp->mflags & MOB_ISHA) == 0) {
		if (cmd == RTM_ADD)
			add_protect(pp);
		else
			del_protect(pp);
	}

	/* can't delete home prefixes */

	if ((cmd == RTM_DELETE) &&
	    (why == ND_OPT_PI_FLAG_AUTO) &&
	    (pp->mflags & MOB_ISHA)) {
		logerr("try to delete home prefix %s/%d on %s",
		       inet6pr(&pp->addr), pp->alen, pp->ifp->name);
		return (1);
	}
#endif
	return (0);
}

void
mobile_newif(ifp)
	struct interface *ifp;
{
#ifdef MOBILE
	struct homeaddr *hap;

	if (!mobile)
		return;
	FORALLHA(hap)
		if (hap->state != HAST_UNKNOWN)
			add_home_prefix(hap, ifp);
#endif
}
#endif
