/* kernel dump tools */

#include "ndpd.h"
#include "mobility.h"

#include <kvm.h>
#include <nlist.h>

static struct nlist nl[] = {
	{ "_pemobrt_head" },
	{ "_pemobha_head" },
	{ "" }
};
static struct nlist rtnl[] = {
	{ "_pemobrt_head" },
	{ "" }
};
#ifdef MOBILE
static struct nlist hanl[] = {
	{ "_pemobha_head" },
	{ "" }
};
#endif
static char buf[2048];

/*
 * Dump all the mobility output policy entries (in verbose mode)
 */

int
kvm_mobiles()
{
	kvm_t *kvmd;
	u_long pep;
	struct plentmob pe;
	struct ip6_plhead head;
	int cc, i = 0;

	if (!verbose)
		return (0);

	kvmd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, buf);
	if (kvmd == NULL) {
		logperror("kvm_openfiles");
		return (-1);
	}
	if (kvm_nlist(kvmd, nl) < 0) {
		logerr("kvm_nlist: %s", kvm_geterr(kvmd));
		return (-1);
	}
	if (nl[0].n_type == 0) {
		logerr("kvm_mobiles: no namelist");
		return (-1);
	}
    again:
	if (nl[i].n_value == 0) {
		logerr("kvm_mobiles: symbol %s not found", nl[i].n_name);
		return (-1);
	}
	cc = kvm_read(kvmd, nl[i].n_value, &head, sizeof(head));
	if (cc != sizeof(head)) {
		logerr("kvm_mobiles: can't get head [%s]: %s",
		       i == 0 ? "RT" : "HA", kvm_geterr(kvmd));
		return (-1);
	}

	logtrace("Dump of mobility entries in kernel (%s)",
		 i == 0 ? "RT" : "HA");

	for (pep = (u_long)head.ph6_head.tqh_first;
	     pep != 0;
	     pep = (u_long)pe.pem_hdr.pe6_link.tqe_next) {
		cc = kvm_read(kvmd, pep, &pe, sizeof(pe));
		if (cc != sizeof(pe)) {
			logerr("kvm_mobiles: can't get pe at %lx: %s",
			       pep, kvm_geterr(kvmd));
			return (-1);
		}

		logtrace("0x%08lx: type=%d[%s] len=%d refcnt=%d",
			 pep, pe.pem_type,
			 pe.pem_type == IP6_NHDR_RT ? "RT" : "HA",
			 pe.pem_len, pe.pem_hdr.pe6_refcnt);

		if (pe.pem_msklen == PCB_MATCHING) {
			logtrace("disabling PCB=%x", pe.pem_inp);
			continue;
		}
		logtrace("%sfilter=%s/%d",
			 pe.pem_len == 0 ? "disabling " : "",
			 inet6pr(&pe.pem_addr), pe.pem_msklen);

		if (pe.pem_len == 0)
			continue;

		if (pe.pem_type == IP6_NHDR_DOPT) {
			logtrace("care-of=%s",
			    inet6pr((struct in6_addr *)&pe.pem_opt_addr1));
			continue;
		}

		logtrace("home=%s",
			 inet6pr((struct in6_addr *)&pe.pem_opt_addr1));
		logtrace("care-of=%s",
			 inet6pr((struct in6_addr *)&pe.pem_opt_addr2));
	}
	if (i == 0) {
		i++;
		goto again;
	}

	return (0);
}

/*
 * Get the reference count of a RT mobility output policy entry
 */

int
kvm_get_rt(mop)
	struct mobile *mop;
{
	kvm_t *kvmd;
	u_long pep;
	struct plentmob pe;
	struct ip6_plhead head;
	int cc;

	kvmd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, buf);
	if (kvmd == NULL) {
		logperror("kvm_openfiles");
		return (-1);
	}
	if (kvm_nlist(kvmd, rtnl) < 0) {
		logerr("kvm_nlist: %s", kvm_geterr(kvmd));
		return (-1);
	}
	if (rtnl[0].n_type == 0) {
		logerr("kvm_get_rt: no namelist");
		return (-1);
	}
	if (rtnl[0].n_value == 0) {
		logerr("kvm_get_rt: symbol %s not found", rtnl[0].n_name);
		return (-1);
	}
	cc = kvm_read(kvmd, rtnl[0].n_value, &head, sizeof(head));
	if (cc != sizeof(head)) {
		logerr("kvm_get_rt: can't get head: %s", kvm_geterr(kvmd));
		return (-1);
	}

	for (pep = (u_long)head.ph6_head.tqh_first;
	     pep != 0;
	     pep = (u_long)pe.pem_hdr.pe6_link.tqe_next) {
		cc = kvm_read(kvmd, pep, &pe, sizeof(pe));
		if (cc != sizeof(pe)) {
			logerr("kvm_get_rt: can't get pe at %lx: %s",
			       pep, kvm_geterr(kvmd));
			return (-1);
		}

		if ((pe.pem_msklen != FULL_MATCHING) ||
		    (pe.pem_len == 0) ||
		    (pe.pem_type != IP6_NHDR_RT) ||
		    !IN6_ARE_ADDR_EQUAL(&pe.pem_addr, &mop->home))
			continue;
		return (pe.pem_hdr.pe6_refcnt);
	}
	return (0);
}

#ifdef MOBILE
/*
 * Get the reference count of a HA mobility output policy entry
 */

int
kvm_get_ha(cop)
	struct corr *cop;
{
	kvm_t *kvmd;
	u_long pep;
	struct plentmob pe;
	struct ip6_plhead head;
	int cc;

	kvmd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, buf);
	if (kvmd == NULL) {
		logperror("kvm_openfiles");
		return (-1);
	}
	if (kvm_nlist(kvmd, hanl) < 0) {
		logerr("kvm_nlist: %s", kvm_geterr(kvmd));
		return (-1);
	}
	if (hanl[0].n_type == 0) {
		logerr("kvm_get_ha: no namelist");
		return (-1);
	}
	if (hanl[0].n_value == 0) {
		logerr("kvm_get_ha: symbol %s not found", hanl[0].n_name);
		return (-1);
	}
	cc = kvm_read(kvmd, hanl[0].n_value, &head, sizeof(head));
	if (cc != sizeof(head)) {
		logerr("kvm_get_ha: can't get head: %s", kvm_geterr(kvmd));
		return (-1);
	}

	for (pep = (u_long)head.ph6_head.tqh_first;
	     pep != 0;
	     pep = (u_long)pe.pem_hdr.pe6_link.tqe_next) {
		cc = kvm_read(kvmd, pep, &pe, sizeof(pe));
		if (cc != sizeof(pe)) {
			logerr("kvm_get_ha: can't get pe at %lx: %s",
			       pep, kvm_geterr(kvmd));
			return (-1);
		}

		if ((pe.pem_msklen != FULL_MATCHING) ||
		    (pe.pem_len == 0) ||
		    (pe.pem_type != IP6_NHDR_DOPT) ||
		    !IN6_ARE_ADDR_EQUAL(&pe.pem_addr, &cop->addr))
			continue;
		return (pe.pem_hdr.pe6_refcnt);
	}
	return (0);
}
#endif
