/* Reception */

#include "ndpd.h"
#include "mobility.h"

#define MAXPACKET	8192	/* max packet size */
static char buf[MAXPACKET];

static struct in6_addr *get_altcoa(u_char *, int);
static int router_recv_destopt(struct ip6_hdr *, u_char *, struct in6_addr *);

/* Receive a destination option header */

void
recv_destopt()
{
	register struct ip6_hdr *ip;
	register struct ipv6_dopthdr *dop;
	register u_char *cp, *ep, *upd = 0;
	struct in6_addr *careof = NULL;
	struct in6_addr *altcoa = NULL;
	int len;

	if ((len = read(m, buf, sizeof(buf))) < 0) {
		if (verbose)
			logperror("read (destopt)");
		return;
	}
	if (len == 0) {
		if (verbose)
			logerr("read destopt return 0");
		return;
	}

	ip = (struct ip6_hdr *)buf;
	if (len < sizeof(*ip) + sizeof(*dop)) {
		if (verbose)
			logerr("destopt: too short (%d)", len);
		return;
	}
	if (ip->ip6_nxt != IP6_NHDR_DOPT) {
		if (verbose)
			logerr("destopt: bad nh (%d)", ip->ip6_nxt);
		return;
	}

	len -= sizeof(struct ip6_hdr);
	dop = (struct ipv6_dopthdr *)(ip + 1);

	if (((int)(dop->io6_hlen) + 1) * 8 < len)
		len = ((int)(dop->io6_hlen) + 1) * 8;

	cp = ((u_char *)dop) + 2 * sizeof(u_int8_t);
	ep = ((u_char *)dop) + len;

	while (cp < ep)
		switch (*cp) {
		    case OPT6_PAD_1:
			cp++;
			break;

		    default:
		    case OPT6_PAD_N:
			cp++;
			len = *cp++;
			cp += len;
			break;

		    case OPT6_BINDING_UPD:
			cp++;
			len = *cp++;
			/* need home and careof addresses: defer! */
			upd = cp;
			if (len > 8)
				altcoa = get_altcoa(cp + 8, len - 8); 
			cp += len;
			break;

		    case OPT6_BINDING_ACK:
			cp++;
			len = *cp++;
			recv_bnd_ack(ip, cp, len);
			cp += len;
			break;

		    case OPT6_BINDING_REQ:
			cp++;
			len = *cp++;
			recv_bnd_req(ip);
			cp += len;
			break;

		    case OPT6_HOME_ADDR:
			cp++;
			len = *cp++;
			/* addresses already swapped by kernel */
			careof = (struct in6_addr *)cp;
			cp += len;
			break;
		}

	if (cp != ep) {
		if (verbose)
			logerr("destop: garbage %d bytes", cp - ep);
		return;
	}
	if (upd) {
		if (careof == NULL) {
			if (verbose)
				logerr("destop: binding update without"
				       "home address?");
			return;
		}
		/* an alternate care-of is better */
		if (altcoa) {
			if (verbose && !IN6_ARE_ADDR_EQUAL(altcoa, careof))
				logtrace("real careof is %s", inet6pr(altcoa));
			careof = altcoa;
		}
		if (router_recv_destopt(ip, upd, careof))
			return;
		if (*upd & OPT6_BNDU_HR)
			recv_home_reg(ip, upd, careof);
		else
			recv_bnd_upd(ip, upd, careof);
	}
}

static struct in6_addr *
get_altcoa(cp, len)
	u_char *cp;
	int len;
{
	struct mob6_sopt_coa *sop;
	int solen = 0;

	while (len > 0) {
		cp += solen;
		sop = (struct mob6_sopt_coa *)cp;
		switch (sop->mscoa_type) {
		    case MOB6_SUBOPT_PAD1:
			solen = 1;
			len -= solen;
			break;
		    case MOB6_SUBOPT_PADN:
		    default:
			solen = 2 + sop->mscoa_len;
			len -= solen;
			break;
		    case MOB6_SUBOPT_COA:
			if (sop->mscoa_len != sizeof(struct in6_addr)) {
				if (verbose)
					logerr("get_altcoa: "
					       "bad alternate care-of");
				return NULL;
			}
			if (len < 2 + sop->mscoa_len) {
				if (verbose)
					logerr("get_altcoa: garbage for "
					       "alternate care-of");
				return NULL;
			}
			return (struct in6_addr *)sop->mscoa_coa;
		}
		if (len < 0) {
			if (verbose)
				logerr("get_altcoa: get garbage");
			return NULL;
		}
	}
	return NULL;
}

#ifdef NDPD_ROUTER

/* Special handler for a router because of anycasts */

static int
router_recv_destopt(ip, upd, careof)
	struct ip6_hdr *ip;
	u_char *upd;
	struct in6_addr *careof;
{
	struct interface *ifp = NULL;
	struct prefix *pp = NULL;
	struct anycast *ap = NULL;

	if (IN6_IS_ADDR_MULTICAST(&ip->ip6_dst))
		goto found;

	FORALLIFP(ifp)
	    FORALLPP(pp, ifp)
		FORALLPANY(ap, pp)
		    if (IN6_ARE_ADDR_EQUAL(&ap->addr, &ip->ip6_dst))
			goto found;
	FORALLIFP(ifp)
	    FORALLOANY(ap, ifp)
		if (IN6_ARE_ADDR_EQUAL(&ap->addr, &ip->ip6_dst))
		    goto found;
	return 0;

	/* Not really for us! */

    found:

	/* Try home agent discovery */

#ifdef HOME_AGENT
	if ((*upd & OPT6_BNDU_HR) &&
	    (ifp != NULL) && (pp != NULL) && (ap != NULL))
	    recv_ha_disc(ip, upd, careof, ap);
#endif

	/* Just drop it ? */

	return 1;
}
#else

/* Void router_recv_destopt() routine */

static int
router_recv_destopt(ip, upd, careof)
	struct ip6_hdr *ip;
	u_char *upd;
	struct in6_addr *careof;
{
	/* Drop too stupid things! */

	if (IN6_IS_ADDR_MULTICAST(&ip->ip6_dst))
		return 1;
	return 0;
}
#endif
