/* Binding acknowledgement processing */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"

/* Receive a binding acknowledgement */

void
recv_bnd_ack(ip, cp, len)
	struct ip6_hdr *ip;
	u_char *cp;
	int len;
{
	struct corr *cop;
	struct timeout *tp;
	u_char status;
	u_short seq;
	u_int life, refresh;
	char *msg;

	(void)gettimeofday(&now, NULL);

	/* decode ack */

	GET8(status, cp);
	GET16(seq, cp);
	GET32(life, cp);
	GET32(refresh, cp);

	/* decode status */

	switch (status) {
	    case OPT6_BNDA_UNSPEC:
		msg = "Reason unspecified";
		break;

	    case OPT6_BNDA_ADMIN:
		msg = "Administratively prohibited";
		break;

	    case OPT6_BNDA_RES:
		msg = "Insufficient resources";
		break;

	    case OPT6_BNDA_UNSUP:
		msg = "Home registration not supported";
		break;

	    case OPT6_BNDA_SUBNET:
		msg = "Not home subnet";
		break;

	    case OPT6_BNDA_DYNDIS:
		msg = "Dynamic home agent address discovery response";
		break;

	    case OPT6_BNDA_BADIFF:
		msg = "Incorrect interface identifier length";
		break;

	    case OPT6_BNDA_REFUSED:
		msg = "Not home agent for this host";
		break;

	    case OPT6_BNDA_DADFAIL:
		msg = "Duplicate Address Detected";
		break;

	    default:
		msg = "Unknown reason?";
	}

	if (verbose) {
		logtrace("receive binding ack from "
			 "%s seq=%d life=%u refresh=%u",
			 inet6pr(&ip->ip6_src),
			 (int)seq, life, refresh);
		if (status != OPT6_BNDA_ACCEPT)
			logtrace("binding ack failed (%d): %s", status, msg);
	}

	/* lookup for a correspondent entry */

	FORALLCOR(cop)
		if (IN6_ARE_ADDR_EQUAL(&ip->ip6_src, &cop->addr))
			break;

	if (cop == NULL) {
		/* no correspondent */

		if (verbose)
			logtrace("can't find acking correspondent %s",
				 inet6pr(&ip->ip6_src));
		return;
	}
	if (cop->seq != seq) {
		if (SEQ_GT(cop->seq, seq)) {
			if (verbose)
				logtrace("old ack (%d > %d)",
					 (int)(cop->seq), (int)seq);
			return;
		}
		if (verbose)
			logerr("ack seq error (%d != %d)",
			       (int)(cop->seq), (int)seq);
		return;
	}

	/* deal with accepts */

	if (status != OPT6_BNDA_ACCEPT)
		return;

	/* cancel timeouts */

	if (cop->tp != NULL) {
		kill_timeout(cop->tp);
		cop->tp = NULL;
	}
	if (cop->rettp)
		cancel_retrans(cop);

	/* reset the ack bit if not a home agent */

	if ((cop->flags & OPT6_BNDU_HR) == 0)
		cop->flags &= ~OPT6_BNDU_ACK;

	/* set the ack bit for multihoming */

	if (multi)
		cop->flags |= OPT6_BNDU_ACK;

	/* close tunnel when the ack is for a home de-registration */

	if ((cop->flags & OPT6_BNDU_HR) &&
	    IN6_ARE_ADDR_EQUAL(&cop->home, &cop->careof)) {
		mdel_tunnel(cop);
		return;
	}

	/* trigger a refresh? (TODO: more cases?) */

	if (((cop->type & CORR_USEBND) == 0) || (cop->lifetime == 0))
		return;

	if (((cop->flags & OPT6_BNDU_HR) == 0) && !multi)
		return;

	/* do it */

	if (Calloc(tp, timeout) == NULL) {
		logerr("out of memory");
		abort();
	}
	cop->tp = tp;
			
	set_timeout(tp, cop->lifetime / 2, 0,
		    (func_to) schedule_refresh, cop);
}

/* Refresh a binding */

void
schedule_refresh(cop)
	struct corr *cop;
{
	/* cancel any timeout */

	if (cop->tp) {
		kill_timeout(cop->tp);
		cop->tp = NULL;
	}
	if (cop->rettp)
		cancel_retrans(cop);

	/* no refresh? */

	if (((cop->type & CORR_USEBND) == 0) ||
	    IN6_ARE_ADDR_EQUAL(&cop->home, &cop->careof))
		return;

	if (verbose)
		logtrace("%s: refresh for %s",
			 inet6pr(&cop->addr),
			 inet6pr(&cop->careof));

	/* do it! */

	cop->seq++;

	if (cop->flags & OPT6_BNDU_ACK)
		cop->retcnt = 0;

	if (cop->flags & OPT6_BNDU_HR)
		send_home_reg(cop);
	else
		send_bnd_upd(cop);

	if (cop->flags & OPT6_BNDU_ACK)
		start_retrans(cop);
}
#endif
