/* Correspondent management for a mobile node */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"

static int prefix_match(struct corr *, struct in6_addr *);

/* Add a new correspondent */

void
add_corr(cop0, getsec)
	struct corr *cop0;
	int getsec;
{
	struct corr *cop = correspondents.tqh_first;

	/* Get security associations */

	if (getsec)
		mget_security(cop0);

	/* Protect special case */

	if (cop0->type & CORR_PROTECT)
		add_homeaddr(cop0);

	/* Insert */

	/* easy case: empty list */
	if (cop == NULL)
		goto at_tail;

	/* a bit less easy case: heads of list have longer prefixes */
	if (cop0->plen > cop->plen)
		goto at_head;

	/* first case: a home agent */
	if (cop0->flags & OPT6_BNDU_HR) {
		/* scan for first non-agent */
		FORALLCOR(cop)
			if ((cop->flags & OPT6_BNDU_HR) == 0)
				break;
		if (cop != NULL)
			goto before;
		goto at_tail;
	}

	/* scan for first prefix with the length */
	FORALLCOR(cop)
		if (cop0->plen >= cop->plen)
			break;
	if (cop == NULL)
		goto at_tail;

	/* disablers are before */
	if ((cop->type & CORR_SETHA) == 0)
		goto before;

	/* scan for a shorter prefix */
	FORALLCOR(cop)
		if (cop0->plen > cop->plen)
			break;
	if (cop == NULL)
		goto at_tail;
	/* fall in before */

	/* insertion cases */

    before:
	TAILQ_INSERT_BEFORE(cop, cop0, cr_list);
	return;

    at_head:
	TAILQ_INSERT_HEAD(&correspondents, cop0, cr_list);
	return;

    at_tail:
	TAILQ_INSERT_TAIL(&correspondents, cop0, cr_list);
	return;
}

/* Delete a correspondent */

void
del_corr(cop)
	struct corr *cop;
{
	/* cancel any timeout */

	if (cop->tp) {
		kill_timeout(cop->tp);
		cop->tp = NULL;
	}

	if (cop->rettp)
		cancel_retrans(cop);

	/* Delete security associations */

	mdel_security(cop);

	/* Delete home-address if needed */

	if ((cop->type & CORR_SETHA) &&
	    !IN6_ARE_ADDR_EQUAL(&cop->careof, &cop->home))
		del_homeaddr(cop);
	else if (cop->type & CORR_PROTECT)
		del_homeaddr(cop);

	/* Remove */

	TAILQ_REMOVE(&correspondents, cop, cr_list);

	/* Free it */

	free(cop);
}

/* clone a prefix for a candidate correspondent */

void
clone_corr(addr)
	struct in6_addr *addr;
{
	struct corr *cop, *pcp;
	struct in6_addr *newca;

	if (verbose)
		logtrace("clone: %s", inet6pr(addr));

	/* get the prefix */

	FORALLCOR(cop)
		if ((cop->type & CORR_PREFIX) == 0) {
			if (IN6_ARE_ADDR_EQUAL(addr, &cop->addr))
				break;
		} else if (prefix_match(cop, addr))
			break;

	/* got nothing? */

	if (cop == NULL) {
		if (verbose)
			logtrace("clone: no match");
		return;
	}
	pcp = cop;
	newca = &correspondents.tqh_first->careof;

	/* got an old entry? */

	if (cop->plen == 128) {
		if (verbose)
			logtrace("clone: get %s with type %x [%s]",
				 inet6pr(addr),
				 cop->type, corr_type(cop->type));
		if (cop->type & CORR_USEBND)
			goto rebind;
		return;
	}

	/* finished? */

	if (cop->type != (CORR_PREFIX | CORR_SETHA | CORR_CLONING))
		return;

	/* clone */

	if (Calloc(cop, corr) == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}
	cop->addr = *addr;
	COPY_ADDR6(pcp->home, cop->home);
	cop->plen = 128;
	/* CORR_SETHA is not mandatory because the prefix will match ??? */
	cop->type = CORR_CLONING | CORR_USEBND;
	cop->last = now.tv_sec;
	if (verbose)
		logtrace("clone: create %s with type %x [%s]",
			 inet6pr(addr), cop->type, corr_type(cop->type));
	add_corr(cop, 1);

	/* bind */

    rebind:
	COPY_ADDR6(pcp->home, cop->careof);
	if (IN6_ARE_ADDR_EQUAL(&cop->home, newca))
		return;
	mobile_move(cop, newca);

	/* TODO: GC on timeout */
}

/* Move */

void
mobile_move(cop, newca)
	struct corr *cop;
	struct in6_addr *newca;
{
	if (verbose)
		logtrace("%s: move from %s to %s",
			 inet6pr(&cop->addr),
			 inet6pr(&cop->careof),
			 inet6pr(newca));

	if (((cop->type & (CORR_SETHA | CORR_USEBND)) == 0) ||
	    IN6_ARE_ADDR_EQUAL(newca, &cop->careof))
		return;
	/* else */

	/* cancel any timeout */

	if (cop->tp) {
		kill_timeout(cop->tp);
		cop->tp = NULL;
	}

	if (cop->rettp)
		cancel_retrans(cop);

	/*
	 * Return to home:
	 */
	if (IN6_ARE_ADDR_EQUAL(newca, &cop->home)) {
		/* delete home-address */

		if (cop->type & CORR_SETHA)
			del_homeaddr(cop);

		/* set */

		COPY_ADDR6(*newca, cop->careof);
		cop->flags |= OPT6_BNDU_ACK;

		/* flush binding */

		if (cop->type & CORR_USEBND) {
			cop->seq++;
			cop->lifetime = 0;
			if (cop->flags & OPT6_BNDU_ACK)
				cop->retcnt = 0;
			if (cop->flags & OPT6_BNDU_HR)
				send_home_reg(cop);
			else
				send_bnd_upd(cop);
			if (cop->flags & OPT6_BNDU_ACK)
				start_retrans(cop);
		}

		return;
	} else

	/*
	 * Move from home:
	 */
	if (IN6_ARE_ADDR_EQUAL(&cop->careof, &cop->home)) {
		/* set care-of address */

		COPY_ADDR6(*newca, cop->careof);
		if ((cop->flags & OPT6_BNDU_HR) == 0)
			cop->flags &= ~OPT6_BNDU_ACK;
		if (multi)
			cop->flags |= OPT6_BNDU_ACK;

		/* add home-address */

		if (cop->type & CORR_SETHA)
			add_homeaddr(cop);

		/* update binding */

		if (cop->type & CORR_USEBND) {
			cop->seq++;
			cop->lifetime = bnd_lifetime;
			if (cop->flags & OPT6_BNDU_ACK)
				cop->retcnt = 0;
			if (cop->flags & OPT6_BNDU_HR)
				send_home_reg(cop);
			else
				send_bnd_upd(cop);
			if (cop->flags & OPT6_BNDU_ACK)
				start_retrans(cop);
		}

		return;
	} else

	/*
	 * Move from a place to another one:
	 */
	{
		/* delete home-address */

		if (cop->type & CORR_SETHA)
			del_homeaddr(cop);

		COPY_ADDR6(*newca, cop->careof);
		cop->flags |= OPT6_BNDU_ACK;

		/* add home-address */

		if (cop->type & CORR_SETHA)
			add_homeaddr(cop);

		/* update binding */

		if (cop->type & CORR_USEBND) {
			cop->seq++;
			cop->lifetime = bnd_lifetime;
			if (cop->flags & OPT6_BNDU_ACK)
				cop->retcnt = 0;
			if (cop->flags & OPT6_BNDU_HR)
				send_home_reg(cop);
			else
				send_bnd_upd(cop);
			if (cop->flags & OPT6_BNDU_ACK)
				start_retrans(cop);
		}

		return;
	}
}

/* Prefix matching utility */

static int
prefix_match(cop, addr)
	struct corr *cop;
	struct in6_addr *addr;
{
	register int i;

	for (i = 0; 8 * i <= cop->plen; i++)
		if (((cop->addr.s6_addr[i] ^ addr->s6_addr[i])
			& cop->mask.s6_addr[i]) != 0)
			return 0;
	return 1;
}
#endif
