/* Read mobile6 configuration file */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"
#include <sys/stat.h>
#include <ctype.h>

/*
 * Syntax
 *
 * mode [multi=(y|n)] [scope=(l|p|s|g)]
 * home [agent=xxx] [addr=xxx]
 * correspondent [addr=xxx | prefix=xxx::/dd] [mode=(n|a|o|y)]
 * removable[=list]
 * movable[=list]
 */

#define LEOF		(-1)
#define LNONE		0
#define LMODE		1
#define LCORR		2
#define LREM		3
#define LMOV		4
#define LHOME		5
#define KMUL		6
#define KPREF		7
#define KAGENT		8
#define KADDR		9
#define KFULL		10
#define KMODE		11
#define KSCOPE		12

static struct kwt {
	char *n;
	char l;
	char delim;
	u_char val;
#define KW(x)	x, sizeof(x)-1
} kwtb[] = {

	{ KW("mod"),	0,	LMODE },
	{ KW("corr"),	0,	LCORR },
	{ KW("remo"),	'=',	LREM },
	{ KW("mova"),	'=',	LMOV },
	{ KW("hom"),	0,	LHOME },
	{ KW("mult"),	'=',	KMUL },
	{ KW("dual"),	'=',	KMUL },
	{ KW("pref"),	'=',	KPREF },
	{ KW("agen"),	'=',	KAGENT },
	{ KW("addr"),	'=',	KADDR },
	{ KW("ful"),	'=',	KFULL },
	{ KW("mod"),	'=',	KMODE },
	{ KW("sco"),	'=',	KSCOPE },
	{ 0,		0,	0 }
};

static char *errsynt = "syntax error";

static char *read_modes(u_char *, int);
static char *read_correspondents(u_char *, int);
static char *read_removables(u_char *, int);
static char *read_movables(u_char *, int);
static char *read_homes(u_char *, int);
static char *read_homeagent(u_char *, int);
static char *read_homeaddr(u_char *, int);
static int scan(u_char *, u_char **, u_char **);

static struct stat oldstat;

void
read_mobile(frc)
	int frc;
{
	struct stat newstat;
	FILE *fp = NULL;
	u_char buf[1024];
	int i, num = 0;
	char *msgerr = NULL;

	/* stat the file */

	if (stat(MOBILE6, &newstat) < 0) {
		if (errno == ENOENT && oldstat.st_ino) {
			if (debug)
				logdebug("%s deleted", MOBILE6);
		}
    erropen:
		bzero(&newstat, sizeof newstat);
	}

	/* something has changed? */

	if (frc > 10 &&
	    oldstat.st_dev == newstat.st_dev &&
	    oldstat.st_ino == newstat.st_ino &&
	    oldstat.st_size == newstat.st_size &&
	    oldstat.st_mtime == newstat.st_mtime)
		return;

	/* open the file */

	if (newstat.st_ino && !(fp = fopen(MOBILE6, "r"))) {
		if (debug)
			logperror("open MOBILE6");
		goto erropen;
	}
	bcopy(&newstat, &oldstat, sizeof oldstat);
	if (frc > 10 && debug)
		logdebug("%s modified", MOBILE6);

	/* parse the file */

	if (fp)
	    while (fgets(buf, sizeof buf, fp)) {
		u_char *p;

		if ((p = index(buf, '#')))
			*p = 0;
		num++;
		i = scan(buf, 0, 0);
		if (i == LEOF)
			continue;
		switch (i) {
		case LMODE:
			if ((msgerr = read_modes(buf, frc)))
				goto err;
			break;
		case LCORR:
			if ((msgerr = read_correspondents(buf, frc)))
				goto err;
			break;
		case LREM:
			if ((msgerr = read_removables(buf, frc)))
				goto err;
			break;
		case LMOV:
			if ((msgerr = read_movables(buf, frc)))
				goto err;
			break;
		case LHOME:
			if ((msgerr = read_homes(buf, frc)))
				goto err;
			break;
		default:
			msgerr = errsynt;
		err:
			logerr("%s in %s line %d", msgerr, MOBILE6, num);
		}
	    }

	/* some scopes need a homeaddress */

	if ((frc == 0) &&
	    (mscope & SCP_PREFIX) &&
	    (PRIMHA == NULL)) {
		logerr("a home address is needed in this scope");
		mscope &= ~SCP_PREFIX;
		if (verbose)
			logtrace("reset the scope to %d", mscope);
	}

	/* post processing */

	if (msgerr == NULL) {
		if (frc != 0)
			read_final();
		else if (mscope & SCP_PREFIX)
			restrict_all();
	}
}

/* Parse mode definitions */

static char *
read_modes(p, frc)
	u_char *p;
	int frc;
{
	u_char *nxt, *arg;

	/* once only */

	if (frc != 0)
		return NULL;

	scan(p, &arg, &nxt);

	for (;;) {
		p = nxt;
		if (!*p)
			break;
		switch (scan(p, &arg, &nxt)) {
		case KMUL:
			switch (tolower(*arg)) {
			case 'n':
				multi = 0;
				break;
			case 'y':
				multi = 1;
				break;
			default:
				return errsynt;
			}
			break;
		case KSCOPE:
			if (mscope != SCP_LOCAL)
				return "multiple scopes";
			switch (tolower(*arg)) {
			case 'l':
				break;
			case 'p':
				mscope = SCP_PREFIX;
				break;
			case 's':
				mscope = SCP_SITE;
				break;
			case 'g':
				mscope = SCP_GLOBAL;
				break;
			default:
				return errsynt;
			}
			break;
		default:
			return errsynt;
		}
	}
	return NULL;
}

/* Parse correspondent definitions */

static char *
read_correspondents(p, frc)
	u_char *p;
	int frc;
{
	struct corr cor, *cop;
	u_char *q, *nxt, *arg;
	int i, n = -1, mode_is_set = 0;

	/* only when a default route is found */

	if (frc <= 0)
		return NULL;

	/* init */

	bzero(&cor, sizeof(cor));
	cor.type = CORR_SETHA;
	if (multi)
		cor.flags = OPT6_BNDU_ACK;
	scan(p, &arg, &nxt);

	/* parse arguments */

	for (;;) {
		p = nxt;
		if (!*p)
			break;
		q = NULL;
		switch (scan(p, &arg, &nxt)) {
		case KPREF:
			if (!IN6_IS_ADDR_UNSPECIFIED(&cor.addr) ||
			    (n != -1)) {
			errmult:
				return "multiple prefixes/addresses";
			}
			if (!(q = index(arg, '/')))
				goto errpref;
			*q++ = 0;
			n = atoi(q);
			if (n < 0 || n >= 128)
				goto errpref;
			if (inet_pton(AF_INET6, arg, &cor.addr) <= 0)
				goto errpref;
			if (IN6_IS_ADDR_MULTICAST(&cor.addr)) {
			errpref:
				return "invalid prefix";
			}
			cor.plen = n;
			cor.type |= CORR_PREFIX;
			/* clean up prefix */
			mk_mask(&cor.mask, n);
			for (i = 0; i < sizeof(cor.mask); i++)
				cor.addr.s6_addr[i] &= cor.mask.s6_addr[i];
			break;
		case KADDR:
			if (!IN6_IS_ADDR_UNSPECIFIED(&cor.addr) ||
			    (n != -1))
				goto errmult;
			if (inet_pton(AF_INET6, arg, &cor.addr) <= 0)
				goto err;
			if (IN6_IS_ADDR_UNSPECIFIED(&cor.addr) ||
			    IN6_IS_ADDR_MULTICAST(&cor.addr) ||
			    IN6_IS_ADDR_LINKLOCAL(&cor.addr) ||
			    IN6_IS_ADDR_SITELOCAL(&cor.addr) ||
			    IN6_IS_ADDR_V4MAPPED(&cor.addr)) {
			err:
				return "invalid address";
			}
			cor.plen = n = 128;
			break;
		case KMODE:
			if (mode_is_set) {
			errmode:
				return "mode is already set";
			}
			switch (tolower(*arg)) {
			case 'n':
				cor.type &= ~CORR_SETHA;
				break;
			case 'a':
				break;
			case 'o':
				cor.type |= CORR_CLONING;
				break;
			case 'y':
				cor.type |= CORR_USEBND;
				break;
			default:
				return errsynt;
			}
			mode_is_set++;
			break;
		case KFULL:
			if (mode_is_set)
				goto errmode;
			switch (tolower(*arg)) {
			case 'n':
				break;
			case 'y':
				cor.type |= CORR_USEBND;
				break;
			default:
				return errsynt;
			}
			mode_is_set++;
			break;
		default:
			return errsynt;
		}
	}
	/* illegal cases */
	if (cor.type & CORR_PREFIX) {
		if (cor.type & CORR_USEBND)
			return "a prefix can't be full";
		if ((cor.type & CORR_SETHA) &&
		    (IN6_IS_ADDR_LINKLOCAL(&cor.addr) ||
		     IN6_IS_ADDR_SITELOCAL(&cor.addr) ||
		     IN6_IS_ADDR_V4MAPPED(&cor.addr)))
			return "invalid prefix for this mode";
	}
	if (n != -1) {
		if (Calloc(cop, corr) == NULL) {
			if (verbose)
				logerr("out of memory");
			return "out of memory";
		}
		*cop = cor;
		COPY_ADDR6(def_if->addr, cop->careof);
		COPY_ADDR6(def_if->addr, cop->home);
		add_corr(cop, 0);
	}
	return NULL;
}

/* Parse removable interface list */

static char *
read_removables(p, frc)
	u_char *p;
	int frc;
{
	register struct dynif *dip;
	u_char *q, *nxt, *arg;
	char *ifn;

	/* once only */

	if (frc != 0)
		return NULL;

	scan(p, &arg, &nxt);

	/* parse list */

	for (ifn = arg; ifn; ifn = q) {
		if ((q = index(ifn, ',')))
			*q++ = 0;
		if (Calloc(dip, dynif) == NULL) {
			if (verbose)
				logerr("out of memory");
			return "out of memory";
		}
		strncpy(dip->name, ifn, IFNAMSIZ);
		TAILQ_INSERT_TAIL(&dynifs, dip, di_list);
	}
	return NULL;
}

/* Parse movable interface list */

static char *
read_movables(p, frc)
	u_char *p;
	int frc;
{
	register struct dynif *dip;
	u_char *q, *nxt, *arg;
	char *ifn;

	/* once only */

	if (frc != 0)
		return NULL;

	scan(p, &arg, &nxt);

	/* parse list */

	for (ifn = arg; ifn; ifn = q) {
		if ((q = index(ifn, ',')))
			*q++ = 0;
		if (Calloc(dip, dynif) == NULL) {
			if (verbose)
				logerr("out of memory");
			return "out of memory";
		}
		strncpy(dip->name, ifn, IFNAMSIZ);
		TAILQ_INSERT_TAIL(&movifs, dip, di_list);
	}
	return NULL;
}

/* Parse home definitions */

static char *
read_homes(p, frc)
	u_char *p;
	int frc;
{
	u_char *nxt, *arg;
	char *msgerr;

	/* incompatible with multi */

	if (multi)
		return "both home and multi: not yet supported";

	scan(p, &arg, &nxt);

	/* parse arguments */

	for (;;) {
		p = nxt;
		if (!*p)
			break;
		switch (scan(p, &arg, &nxt)) {
		case KAGENT:
			if ((msgerr = read_homeagent(arg, frc)))
				return msgerr;
			break;
		case KADDR:
			if ((msgerr = read_homeaddr(arg, frc)))
				return msgerr;
			break;
		default:
			return errsynt;
		}
	}
	return NULL;
}

/* Parse a home agent definition */

static char *
read_homeagent(arg, frc)
	u_char *arg;
	int frc;
{
	struct corr *cop;
	struct in6_addr addr;
	static int agent_count = 0;

	/* only when a default route is found */

	if (frc <= 0)
		return NULL;

	if (inet_pton(AF_INET6, arg, &addr) <= 0)
		goto err;

	if (IN6_IS_ADDR_UNSPECIFIED(&addr) ||
	    IN6_IS_ADDR_MULTICAST(&addr) ||
	    IN6_IS_ADDR_LINKLOCAL(&addr) ||
	    IN6_IS_ADDR_SITELOCAL(&addr) ||
	    IN6_IS_ADDR_V4MAPPED(&addr)) {
	err:
		return "invalid home agent address";
	}

	if (agent_count++)
		return "at most one home agent is supported";

	if (Calloc(cop, corr) == NULL) {
		if (verbose)
			logerr("out of memory");
		return "out of memory";
	}
	COPY_ADDR6(addr, cop->addr);
	COPY_ADDR6(def_if->addr, cop->careof);
	COPY_ADDR6(def_if->addr, cop->home);
	cop->type = CORR_SETHA | CORR_USEBND;
	cop->flags = OPT6_BNDU_HR | OPT6_BNDU_ACK;
	cop->plen = 128;
	cop->lifetime = reg_lifetime;
	add_corr(cop, 0);

	if (verbose)
		logtrace("add home agent %s", inet6pr(&cop->addr));

	return NULL;
}

/* Parse a home address definition */

static char *
read_homeaddr(arg, frc)
	u_char *arg;
	int frc;
{
	struct homeaddr *hap;
	struct interface *ifp;
	struct in6_addr addr;

	if (mscope & SCP_PREFIX) {
		/* as soon as boot */
		if (frc != 0)
			return NULL;
	} else {
		/* only when a default route is found */
		if (frc == 0)
			return NULL;
	}

	/* get address */

	if (inet_pton(AF_INET6, arg, &addr) <= 0)
		goto err;

	if (IN6_IS_ADDR_UNSPECIFIED(&addr) ||
	    IN6_IS_ADDR_MULTICAST(&addr) ||
	    IN6_IS_ADDR_LINKLOCAL(&addr) ||
	    IN6_IS_ADDR_SITELOCAL(&addr) ||
	    IN6_IS_ADDR_V4MAPPED(&addr)) {
	err:
		return "invalid home address";
	}

	if (PRIMHA) {
		/* already? */

		FORALLHA(hap)
			if (IN6_ARE_ADDR_EQUAL(&hap->addr, &addr)) {
				logerr("home address %s is already known",
				       inet6pr(&addr));
				return NULL;
			}
		logerr("at most one home address is supported");
	}

	/* alloc & init */

	if (Calloc(hap, homeaddr) == NULL) {
		if (verbose)
			logerr("out of memory");
		return "out of memory";
	}
	COPY_ADDR6(addr, hap->addr);
	TAILQ_INIT(&hap->homeprefs);
	TAILQ_INSERT_TAIL(&homeaddrs, hap, ha_list);

	/* where is the mobile ? */

	where_at_boot(hap);

	/* set up interfaces */

	FORALLIFP(ifp)
		add_home_prefix(hap, ifp);

	return NULL;
}

static int
scan(s, arg, nxt)
	u_char *s;
	u_char **arg, **nxt;
{
	u_char *p, d = 0;
	struct kwt *k;

	while (isspace(*s))
		s++;
	if (!*s)
		return LEOF;
	p = s;
	while (isalnum(*p))
		p++;
	if (*p && !isspace(*p))
		d = *p++;
	if (arg)
		*arg = p;
	while (*p && !isspace(*p))
		p++;
	if (nxt) {
		if (arg && *p)
			*p++ = 0;
		while (isspace(*p))
			p++;
		*nxt = p;
	}
	for (k = kwtb; k->n; k++)
		if (k->delim == d && strncasecmp(k->n, s, k->l) == 0)
			break;
	if (k->n)
		return k->val;
	else
		return LNONE;
}
#endif
