/* Deal with the "mobile Ethernet" case */

#include "ndpd.h"
#include "mobility.h"

#ifdef MOBILE
static void mld_restart(struct interface *);

/*
 * Restart Multicast Listener Discovery
 * (if we have moved MLD state is still for the old instance)
 */

void
mld_restart(ifp)
	struct interface *ifp;
{
#ifdef IPV6_RESTART_MLD
	struct ipv6_mreq mreq;

	bzero(&mreq, sizeof(mreq));
	mreq.ipv6mr_interface = ifp->index;
	(void) setsockopt(u, IPPROTO_IPV6, IPV6_RESTART_MLD,
			  &mreq, sizeof(mreq));
#endif
}
#endif

/*
 * A Router Advertisement with totally new things have been received
 * (called after update_prefix()/sort_prefix() but before update_router())
 */

void
mobile_newrtadv()
{
#ifdef MOBILE
	struct interface *ifp;
	struct prefix *pp;
	struct router *rp, *rp1 = NULL;

	/* current router has sent this RtAdv */

	if (cur_rt == NULL) {
		if (verbose)
			logerr("mobile_newrtadv: no current router");
		return;
	}
	ifp = cur_rt->ifp;
	if (BEST(ifp) == NULL) {
		if (verbose)
			logerr("mobile_newrtadv: no router on %s", ifp->name);
		return;
	}

	/* movable interface? */

	if (!is_mov(ifp))
		return;

	/* we have a "mobile Ethernet" */

	if (verbose)
		logtrace("mobile_newrtadv: interface %s can have moved!",
			 ifp->name);

	/* try to know more */

	if_restart(ifp);

	/* restart MLD (a bit aggressive?) */

	mld_restart(ifp);

	/* make all old real prefix deprecated/invalid in 10/15 seconds */

	FORALLPP(pp, ifp)
		if (!SIMULT(pp->last, last_rtadv) &&
		    ((pp->mflags & MOB_ISFA) == 0)) {
			struct timeval delta;
			long age;

			/* compute the age */
			delta.tv_sec = last_rtadv.tv_sec - pp->last.tv_sec - 1;
			delta.tv_usec = last_rtadv.tv_usec - pp->last.tv_usec;
			delta.tv_usec += 1000000;
			NORMTIME(delta);
			age = delta.tv_sec;
			if (delta.tv_usec >= 500000)
				age += 1;

			/* ignore short term */
			if ((pp->deplife != INFLIFE) &&
			    (age + 10 >= pp->deplife))
				continue;
			if ((pp->invlife != INFLIFE) &&
			    (age + 15 >= pp->invlife))
				continue;

			/* reset timeout with short values */
			pp->deplife = 10;
			pp->invlife = 15;

			if (pp->tp) {
				kill_timeout(pp->tp);
				pp->tp = NULL;
			}

			pp->tp = new_timeout(pp->deplife, 0,
					     (func_to) prefix_deprec, pp);
			if (verbose)
				logtrace("mobile_newrtadv: make %s/%d old",
					 inet6pr(&pp->addr), pp->plen);
		}

	/* aggressive router cleanup is possible? */

	rp = BEST(ifp);
	if ((rp == cur_rt) && (rp->rt_list.tqe_next == NULL)) {
		if (verbose)
			logtrace("mobile_newrtadv: virgin interface %s "
				 "(no old router)", ifp->name);
		return;
	}
	if ((cur_rt->lifetime == 0) || (cur_rt->phy.sdl_alen == 0))
		goto smooth;
	if (ifp->hasphy && (rp->phy.sdl_alen == 0))
		goto smooth;

	/* cleanup old routers */

	if (verbose)
		logtrace("mobile_newrtadv: purge routers on %s (hard)",
			 ifp->name);

	/* first phase: invalid unreachables */

    again1:
	FORALLRP(rp, ifp)
		if ((rp != cur_rt) && (rp->phy.sdl_alen == 0))
			rp1 = rp;
	if (rp1 != NULL) {
		router_invalid(rp1);
		rp1 = NULL;
		goto again1;
	}

	/* second phase: make not current unreachable */

    again2:
	FORALLRP(rp, ifp)
		if ((rp != cur_rt) && (rp->phy.sdl_alen != 0))
			rp1 = rp;
	if (rp1 != NULL) {
		router_unreachable(rp1, 0);
		del_router(rp1, 0);
		rp1 = NULL;
		goto again2;
	}

	/* last phase: make not current old */

    smooth:
	if (verbose)
		logtrace("mobile_newrtadv: purge routers on %s (smooth)",
			 ifp->name);

	FORALLRP(rp, ifp)
		if (rp != cur_rt) {
			struct timeval delta;
			long age;

			/* compute the age */
			delta.tv_sec = last_rtadv.tv_sec - rp->last.tv_sec - 1;
			delta.tv_usec = last_rtadv.tv_usec - rp->last.tv_usec;
			delta.tv_usec += 1000000;
			NORMTIME(delta);
			age = delta.tv_sec;
			if (delta.tv_usec >= 500000)
				age += 1;

			/* ignore short term */
			if (age + 20 >= rp->lifetime)
				continue;

			/* reset timeout with a short value */
			rp->lifetime = 20;

			if (rp->tp) {
				kill_timeout(rp->tp);
				rp->tp = NULL;
			}

			rp->tp = new_timeout(rp->lifetime, 0,
					     (func_to) router_invalid, rp);
			if (verbose)
				logtrace("mobile_newrtadv: make %s old",
					 inet6pr(&rp->addr.sin6_addr));
		}
#endif
}
