/* Manage tunnel end at the mobile */

#include "ndpd.h"
#include "mobility.h"

#ifdef MOBILE

#include <net/bpf.h>
#include <netinet/if_tug6.h>

#define CAPLEN		72
#define TUGCSIZE	8

static char bbuf[BUFSIZ];
static struct bpf_insn bi[] = { BPF_STMT(BPF_RET+BPF_K, CAPLEN) };
static struct bpf_program bp = { 1, bi };

/* Add a tunnel */

void
madd_tunnel(agp)
	struct corr *agp;
{
	struct if_laddrreq ifr6;
	struct ifreq ifr;
	struct sockaddr_in6 *sin;
	char dname[12];
	int d, unit = 0;
	u_int val;

	if (verbose)
		logtrace("will set up tunnel end on %s from %s",
			 "tug0", inet6pr(&agp->addr));

	/* UDP/IPv6 socket for interface ioctls */
	if ((d = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		logperror("socket (home-agent/add)");
		return;
	}

	/* just set the end point of tunnel, making it RUNNING... */

	bzero(&ifr6, sizeof(ifr6));
	strncpy(ifr6.iflr_name, "tug0", IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&ifr6.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(agp->addr, sin->sin6_addr);

	if (ioctl(d, SIOCSTUGLADDR, &ifr6) < 0) {
		logperror("ioctl (SIOCSTUGLADDR)");
		(void) close(d);
		return;
	}

	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, "tug0", IFNAMSIZ);
	
	if (ioctl(d, SIOCGIFFLAGS, &ifr) < 0) {
		logperror("ioctl (SIOCGIFFLAGS)");
		(void) close(d);
		return;
	}

	ifr.ifr_flags |= IFF_UP;

	if (ioctl(d, SIOCSIFFLAGS, &ifr) < 0) {
		logperror("ioctl (SIOCSIFFLAGS)");
		(void) close(d);
		return;
	}

	/* BPF stuff */

	if (b >= 0) {
		if (verbose)
			logtrace("reset BPF on %s", "tug0");
		bzero(&ifr, sizeof(ifr));
		strncpy(ifr.ifr_name, "tug0", IFNAMSIZ);
		ifr.ifr_csize = TUGCSIZE;
		if (ioctl(d, SIOCSTUGCSIZE, &ifr) < 0)
			logperror("ioctl (SIOCSTUGCSIZE)");
		(void) ioctl(b, BIOCFLUSH, NULL);
		(void) close(d);
		return;
	}
    retry:
	(void) snprintf(dname, sizeof(dname) - 1, "/dev/bpf%d", unit++);
	if ((b = open(dname, O_RDONLY)) < 0) {
		if (errno == EBUSY)
			goto retry;
		logperror("open (BPF)");
		(void) close(d);
		return;
	}
	if (verbose)
		logtrace("open BPF on %s", dname);
	val = BUFSIZ;
	if (ioctl(b, BIOCSBLEN, &val) < 0)
		logperror("ioctl (BIOCSBLEN)");
	val = 1;
	if (ioctl(b, BIOCIMMEDIATE, &val) < 0)
		logperror("ioctl (BIOCIMMEDIATE)");
	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, "tug0", IFNAMSIZ);
	if ((ioctl(b, BIOCSETIF, &ifr) < 0) && verbose)
		logperror("ioctl (BIOCSETIF)");
	if (ioctl(b, BIOCSETF, &bp) < 0)
		logperror("ioctl (BIOCSETF)");
	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, "tug0", IFNAMSIZ);
	ifr.ifr_csize = TUGCSIZE;
	if (ioctl(d, SIOCSTUGCSIZE, &ifr) < 0)
		logperror("ioctl (SIOCSTUGCSIZE)");
	(void) close(d);
}

/* Delete a tunnel */

void
mdel_tunnel(agp)
	struct corr *agp;
{
	/* TODO ??? down and unconfig */

	if (verbose)
		logtrace("will release tunnel end on %s from %s",
			 "tug0", inet6pr(&agp->addr));

	/* close BPF */

	if (b < 0)
		return;
	(void) ioctl(b, BIOCFLUSH, NULL);
	(void) close(b);
	b = -1;
}

/* Read a packet from BPF */

void
recv_bpf()
{
	struct bpf_hdr *hp;
	struct ip6_hdr *ip;
	struct homeaddr *hap;
	u_int *afp;
	char *p;
	int cc;

	cc = read(b, bbuf, sizeof(bbuf));
	if (cc < 0) {
		logperror("read (BPF)");
		return;
	}
	if (cc < sizeof(*hp)) {
		if (verbose)
			logerr("recv_bpf: very short (%d) read", cc);
		return;
	}

	p = bbuf;

	for (;;) {
		hp = (struct bpf_hdr *)p;
		afp = (u_int *)(p + hp->bh_hdrlen);
		ip = (struct ip6_hdr *)(afp + 1);

		if (hp->bh_caplen < sizeof(*ip) + sizeof(*afp))
			goto next;
		if ((char *)(ip + 1) > bbuf + cc)
			return;
		if (*afp != AF_INET6)
			return;
		if ((ip->ip6_flow & IPV6_FLOWINFO_VERSION) != IPV6_VERSION)
			return;
		FORALLHA(hap)
			if (IN6_ARE_ADDR_EQUAL(&ip->ip6_dst, &hap->addr))
				break;
		if (hap == NULL)
			return;

		if (verbose)
			logtrace("recv_bpf: got %s", inet6pr(&ip->ip6_src));
		clone_corr(&ip->ip6_src);

	    next:
		p += BPF_WORDALIGN(hp->bh_hdrlen + hp->bh_caplen);
		if (p >= bbuf + cc)
			return;
	}
}

#else

void
recv_bpf()
{
	logerr("recv_bpf ???");
	abort();
}

#endif
