/* Include for IPv6 mobility */

#include <netinet/in_pcb.h>
#include <netinet/ip6_policy.h>
#include <netinet/ip6_mob.h>

#define BND_LIFETIME	(20 * 60)	/* 20mn for default binding lifetime */
#define REG_LIFETIME	(60 * 60)	/* 1h for default home reg lifetime */

struct mobile {
	TAILQ_ENTRY(mobile) mb_list;	/* mobile list */
	struct in6_addr home;		/* home address */
	struct in6_addr careof;		/* care-of address */
	struct in6_addr me;		/* my address */
	struct timeout *tp;		/* timeout */
	u_long refresh;			/* last refresh time */
	u_int lifetime;			/* life time */
	u_short seq;			/* sequence number */
	u_char flags;			/* flags (from update) */
	u_char plen;			/* prefix length */
	caddr_t sap;			/* pointer to SA (TODO) */
#ifdef HOME_AGENT
	struct prefix *pp;		/* associated prefix */
	struct anycast *ap;		/* proxy entry */
	struct tunnel *tup;		/* tunnel to client */
#endif
};
TAILQ_HEAD(, mobile) mobiles;
#define FORALLMOB(mop)	\
	for(mop = mobiles.tqh_first; mop; mop = mop->mb_list.tqe_next)

extern	char *ipsec_policy;		/* IPsec policy (binary) */
extern	int ipsec_len;			/* IPsec policy binary length */
extern	int is_secure;			/* IPsec usage flag */

#ifdef HOME_AGENT

struct tunnel {
	TAILQ_ENTRY(tunnel) tn_list;	/* tunnel list (free first) */
	char name[IFNAMSIZ];		/* name */
	int state;			/* state */
#define TUN_FREE	0
#define TUN_USED	1
#define TUN_RES		2
};
TAILQ_HEAD(, tunnel) tunnels;
#define FORALLTUN(tup)	\
	for(tup = tunnels.tqh_first; tup; tup = tup->tn_list.tqe_next)

#endif

#ifdef MOBILE

struct corr {
	TAILQ_ENTRY(corr) cr_list;	/* correspondent list */
	struct in6_addr addr;		/* correspondent address */
	struct in6_addr mask;		/* prefix mask */
	struct in6_addr careof;		/* care-of address */
	struct in6_addr home;		/* home address */
	struct timeout *tp;		/* timeout */
	u_int type;			/* type */
#define CORR_PREFIX	0x01
#define CORR_SETHA	0x02
#define CORR_CLONING	0x04
#define CORR_USEBND	0x08
#define CORR_PROTECT	0x10
	u_char flags;			/* flags to use in updates */
	u_char plen;			/* prefix length */
	u_short seq;			/* sequence number */
	u_long last;			/* last seen time */
	u_int lifetime;			/* life time */
	u_int retcnt;			/* retrans counter */
	struct timeout *rettp;		/* retrans timeout */
	struct csa *sap;		/* pointer to SA */
};
TAILQ_HEAD(, corr) correspondents;
#define FORALLCOR(cop)	\
	for(cop = correspondents.tqh_first; cop; cop = cop->cr_list.tqe_next)

struct csa {
	struct timeout *tp;		/* repoll timeout */
	u_char acquired;		/* acquiring or acquired */
};

extern	u_int bnd_lifetime;		/* default binding lifetime */
extern	u_int reg_lifetime;		/* default registration lifetime */
extern	int multi;			/* multi-homed mode */
extern	int mscope;			/* strictness scope */

#define SCP_LOCAL	0x0		/* enforce nothing */
#define SCP_PREFIX	0x1		/* visited => off-link */
#define SCP_SITE	0x2		/* (to) do site stuff */
#define SCP_GLOBAL	0x3		/* enforce everything */

#define MOBILE6		"/etc/mobile6"

#endif

#define SEQ_GT(a, b)	((short)((a) - (b)) > 0)

#define GET8(val, cp)	((val) = *(cp)++)

#define GET16(val, cp)	\
	{ \
		(val) = *(cp)++ << 8; \
		(val) |= *(cp)++; \
	}

#define GET32(val, cp)	\
	{ \
		(val) = *(cp)++ << 24; \
		(val) |= *(cp)++ << 16; \
		(val) |= *(cp)++ << 8; \
		(val) |= *(cp)++; \
	}

#define GETADDR(val, cp)	\
	{ \
		bcopy((cp), &(val), 16); \
		(cp) += 16; \
	}

#define SET8(val, cp)	(*(cp)++ = val)

#define SET16(val, cp)	\
	{ \
		*(cp)++ = (val) >> 8; \
		*(cp)++ = (val); \
	}

#define SET32(val, cp)	\
	{ \
		*(cp)++ = (val) >> 24; \
		*(cp)++ = (val) >> 16; \
		*(cp)++ = (val) >> 8; \
		*(cp)++ = (val); \
	}

#define SETADDR(val, cp)	\
	{ \
		bcopy(&(val), (cp), 16); \
		(cp) += 16; \
	}

void add_home_agent(struct in6_addr *, int);

void add_binding(struct mobile *);
void del_binding(struct mobile *);
void change_binding(struct mobile *);

void del_route(struct in6_addr *);

int kvm_mobiles(void);
int kvm_get_rt(struct mobile *);

int check_ipsec(void);
int get_security(struct mobile *);
void del_security(struct mobile *);

void recv_bnd_upd(struct ip6_hdr *, u_char *, struct in6_addr *);
void recv_bnd_ack(struct ip6_hdr *, u_char *, int);
void recv_bnd_req(struct ip6_hdr *);
void clean_bindings(struct mobile *);

void recv_home_reg(struct ip6_hdr *, u_char *, struct in6_addr *);

void send_bnd_ack(struct mobile *, u_char, u_int);
void send_bnd_req(struct mobile *);

char *bnd_upd_flg(u_char);

#ifdef MOBILE

void read_mobile(int);

void where_at_boot(struct homeaddr *);
void add_home_prefix(struct homeaddr *, struct interface *);
void read_final(void);

extern int discarding;
void homeroute(struct interface *, int);

void add_corr(struct corr *, int);
void del_corr(struct corr *);
void clone_corr(struct in6_addr *);
void mobile_move(struct corr *, struct in6_addr *);

void add_homeaddr(struct corr *);
void del_homeaddr(struct corr *);

int kvm_get_ha(struct corr *);

void schedule_refresh(struct corr *);

void mget_security(struct corr *);
void mdel_security(struct corr *);

void start_retrans(struct corr *);
void cancel_retrans(struct corr *);

void multi_move(struct interface *, struct prefix *);
void minor_move(struct interface *, struct prefix *);
void major_move(struct prefix *, int);

void protect_all(void);
void add_protect(struct prefix *);
void del_protect(struct prefix *);

void restrict_all(void);
void restrict(struct prefix *, int);

void send_bnd_upd(struct corr *);
void send_home_reg(struct corr *);
void madd_tunnel(struct corr *);
void mdel_tunnel(struct corr *);

char *corr_type(u_int);

#endif

#ifdef HOME_AGENT

void recv_ha_disc(struct ip6_hdr *, u_char *,
		  struct in6_addr *, struct anycast *);
void send_ha_disc(struct mobile *, struct prefix *);
void del_hagent(struct hagent *);

struct prefix *loc_client(struct mobile *);

void del_client(struct mobile *);

void init_tunnels(void);
int add_tunnel(struct mobile *);
void del_tunnel(struct mobile *);

#endif
