#include "ndpd.h"
#include <sys/stat.h>

static char *errsynt = "syntax error";
static char *read_filter(u_char *);
static char *read_opt(u_char *, int *);
static char *read_prefix(u_char *);
static char *read_route(u_char *);
static int scan(u_char *, u_char **, u_char **);
static void setopt(struct interface *);

/* gateway syntax
option [if=n1,n2] ripin=(y|n) ripout=(y|n|S|R) rtadv=(y|n|min[/max])
	flag=[MOH] life=dd reach=dd retrans=dd mtu[=mtuval]
	prefix=(y|n) deprec=dd valid=dd
prefix if=n pref=(none|xxx::/dd|import[=xx/d]) flag=[LAR]
	site=dd valid=dd deprec=dd
filter=(in|out) [if=n1,n2] prefix=xx::/dd gateway=xxx length=(=|>=|<=|<|>)dd
	action=(accept|reject|truncate/xx)
xxx::/%d metric %d [gateway xxxx ifname]
*/
#define LEOF		(-1)
#define LNONE		0
#define LOPT		1
#define LPREF		2
#define LFILT		3
#define KACT		4
#define KFLAG		5
#define KGW		6
#define KIF		7
#define KLEN		8
#define KLIFE		9
#define KPREF		10
#define KREACH		11
#define KRETRANS	12
#define KRIN		13
#define KROUT		14
#define KRTADV		15
#define KVAL		16
#define KDEPREC		17
#define KMTU		18
#define KMTUIF		19
#define KHAADV		20
#define KSITE		21

static struct kwt {
	char *n;
	char l;
	char delim;
	u_char val;
#define KW(x)	x, sizeof(x)-1
} kwtb[] = {
	{ KW("opt"),	0,	LOPT },
	{ KW("pref"),	0,	LPREF },
	{ KW("filt"),	'=',	LFILT },
	{ KW("act"),	'=',	KACT },
	{ KW("dep"),	'=',	KDEPREC },
	{ KW("flag"),	'=',	KFLAG },
	{ KW("gate"),	'=',	KGW },
	{ KW("gw"),	'=',	KGW },
	{ KW("if"),	'=',	KIF },
	{ KW("interf"),	'=',	KIF },
	{ KW("len"),	'=',	KLEN },
	{ KW("lif"),	'=',	KLIFE },
	{ KW("mtu"),	0,	KMTUIF },
	{ KW("mtu"),	'=',	KMTU },
	{ KW("pref"),	'=',	KPREF },
	{ KW("reach"),	'=',	KREACH },
	{ KW("retr"),	'=',	KRETRANS },
	{ KW("ripin"),	'=',	KRIN },
	{ KW("ripout"),	'=',	KROUT },
	{ KW("rtadv"),	'=',	KRTADV },
	{ KW("valid"),	'=',	KVAL },
	{ KW("haadv"),	'=',	KHAADV },
	{ KW("site"),	'=',	KSITE },
	{ 0,		0,	0 }
};

static struct stat oldstat;
static int started;

void
read_gateway(frc)
	int frc;
{
	struct stat newstat;
	FILE *fp = NULL;
	struct Route *rp;
	struct prefix *pp;
	struct interface *ifp;
	u_char buf[1024];
	int num, i;
	int mode; /* 0: start, 1: after opt_def, 2: after opt; 3: route */
	char *msgerr;

	if (stat(gateway6, &newstat) < 0) {
		if (errno == ENOENT && oldstat.st_ino) {
			if (debug)
				logdebug("%s deleted", gateway6);
		}
    erropen:
		bzero(&newstat, sizeof newstat);
	}
        if (!frc &&
	    oldstat.st_dev == newstat.st_dev &&
            oldstat.st_ino == newstat.st_ino &&
            oldstat.st_size == newstat.st_size &&
            oldstat.st_mtime == newstat.st_mtime)
		return;
	if (newstat.st_ino && !(fp = fopen(gateway6, "r"))) {
		if (debug)
			logperror("open GATEWAY6");
		goto erropen;
	}
	bcopy(&newstat, &oldstat, sizeof oldstat);
	if (!frc && debug)
		logdebug("%s modified", gateway6);
	mode = num = 0;
	started = frc;
	defopt = defoptini;
	FORALLIFP(ifp) {
		if (ifp->filtin)
			free(ifp->filtin);
		if (ifp->filtout)
			free(ifp->filtout);
		ifp->filtin = ifp->filtout = 0;
		while ((pp = FFIRST(ifp)))
			TAILQ_REMOVE(&ifp->fprefixes, pp, prf_list);
		ifp->usefpref = 0;
		ifp->oldopt = ifp->opt.difopt;
		ifp->opt = defopt;
	}
	if (filtin)
		free(filtin);
	if (filtout)
		free(filtout);
	filtin = filtout = 0;

	if (fp)
	    while (fgets(buf, sizeof buf, fp)) {
		u_char *p;
		if ((p = index(buf, '#')))
			*p = 0;
		num++;
		i = scan(buf, 0, 0);
		if (i == LEOF)
			continue;
		if (i == LNONE) {
			if ((msgerr = read_route(buf)))
				goto err;
			mode=3;
			continue;
		}
		if (mode > 2) {
			msgerr = "option/pref/filter after gateway";
			goto err;
		}
		switch (i) {
		case LOPT:
			if (mode>1) {
				msgerr = "option too late";
				goto err;
			}
			if ((msgerr = read_opt(buf, &mode)))
				goto err;
			break;
		case LPREF:
			mode = 2;
			if ((msgerr = read_prefix(buf)))
				goto err;
			break;
		case LFILT:
			mode = 2;
			if ((msgerr = read_filter(buf)))
				goto err;
			break;
		default:
			msgerr = errsynt;
		err:
			logerr("%s in %s line %d", msgerr, gateway6, num);
		}
	}
	FORALLIFP(ifp) {
		FORALLPP(pp, ifp)
			pp->valpref = ifp->opt.valpref;
		if (ifp->usefpref == 0 && ifp->opt.nopref)
			ifp->usefpref = -1;
		if (ifp->usefpref < 0)
			while ((pp = FFIRST(ifp)))
				TAILQ_REMOVE(&ifp->fprefixes, pp, prf_list);
		setopt(ifp);
	}
	if (optg)			/* may be filter changes */
		decl_route_default(RT_SEEN);
	FORALLIFP(ifp)
	    if (IS_UP(ifp))
		FORALLPP(pp, ifp) {
			add_route_prefix(pp, ifp, RT_SEEN);
#ifdef notyet
			add_anycast_prefix(ifp, pp, NULL);
#endif
		}
	(void)gettimeofday(&now, NULL);
	FORALLRP(rp) {
		if (rp->flags & RT_MANUAL) {
			if (rp->flags & RT_SEEN)
				rp->flags &= ~RT_SEEN;
			else if (rp->metric != HOPCNT_INFINITY) {
				del_kern_route(rp);
				rp->refresh = now.tv_sec;
				rp->metric = HOPCNT_INFINITY;
				ripchanges = rp->flags |= RT_CHANGED;
			}
			continue;
		}
	/* don't bother to check the received routes, they will expire */
	}
}

static char *
read_filter(p)
	u_char *p;
{
	struct Filter f;
	struct TFilter **pfilt = 0;
	int action = 0, sens = 0, isgw = 0;
	int n = 0;
	char *ifn = 0;
	u_char *q = NULL, *nxt, *arg;
	struct interface *ifp = NULL;

	bzero(&f, sizeof f);
	f.gateway.s6_addr[0] = NO_GATEWAY;
	scan(p, &arg, &nxt);
	sens = tolower(*arg);
	if (sens != 'i' && sens != 'o')
		return errsynt;

	for (;;) {
		p = nxt;
		if (!*p)
			break;
		switch (scan(p, &arg, &nxt)) {
		    default:
			return errsynt;
		    case KIF:
			ifn = arg;
			break;
		    case KPREF:
			if (!(q = index(arg, '/')))
				goto errpref;
			*q++ = 0;
			n = atoi(q);
			if (inet_pton(AF_INET6, arg, &f.addr) <= 0 ||
			    (n < 0 || n > 128)) {
			errpref:
				return "invalid prefix";
			}
			f.plen = n;
			/* clean up prefix */
			mk_mask(&f.mask, n);
			for (n = 0; n < sizeof(f.mask); n++)
				f.addr.s6_addr[n] &= f.mask.s6_addr[n];
			break;
		    case KGW:
			isgw = 1;
			if (inet_pton(AF_INET6, arg, &f.gateway) <= 0)
				return "invalid gateway address";
			break;
		    case KLEN:
			n = 0;
			switch (*arg++) {
			    case '=':
				f.typlg = LG_EQ;
				break;
			    case '>':
				f.typlg = LG_GE;
				if (*arg == '=')
					arg++;
				else
					n = 1;
				break;
			    case '<':
				f.typlg = LG_LE;
				if (*arg == '=')
					arg++;
				else
					n = -1;
				break;
			    default:
			    errlen:
				return "invalid len field";
			}
			n += atoi(arg) ;
			if (n < 0 || n > 128)
				goto errlen;
			f.lg = n;
			break;
		    case KACT:
			switch (action = tolower(*arg++)) {
			    case 'a':
				f.action = FACCEPT;
				break;
			    case 'r':
				f.action = FREJECT;
				break;
			    case 't':
				while (isalpha(*arg))
					arg++;
				if (*arg == '/')
					arg++;
				if (!isdigit(*arg))
					goto erract;
				n = atoi(arg);
				if (n < 0 || n > 128)
					goto erract;
				f.action = n;
				break;
			    default:
			    erract:
				return "invalid act field";
			}
		}
	}
	if (!action || !sens || (sens=='o' && isgw))
		return "missing/conflicting fields";
	if (!ifn) {
		pfilt = (sens =='i') ? &filtin : &filtout;
		goto set;
	} else for (; ifn; ifn = q) {
		if ((q = index(ifn, ',')))
			*q++ = 0;
		if (!(n = strlen(ifn)))
			continue;
		n = (ifn[n-1] == '*') ? n-1 : IFNAMSIZ+1;
		FORALLIFP(ifp) {
			if (strncmp(ifp->name, ifn, n)!=0)
				continue;
			if (sens =='i')
				pfilt = &ifp->filtin;
			else
				pfilt = &ifp->filtout;
    set:
			if (!*pfilt) {
				if (Calloc(*pfilt, TFilter) == NULL) {
    nomem:
					if (verbose)
						logerr("out of memory");
					return "out of memory";
				}
			} else if ((*pfilt)->nb%FILTALLOC == 0) {
				struct TFilter *p = (struct TFilter *)
					realloc(*pfilt, sizeof **pfilt +
					    (*pfilt)->nb*sizeof(struct Filter));
				if (!p)
					goto nomem;
				*pfilt = p;
			}
			(*pfilt)->filter[(*pfilt)->nb] = f;
			(*pfilt)->nb++;
			if (!ifn)
				return (char *)0;
		}
	}
	return (char *)0;
}

static char *
read_prefix(p)
	u_char *p;
{
	struct prefix pf, setpf, *pp, *pp2;
	u_int n;
	u_char *q, *nxt, *arg;
	struct interface *ifp;
	char *ifn = 0;
	struct valpref prvalpref;
	int opt = 0, mode = 0; /* 'n' 'i' 'm' 'p' */
	int notaprefix = 0;

	bzero(&pf, sizeof pf);
	pf.valpref.pref_date = pf.valpref.val_date = -1; /* not set */
	pf.flags = ND_OPT_PI_FLAG_AUTO | ND_OPT_PI_FLAG_ONLINK;

	scan(p, &arg, &nxt);

	for (;;) {
		p = nxt;
		if (!*p)
			break;
		switch (scan(p, &arg, &nxt)) {
		    default:
			return errsynt;
		    case KIF:
			ifn = arg;
			break;
		    case KPREF:
			if (mode)
				return errsynt;
			mode = tolower(*arg);
			if (mode == 'n')
				break;
			if (mode == 'i') {
				while (isalpha(*arg))
					arg++;
				if (*arg == 0)
					break;
				if (*arg++ != '=')
					return errsynt;
				mode = 'm';
			} else
				mode = 'p';
			if (!(q = index(arg, '/')))
				goto errpref;
			*q++ = 0;
			n = atoi(q);
			if (inet_pton(AF_INET6, arg, &pf.addr) <= 0 ||
			    (n < 0 || n > 128)) {
			errpref:
				return "invalid prefix";
			}
			/* Hack alert: to forbid eui64 insertion, use an
			 * address ending by a digit, even with 0:
			 *  x:: -> insert eui, x::0 -> do not insert ! */
			if (q[-2] != ':')
				notaprefix = 1;
			if (mode == 'p' &&
			    (IN6_IS_ADDR_UNSPECIFIED(&pf.addr) ||
			     IN6_IS_ADDR_MULTICAST(&pf.addr) ||
			     IN6_IS_ADDR_LINKLOCAL(&pf.addr) ||
			     (pf.addr.s6_addr[0] == 0)))
				goto errpref;
			/* clean up prefix */
			pf.plen = n;
			COPY_ADDR6(pf.addr, pf.pref);
			mk_pref(&pf);
			break;
		    case KFLAG:
			pf.flags &= ~ND_OPT_PI_FLAG_CFG;
			opt = 1;
			while ((n = (u_char)*arg++)) switch (tolower(n)) {
			    case 'l':
				pf.flags |= ND_OPT_PI_FLAG_ONLINK;
				break;
			    case 'a':
				pf.flags |= ND_OPT_PI_FLAG_AUTO;
				break;
			    case 'r':
				pf.flags |= ND_OPT_PI_FLAG_RTADDR;
				break;
			    default:
				return "invalid flag field";
			}
			break;
		    case KSITE:
			pf.flags |= ND_OPT_PI_FLAG_SITE;
			if (!isdigit(*arg))
				goto errpref;
			n = atoi(arg);
			if (n < 0 || n > 128)
				goto errpref;
			pf.slen = n;
			opt = 1;
			break;
		    case KVAL:
			if (convvalpref(arg,
					&pf.valpref.val_date,
					&pf.valpref.val) < 0)
				goto errpref;
			opt = 1;
			break;
		    case KDEPREC:
			if (convvalpref(arg,
					&pf.valpref.pref_date,
					&pf.valpref.pref) < 0)
				goto errpref;
			opt = 1;
			break;
		}
	}

	if (ifn == 0)
		return "missing interface";
	if (mode == 'n' && opt)
		return "options with prefix=none";

	prvalpref = pf.valpref;
	for (; ifn; ifn = q) {
	    if ((q = index(ifn, ',')))
		*q++ = 0;
	    if (!(n = strlen(ifn)))
		continue;
	    n = (ifn[n-1] == '*') ? n - 1 : IFNAMSIZ + 1;
	    FORALLIFP(ifp) {
		if (ifp->usefpref < 0 ||
		    strncmp(ifp->name, ifn, n) != 0)
			continue;
		if (mode == 'n') {
		    ifp->usefpref = -1;
		    continue;
		}
		ifp->usefpref = 1;
		if (pf.valpref.val_date < 0) {
		    pf.valpref.val = ifp->opt.valpref.val;
		    pf.valpref.val_date = ifp->opt.valpref.val_date;
		}
		if (pf.valpref.pref_date < 0) {
		    pf.valpref.pref = ifp->opt.valpref.pref;
		    pf.valpref.pref_date = ifp->opt.valpref.pref_date;
		}
		if (mode == 'p') {
		    pp2 = 0;
		    setpf = pf;
		    if (notaprefix || ifp->ltoken == 0)
			goto setpref;
		    for (n = sizeof(setpf.addr) - ifp->ltoken;
				n < sizeof(setpf.addr); n++)
			if (setpf.addr.s6_addr[n])
			    goto setpref;
		    for (n = sizeof(setpf.addr) - ifp->ltoken;
				n < sizeof(setpf.addr); n++)
			setpf.addr.s6_addr[n] = ifp->llocal.s6_addr[n];
			/* no mk_pref, we will test plen short enough */
		    goto setpref;
		}
		FORALLPP(pp2, ifp) {
		    if (mode == 'm') {
			if (pp2->plen < pf.plen)
			    goto nxt;
			for (n = 0; n < sizeof(pf.pmask); n++)
			    if ((pf.addr.s6_addr[n] & pf.pmask.s6_addr[n]) !=
				(pp2->pref.s6_addr[n] & pf.pmask.s6_addr[n]))
				goto nxt;
		    }
		    setpf = *pp2;
		    setpf.valpref = pf.valpref;
		    setpf.valpref.imported = 1;
		    setpf.flags = pf.flags;
		    setpf.slen = pf.slen;
	setpref:
		    LOOPALLPP(pp = FFIRST(ifp), pp)
			if (pp->plen == setpf.plen &&
			    SAME_ADDR6(pp->pref, setpf.pref)) {
			    return "prefix already announced";
		    }
		    if (Calloc(pp, prefix) == NULL) {
			if (verbose)
			    logerr("out of memory");
			return "out of memory";
		    }
		    if ((setpf.plen < 16) ||
			((setpf.plen != 128) &&
			 (setpf.plen > 128 - 8 * ifp->ltoken))) {
			if (verbose)
				logerr("bad prefix length %d (should be %d)",
				       setpf.plen, 128 - 8 * ifp->ltoken);
		    }
		    *pp = setpf;
		    pp->ifp = ifp;
		    TAILQ_INIT(&pp->anycasts);
		    TAILQ_INIT(&pp->hagents);
		    pp->hacnt = -1;
		    TAILQ_INSERT_TAIL(&ifp->fprefixes, pp, prf_list);
		    if (pp2 == 0)
			goto nxt2;
	nxt: ;
		}
	nxt2:
		pf.valpref = prvalpref;
	    }
	}
	return (char *)0;
}

static char *
read_opt(p, mode)
	u_char *p;
	int *mode;
{
	struct option of;
	u_char *q, *nxt, *arg;
	struct interface *ifp;
	char *ifn = 0;
	u_int n;

	of = defopt;
	scan(p, &arg, &nxt);

	for (;;) {
		p = nxt;
		if (!*p)
			break;
		switch (scan(p, &arg, &nxt)) {
		    default:
			return errsynt;
		    case KIF:
			ifn = arg;
			break;
		    case KFLAG:
			of.flags = 0;
			while ((n = (u_char)*arg++)) switch (tolower(n)) {
			    case 'm':
				of.flags |= ND_RA_FLAG_MANAGED;
				break;
			    case 'o':
				of.flags |= ND_RA_FLAG_OTHER;
				break;
			    case 'h':
				of.flags |= ND_RA_FLAG_HAGENT;
				break;
			    default:
				return "invalid flag field";
			}
			break;
		    case KRIN:
			switch (tolower(*arg)) {
			    case 'n':
				of.no_ripin = 1;
				break;
			    case 'y':
				of.no_ripin = 0;
				break;
			    default:
				return "invalid ripin field";
			}
			break;
		    case KROUT:
			switch (tolower(*arg)) {
			    case 'n':
				of.no_ripout = 1;
				break;
			    case 'y':
				of.optsplit = 1;
				of.no_ripout = 0;
				break;
			    case 'r':
				of.optsplit = -1;
				of.no_ripout = 0;
				break;
			    case 's':
				of.optsplit = 0;
				of.no_ripout = 0;
				break;
			    default:
				return "invalid ripout field";
			}
			break;
		    case KRTADV:
			switch (tolower(*arg)) {
			    case 'n':
				of.no_rtadv = 1;
				break;
			    case 'y':
				of.no_rtadv = 0;
				break;
			    default:
				of.no_rtadv = 0;
				if (!isdigit(*arg))
					return "invalid rtadv field";
                                of.maxrtadv = strtoul(arg, (char **)&q, 0);
                                if (*q)
                                        of.minrtadv = strtoul(q+1, 0, 0);
                                else
                                        of.minrtadv = 0;
				of.maxrtadv *= M_SECOND; /* milliseconds */
				of.minrtadv *= M_SECOND; /* milliseconds */
				if (of.maxrtadv < MIN_RTR_ADV_INTERVAL ||
				    of.maxrtadv > MAX_RTR_ADV_INTERVAL)
					of.maxrtadv = DEF_RTR_ADV_INTERVAL;
				if (of.minrtadv < (1 * M_SECOND) ||
				    of.minrtadv > ((of.maxrtadv * 3) / 4))
					of.minrtadv = of.maxrtadv / 3;
				break;
                        }
			break;
		    case KLIFE:
			if (!isdigit(*arg))
				return "invalid life field";
			n = strtoul(arg, 0, 0);
			if (n > MAX_RT_LIFE)
				n = MAX_RT_LIFE;	/* max in seconds */
			of.life = n;
			break;
		    case KREACH:
			if (!isdigit(*arg))
				return "invalid reach field";
			n = strtoul(arg, 0, 0);
			if (n && n < 2)
				n = 2;
			if (n > 3600)
				n = 3600;
			of.reach = n * 1000; /* milliseconds */
			break;
		    case KRETRANS:
			if (!isdigit(*arg))
				return "invalid retrans field";
			n = strtoul(arg, 0, 0);
			if (n && n < 2)
				n = 2;
			if (n > 3600)
				n = 3600;
			of.retrans = n * 1000; /* milliseconds */
			break;
		    case KMTUIF:
			of.extra |= ADVMTU | ADVMTUIF;
			break;
		    case KMTU:
			if (!isdigit(*arg)) {
			errmtu:
				return "invalid mtu field";
			}
			n = strtoul(arg, 0, 0);
			if (n && n < IP6_MMTU)
				goto errmtu;
			of.mtuadv = n;
			of.extra |= ADVMTU;
			break;
		    case KHAADV:
			switch (tolower(*arg)) {
			    case 'n':
				of.extra &= ~(ADVADVINT | ADVHAINF);
				break;
			    case 'y':
				of.extra |= ADVADVINT;
				break;
			    default:
				of.extra |= ADVADVINT | ADVHAINF;
				if (!isdigit(*arg) && (*arg != '-'))
					return "invalid haadv field";
				of.hapref = strtol(arg, (char **)&q, 0);
				if (*q)
					of.halife = strtoul(q+1, 0, 0);
				break;
			}
			break;
		    case KPREF:
			switch (tolower(*arg)) {
			    case 'n':
				of.nopref = 1;
				break;
			    case 'y':
				of.nopref = 0;
				break;
			    default:
			    errpref:
				return "invalid prefix options";
			}
			break;
		    case KVAL:
			if (convvalpref(arg,
					&of.valpref.val_date,
					&of.valpref.val) < 0)
				goto errpref;
			break;
		    case KDEPREC:
			if (convvalpref(arg,
					&of.valpref.pref_date,
					&of.valpref.pref) < 0)
				goto errpref;
			break;
		}
	}
	/* check and fix home-agent infos */
	if (of.extra & ADVHAINF) {
		if (of.halife == 0)
			of.halife = of.life;
		if (of.halife == 0)
			of.extra &= ~ADVHAINF;
	}
	if (!ifn) {
		if (*mode) {
			return "default option must be first line";
		}
		defopt = of;
		FORALLIFP(ifp)
			ifp->opt = of;
		return (char *)0;
	}
	*mode = 1;
	for (; ifn; ifn = q) {
		if ((q = index(ifn, ',')))
			*q++ = 0;
		if (!(n = strlen(ifn)))
			continue;
		n = (ifn[n - 1] == '*') ? n - 1 : IFNAMSIZ + 1;
		FORALLIFP(ifp)
			if (strncmp(ifp->name, ifn, n) == 0)
				ifp->opt = of;
	}
	return (char *)0;
}

static char *
read_route(p)
	u_char *p;
{
	char aa[60], ag[60], x[4];
	struct in6_addr addr, gateway;
	char interf[20];
	char kw[16];
	int prefix, metric, n;
	int flag;
	struct interface *ifp;

	n = sscanf(p, "%60[0-9a-fA-F:]/%d metric %d %s %60s %20s %2s",
			aa, &prefix, &metric, kw, ag, interf, x);
	if (n != 3 && n != 6)
		return errsynt;
	if (n != 3) {
		if (strcmp(kw, "gateway") != 0)
			return errsynt;
		if (inet_pton(AF_INET6, ag, &gateway) <= 0)
			return "invalid gateway address";
		FORALLIFP(ifp)
			if (strcmp(interf, ifp->name) == 0)
				break;
		if (ifp == NULL) {
			return "unsupported interface";
		}
		flag = RT_MANUAL | RT_FILE | RT_SEEN;
	} else {
		bcopy(&localH, &gateway, sizeof gateway);
		ifp = NULL;
		flag = RT_MANUAL | RT_FILE | RT_SEEN | RT_DONTINST;
	}
	if (inet_pton(AF_INET6, aa, &addr) <= 0 || prefix < 0 || prefix > 128)
		return "invalid prefix";
	if (metric < 0 || metric >= HOPCNT_INFINITY-1)
		return "invalid metric";

	if (optrip)
		recv_one_route(&addr, prefix, metric, flag, &gateway, ifp);
	return (char *)0;
}

static void
setopt(ifp)
	struct interface *ifp;
{
	if (started >= 0) {
		if (IS_UP(ifp) && ifp->opt.no_rtadv != ifp->oldopt.d_no_rtadv)
			if_rtadv(ifp, ifp->oldopt.d_no_rtadv);
/*	XXX NO_ANYCAST(ifp) == ifp->opt.no_rtadv */
		if (IS_UP(ifp) && ifp->opt.no_rtadv != ifp->oldopt.d_no_rtadv)
			if_apadv(ifp, ifp->oldopt.d_no_rtadv);
		if (IS_UP(ifp) && ifp->opt.no_ripin != ifp->oldopt.d_no_ripin)
			if_ripin(ifp, ifp->oldopt.d_no_ripin);
		if (IS_UP(ifp) && ifp->opt.no_ripout != ifp->oldopt.d_no_ripout)
			if_ripout(ifp, ifp->oldopt.d_no_ripout);
	}
}

static int
scan(s, arg, nxt)
	u_char *s;
	u_char **arg, **nxt;
{
	u_char *p, d = 0;
	struct kwt *k;

	while (isspace(*s))
		s++;
	if (!*s)
		return LEOF;
	p = s;
	while (isalnum(*p))
		p++;
	if (*p && !isspace(*p))
		d = *p++;
	if (arg)
		*arg = p;
	while (*p && !isspace(*p))
		p++;
	if (nxt) {
		if (arg && *p)
			*p++ = 0;
		while (isspace(*p))
			p++;
		*nxt = p;
	}
	for (k = kwtb; k->n; k++)
		if (k->delim == d && strncasecmp(k->n, s, k->l) == 0)
			break;
	if (k->n)
		return k->val;
	else
		return LNONE;
}

/*
 * interpret a valid/preferred field, set type, value in arg 2/3
 * inspired form touch and bind
 * return : -1 on err, no arg modified
 */
#define	ATOI2(ar)	((ar)[0] - '0') * 10 + ((ar)[1] - '0'); (ar) += 2;

int
convvalpref(arg, type, result)
	char *arg;
	char *type;
	union datelife *result;
{
	u_long val;

	if (*arg == '@') {	/* @[CC]YYMMDDhhmm[.SS] */
		time_t nows, thens;
		struct tm *t;
		int v;
		char *p, *dot = NULL;

		arg++;
		for (p = arg; *p; ++p) {
			if (isdigit(*p))
				continue;
			if (*p == '.' && dot == NULL) {
				dot = p;
				continue;
			}
			return -1;
		}
		if (dot) {
			if (p != dot + 3)
				return -1;
			p =  dot++;
		}

		nows = now.tv_sec;
		if ((t = localtime(&nows)) == NULL)
			return -1;

		t->tm_isdst = -1;		/* Figure out DST. */
		if (dot) {
			t->tm_sec = ATOI2(dot);
		} else
			t->tm_sec = 0;
		switch(p - arg) {
		case 12:			/* CCYYMMDDhhmm */
			v = ATOI2(arg);
			v = 100 * v - 1900;
			v += ATOI2(arg);
			goto ydone;
		case 10:			/* YYMMDDhhmm */
			v = ATOI2(arg);
			if (v < 69)	/* y2k ! */
				v += 100;
		ydone:
			t->tm_year = v;
			t->tm_mon = ATOI2(arg); /* from 01-12 to 00-11 */
			t->tm_mon -= 1; /* from 01-12 to 00-11 */
			t->tm_mday = ATOI2(arg);
			t->tm_hour = ATOI2(arg);
			t->tm_min = ATOI2(arg);
			break;
		default:
			return -1;
		}
		if (t->tm_mon > 12 || t->tm_mday > 31 ||
		    t->tm_hour > 23 ||t->tm_min > 59 || t->tm_sec > 59)
			return -1;

		thens = mktime(t);
		if (thens == -1)
			return -1;

		result->date = thens;
		*type = 1;
		return 1;

	} else if (strpbrk(arg, "wdhmsWDHMS") && !strpbrk(arg, "xX")) {
		u_long v;
		int c;

		val = 0;
		while (*arg != '\0') {
			if (!isdigit(*arg))
				return -1;
			v = 0;
			while ((c = *arg++) && isdigit(c))
				v = 10 * v + (c - '0');
			if (islower(c))
				c = toupper(c);
			switch (c) {
			case 'W':  v *= 7;
			case 'D':  v *= 24;
			case 'H':  v *= 60;
			case 'M':  v *= 60;
			case 'S':  break;
			default:
				return -1;
			}
			val += v;
		}
	} else  if (*arg == '-' || isdigit(*arg)) {
		char *p;

		val = strtoul(arg, &p, 0);
		if (*p)
			return -1;
	} else
		return -1;

	result->life = val;
	*type = 0;
	return 0;
}
