/* Neighbor discovery specific routines for ATM */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>
#include <net/if.h>
#include <netinet/in.h>
#include <net/route.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "atmcall.h"
#include "scl_io.h"

#include <ndpd.h>

/* type-length encoding in link-layer infos */

#define MAX_ATM_LLSIZE		5
#define XTL_RESERVED		0x80
#define XTL_EASA		0x00
#define XTL_E164		0x40
#define XTL_LMASK		0x3f

static struct nd_opt_hdr *atm_addr2llinfo(struct svc_addr *);
static int atm_llinfo2addr(struct nd_opt_hdr *, struct svc_addr *);
static struct neighbor *atm_search_ndp(struct svc_addr *);
static int atm_verify_ndp(struct neighbor *, struct svc_addr *);

/*
 * Hardware dependent interface initialisations
 */
int
new_atm_if(ifm, first)
	struct if_msghdr *ifm;
	int first;
{
	struct sockaddr_dl *sdl;
	int hasphy = 0, ltoken = 0, lmac = 0;

	/* first time? */

	if (first != 1)
		return;

	/* addresses should be interface only */

	if (ifm->ifm_addrs != RTA_IFP) {
		logerr("bad ifinfo (no ifp#%d)", ifm->ifm_index);
		return 0;
	}

	/* get and check the data-link address */

	sdl = (struct sockaddr_dl *)((char *)ifm + sizeof(struct if_msghdr));
	if ((sdl->sdl_family != AF_LINK) || (sdl->sdl_nlen == 0)) {
		logerr("bad sdl address");
		return 0;
	}

	/* verify flags and MTU */

	if ((ifm->ifm_flags & IFF_LOOPBACK) ||
	    (ifm->ifm_flags & IFF_POINTOPOINT) ||
	    (ifm->ifm_flags & IFF_BROADCAST) ||
	    ((ifm->ifm_flags & IFF_MULTICAST) != IFF_MULTICAST) ||
	    (ifm->ifm_data.ifi_mtu < IP6_MMTU)) {
		logerr("incompatible flags 0x%x for interface %.*s",
		       (unsigned)ifm->ifm_flags,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* verify type (IFT_AAL5 is not used) */

	if (ifm->ifm_data.ifi_type != IFT_ATM) {
		logerr("incompatible type %d for interface %.*s",
		       ifm->ifm_data.ifi_type,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* everything has 48 bit MAC address... */

	ltoken = MACTOKEN;
	lmac = MACMAC;
	if (ifm->ifm_data.ifi_addrlen != lmac) {
		logerr("bad MAC address length (%d != 6) for interface %.*s",
		       ifm->ifm_data.ifi_addrlen,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	if (sdl->sdl_alen != lmac) {
		logerr("unsupported type for interface %.*s",
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* fill the interface structure */

	itf.mtu = ifm->ifm_data.ifi_mtu;
	itf.flags = ifm->ifm_flags;
	itf.ltoken = ltoken;
	itf.lmac = lmac;
	bcopy(sdl->sdl_data, itf.name, sdl->sdl_nlen);
#ifdef notyet
	itf.ntosend = MAX_RTR_SOLICITATIONS;
#endif

	/* get the link-layer info */

	itf.llinfo = atm_addr2llinfo(&my_local_atm_addr);
	if (itf.llinfo == NULL)
		return 0;

	/* fill methods */

	itf.lladdr_get = atm_addr_get;
	itf.lladdr_cmp = atm_addr_cmp;
	itf.lladdr_updated = atm_addr_updated;
	itf.circuit_close = atm_circuit_close;
	itf.lladdr_pr = atm_addr_pr;
	itf.sdl_pr = atm_sdl_pr;

	/* success */

	if (verbose)
		logtrace("got interface %s#%d", itf.name, itf.index);
	return 1;
}

/*
 * Transform an ATM link-layer info in an usable stuff
 */
struct nd_opt_hdr *
atm_addr_get(lp)
	struct nd_opt_hdr *lp;
{
	u_char *cp = (u_char *)lp;
	int len = 4;

	/* checks (logerr: errors, logtrace: unsupported) */

	if ((lp->nd_opt_len < 1) || (lp->nd_opt_len > MAX_ATM_LLSIZE)) {
		/* bad length */
		if (verbose)
			logerr("ATM llinfo with bad length %d",
			       lp->nd_opt_len);
		return NULL;
	}
	if (cp[2] & XTL_RESERVED) {
		/* reserved bit set in NTL */
		if (verbose)
			logerr("bad NTL field (%x)", cp[2]);
		return NULL;
	}
	if (cp[2] & XTL_E164) {
		/* E.164 address */
		if (verbose)
			logtrace("unsupported E.164 address");
		return NULL;
	}
	if ((cp[2] & XTL_LMASK) > ATM_ESA_LEN) {
		/* EASE length should be 20 (ATM_ESA_LEN) */
		if (verbose)
			logerr("too large EASA (%d)", cp[2] & XTL_LMASK);
		return NULL;
	}
	if ((cp[2] != ATM_ESA_LEN) && verbose)
		logtrace("strange EASA length (%d)", cp[2]);
	if (cp[3] & XTL_RESERVED) {
		/* reserved bit set in STL */
		if (verbose)
			logerr("bad STL field (%x)", cp[3]);
		return NULL;
	}
	if (cp[3] != 0) {
		/* subaddress is present */
		if (verbose)
			logtrace("unsupported sub address");
		return NULL;
	}

	/* compute the length from addresses */

	len += cp[2] & XTL_LMASK;
	len += cp[3] & XTL_LMASK;
	len = 1 + ((len - 1) | 7);

	/* check the length */

	if (lp->nd_opt_len < (len >> 3)) {
		if (verbose)
			logerr("bad llinfo length (%d != %d)",
			       lp->nd_opt_len << 3, len);
		return NULL;
	}
	if (lp->nd_opt_len != (len >> 3)) {
		if (verbose)
			logtrace("strange llinfo length (%d != %d)",
				 lp->nd_opt_len << 3, len);
	}

	/* done */

	if (debug)
		logdebug("get ATM address %s", atm_addr_pr(lp));
	lp->nd_opt_type = 0;
	return lp;
}

/*
 * Compare two ATM link-layer infos (specialized bcmp, return 0 if equal)
 */
int
atm_addr_cmp(l1, l2)
	struct nd_opt_hdr *l1, *l2;
{
	u_char *cp1, *cp2;
	int len = 2;

	/* both null: equal */

	if ((l1 == NULL) && (l2 == NULL))
		return 0;

	/* only one is null: different */

	if ((l1 == NULL) || (l2 == NULL))
		return 1;

	/* binary comparison */

	cp1 = (u_char *)l1;
	cp2 = (u_char *)l2;
	len += cp1[2] & XTL_LMASK;
	len += cp1[3] & XTL_LMASK;
	return (bcmp(cp1 + 2, cp2 + 2, len));
}

/*
 * Update a data-link address from an ATM link-layer info
 * Four cases:
 *  - the neighbor is ourself
 *  - we have already the circuit (opened by the peer)
 *  - llinfo/ATM address has changed
 *  - no circuit and llinfo/ATM address before
 */
void
atm_addr_updated(ndp)
	struct neighbor *ndp;
{
	ATMCALL *call = (ATMCALL *)ndp->circuit;
	struct svc_addr addr;

	/* special case for myself */

	if (ndp->flags & ND6F_IS_SELF) {
		ndp->flags |= ND6F_HAS_LLADDR;
		return;
	}

	/* translate llinfo to an ATM address */

	bzero(&addr, sizeof (struct svc_addr));
	if (!atm_llinfo2addr(ndp->llinfo, &addr)) {
		if (verbose)
			logerr("can't convert llinfo to ATM addr");
		return;
	}

	/* NdAdv on a new circuit? */

	if (call != NULL)
		goto cleanup;

	call = atmcall_lookupa(&addr);
	if (call == NULL)
		goto cleanup;

	/* passive open has already got the circuit */

	if (verbose)
		logtrace("neighbor %s get passive circuit %u",
			 inet6pr(&ndp->addr), call->vcc);

	if (call->state != connected) {
		logerr("connecting_circuit: not connected");
		goto cleanup;
	}

	/* link/update then done */

	ndp->circuit = call;
	call->ndp = ndp;
	bcopy(&ndp->addr, &call->rin6, sizeof(struct in6_addr));
	atm_sdl_update(ndp);
	return;

	/* get rid of past */

    cleanup:

	if (ndp->flags & ND6F_HAS_ROUTE)
		del_route(ndp);
	if (call != NULL)
		(void) active_close_atmcall(0, 0, call);
	call = (ATMCALL *)ndp->circuit;
	if (call != NULL) {
		ndp->state &= ~ND6F_HAS_LLADDR;
		ndp->circuit = NULL;
		call->ndp = NULL;
	}

	/* try to get a circuit */

	(void) atmcall_setup(&addr, NULL,
			     &ndp->addr, &qos_default_p2p,
			     -1, NULL, NULL);

	/* ND6F_HAS_LLADDR will be set by connecting_circuit() callback */
}

/*
 * Close a circuit
 */
void
atm_circuit_close(ndp)
	struct neighbor *ndp;
{
	ATMCALL *call;

	/* special case: NULL argument: close_all() */

	if (ndp == NULL) {
		(void) close_all(sig_down);
		return;
	}

	/* initiate release */

	call = (ATMCALL *)ndp->circuit;
	if (call == NULL)
		return;
	(void) active_close_atmcall(sig_down, 0, call);

	/* should be closed and closing_circuit() called */

	call = (ATMCALL *)ndp->circuit;
	if (call == NULL)
		return;

	/* finished */

	ndp->state &= ~ND6F_HAS_LLADDR;
	ndp->circuit = NULL;
	call->ndp = NULL;
}

static char atmname[5][MAX_ATM_ADDR_LEN + 1];
static int atmcnt = 0;

/*
 * Print an ATM link-layer info
 */
char *
atm_addr_pr(lp)
	struct nd_opt_hdr *lp;
{
	struct svc_addr addr;

	/* get the svc_addr from the link-layer info */

	if (lp == NULL)
		return (char *)"[NULL]";
	bzero(&addr, sizeof(struct svc_addr));
	if (!atm_llinfo2addr(lp, &addr))
		return (char *)"[XXX?]";

	/* print it in the next buffer */

	atmcnt++;
	atmcnt &= 3;
	if (atm2text_svc(atmname[atmcnt], MAX_ATM_ADDR_LEN + 1,
			 &addr, A2T_PRETTY | A2T_LOCAL) < 0)
		return (char *)"[XX??]";

	/* done */

	return atmname[atmcnt];
}

/*
 * Print an ATM data-link address (ATM pseudo header)
 */
char *
atm_sdl_pr(sdl)
	struct sockaddr_dl *sdl;
{
	static char s[sizeof("[bad-lenXXX]")];
	char aph[4];

	/* get the ATM pseudo header */

	if (sdl == NULL)
		return (char *)"[NULL]";
	if (sdl->sdl_alen != sizeof(aph)) {
		(void) snprintf(s, sizeof(s), "[bad-len%d]", sdl->sdl_alen);
		return s;
	}
	bcopy(LLADDR(sdl), aph, sizeof(aph));

	/* print it */

	if (aph[0] != 3) {
		(void) snprintf(s, sizeof(s), "[bad-aph%d]", aph[0] & 0xff);
		return s;
	}
	(void) snprintf(s, sizeof(s), "%d.%d",
			aph[1], (aph[2] << 8) | (aph[3] & 0xff));
	return s;
}

/*
 * Update sockaddr_dl sdl structure
 */
void
atm_sdl_update(ndp)
	struct neighbor *ndp;
{
	ATMCALL *call = (ATMCALL *)ndp->circuit;
	char aph[4];

	/* data-link address is an ATM pseudo header */

	aph[0] = 3;
	aph[1] = call->vpi;
	aph[2] = call->vci >> 8;
	aph[3] = call->vci & 0xff;
	bcopy(aph,  LLADDR(&ndp->sdl), sizeof(aph));
	ndp->sdl.sdl_alen = sizeof(aph);
	ndp->flags |= ND6F_HAS_LLADDR;
}

/*
 * A circuit has been connected
 */
void
connecting_circuit(ccp)
	void *ccp;
{
	ATMCALL *call = (ATMCALL *)ccp;
	struct neighbor *ndp;
	char name[MAX_ATM_ADDR_LEN + 1];

	/* get printable form of the ATM address of the peer */

	bzero(name, MAX_ATM_ADDR_LEN + 1);
	(void) strcpy(name, "[bad ATM address]");
	(void) atm2text_svc(name, MAX_ATM_ADDR_LEN + 1,
			    &call->rsvcaddr, A2T_PRETTY | A2T_LOCAL);

	if (verbose)
		logtrace("connecting circuit to %s [%s]",
			 name, inet6pr(&call->rin6));

	/* sanity check */

	if (call->state != connected) {
		logerr("connecting_circuit: not connected");
		return;
	}

	/* get the neighbor entry */

	if (IN6_IS_ADDR_UNSPECIFIED(&call->rin6)) {
	    retry:
		/* can't use the IPv6 address, try the ATM address */
		ndp = atm_search_ndp(&call->rsvcaddr);
		if (ndp == NULL) {
			if (verbose)
				logtrace("connecting_circuit: unknown peer");
			return;
		}
	} else
		ndp = ndplookup(&call->rin6, 1);

	/* easy cases */

	if (ndp == NULL)
		return;
	if (ndp->circuit != NULL) {
		if ((ndp->circuit != ccp) && verbose)
			logtrace("connecting_circuit: call collision");
		return;
	}

	/* verify llinfo */

	if (!atm_verify_ndp(ndp, &call->rsvcaddr)) {
		if (verbose)
			logerr("connecting_circuit: mismatch %s [%s] %s [%s]",
			       inet6pr(&ndp->addr), atm_addr_pr(ndp->llinfo),
			       inet6pr(&call->rin6), name);
		/* retry without an IPv6 address */
		goto retry;
	}

	/* link the neighbor entry and the ATM circuit together */

	ndp->circuit = ccp;
	call->ndp = ndp;
	bcopy(&ndp->addr, &call->rin6, sizeof(struct in6_addr));

	/* we have got a circuit to the peer */

	if ((ndp->flags & ND6F_HAS_LLADDR) && verbose)
		logerr("connecting_circuit: new circuit but sdl was set?");
	atm_sdl_update(ndp);

	/* compute/execute action according to the neighbor state */

	switch (ndp->state) {
	    case ND6S_REJECT:
		/* some progress done */
		(void) del_route(ndp);
		/* fall in */

	    case ND6S_REACHABLE:
	    case ND6S_STALE:
	    case ND6S_DELAY:
	    case ND6S_PROBE:
		/* should not happen */
		if (verbose)
			logerr("connecting_circuit: %s in state %d?",
			       inet6pr(&ndp->addr), ndp->state);
		/* fall in */

	    case ND6S_INCOMPLETE:
		/* new and free neighbor */
		if (ndp->flags & ND6F_HAS_ROUTE) {
			/* update the route */
			change_route(ndp);
			enter_state(ndp, ND6S_STALE);
		} else {
			/* wait for the route */
			enter_pending(ndp, ND6S_STALE);
		}
		return;

	    case ND6S_PENDING:
		/* likely waiting for the circuit */
		if (ndp->flags & ND6F_HAS_ROUTE) {
			/* update the route */
			change_route(ndp);
			leave_pending(ndp);
		}
		return;

	    case ND6S_LOCKED:
		/* should not happen? */
		if (verbose)
			logtrace("connecting_circuit: locked %s",
				 inet6pr(&ndp->addr));
		return;

	    default:
		/* should not happen */
		logerr("connecting_circuit: %s has impossible state %d",
		       inet6pr(&ndp->addr), ndp->state);
		del_neighbor(ndp);
	}
}

/*
 * A circuit is being closed
 */
void
closing_circuit(ccp, freeing)
	void *ccp;
	int freeing;
{
	ATMCALL *call = (ATMCALL *)ccp;
	struct neighbor *ndp = call->ndp;
	char name[MAX_ATM_ADDR_LEN + 1];

	/* get printable form of the ATM address of the peer */

	bzero(name, MAX_ATM_ADDR_LEN + 1);
	(void) strcpy(name, "[bad ATM address]");
	(void) atm2text_svc(name, MAX_ATM_ADDR_LEN + 1,
			    &call->rsvcaddr, A2T_PRETTY | A2T_LOCAL);

	if (verbose)
		logtrace("%s circuit to %s [%s]",
			 freeing ? "freeing" : "closing",
			 name, inet6pr(&call->rin6));

	if (ndp != NULL)
		goto found;

	/* get the neighbor entry */

	if (IN6_IS_ADDR_UNSPECIFIED(&call->rin6)) {
		/* the IPv6 address is not available */
		ndp = atm_search_ndp(&call->rsvcaddr);
		if (ndp != NULL)
			goto found;
		if (verbose && !freeing)
			logerr("closing_circuit: no neighbor");
		return;
	} else
		ndp = ndplookup(&call->rin6, 0);
	if (ndp != NULL)
		goto found;

	/* the IPv6 address gave nothing, try with the ATM address */

	ndp = atm_search_ndp(&call->rsvcaddr);
	if (ndp != NULL)
		goto found;
	if (verbose && !freeing)
		logerr("closing_circuit: no neighbor %s",
		       inet6pr(&call->rin6));
	return;

    found:

	/* verify llinfo */

	if (!atm_verify_ndp(ndp, &call->rsvcaddr)) {
		if (verbose)
			logerr("connecting_circuit: mismatch %s [%s] %s [%s]",
			       inet6pr(&ndp->addr), atm_addr_pr(ndp->llinfo),
			       inet6pr(&call->rin6), name);
		return;
	}

	/* already in del_neighbor() ? */

	if (ndp->ngb_list.tqe_next == NULL)
		return;

	/* cleanup */

	if (ndp->state & ND6F_HAS_ROUTE)
		(void) del_route(ndp);
	ndp->state &= ~(ND6F_HAS_ROUTE | ND6F_HAS_LLADDR);
	ndp->circuit = NULL;
	call->ndp = NULL;
	del_neighbor(ndp);
	return;
}

/*
 * Transform an ATM address in a link-layer info structure
 */
static struct nd_opt_hdr *
atm_addr2llinfo(addr)
	struct svc_addr *addr;
{
	struct nd_opt_hdr *lp = NULL;
	u_char *cp;
	int len = 4;

	/* no address? */

	if ((addr == NULL) || !atmsvc_addr_in_use(*addr)) {
		if (verbose)
			logerr("atm_addr2llinfo: no address");
		return NULL;
	}

	/* E.164 address? */

	if (addr->pub[0]) {
		if (verbose)
			logtrace("atm_addr2llinfo: unsupported E.164");
		return NULL;
	}

	/* allocate the link-layer info structure */

	len += ATM_ESA_LEN;
	lp = (struct nd_opt_hdr *)calloc(1, len);
	if (lp == NULL) {
		logerr("out of memory");
		return NULL;
	}

	/* fill it */

	lp->nd_opt_len = len >> 3;
	cp = (u_char *)lp;
	cp[2] = ATM_ESA_LEN;
	bcopy(addr->prv, cp + 4, ATM_ESA_LEN);
	return lp;
}

/*
 * Transform a link-layer info structure in an ATM address
 */
static int
atm_llinfo2addr(lp, addr)
	struct nd_opt_hdr *lp;
	struct svc_addr *addr;
{
	u_char *cp = (u_char *)lp;

	/* checks */

	if (lp == NULL)
		return 0;
	if (addr == NULL)
		return 0;
	if (cp[2] != ATM_ESA_LEN)
		return 0;
	if (cp[3] != 0)
		return 0;

	/* do it */

	bcopy(cp + 4, addr->prv, ATM_ESA_LEN);
	addr->pub[0] = 0;
	return 1;
}

/*
 * Search a neighbor entry matching an ATM address
 */
static struct neighbor *
atm_search_ndp(addr)
	struct svc_addr *addr;
{
	struct neighbor *ndp;
	struct nd_opt_hdr *lp;

	/* get the link-layer info structure from the ATM address */

	lp = atm_addr2llinfo(addr);
	if (lp == NULL)
		return NULL;

	/* search for a neighbor entry with the same link-layer info */

	FORALLND(ndp)
		if (atm_addr_cmp(ndp->llinfo, lp) == 0)
			break;

	/* done */

	free(lp);
	return ndp;
}

/*
 * Verify the match between:
 *     the neighbor entry link-layer info
 *     an ATM address
 */
static int
atm_verify_ndp(ndp, addr)
	struct neighbor *ndp;
	struct svc_addr *addr;
{
	struct nd_opt_hdr *lp;
	int same = 0;

	/* get the link-layer info structure from the ATM address */

	lp = atm_addr2llinfo(addr);
	if (lp == NULL)
		return 0;

	/* compare link-layer infos */

	if (atm_addr_cmp(ndp->llinfo, lp) == 0)
		same = 1;

	/* done */

	free(lp);
	return same;
}

/*
 * Definitions for functions called by timeout
 */

void
T(del_neighbor) (arg)
	void *arg;
{
	struct neighbor *ndp = (struct neighbor *)arg;

	ndp->tp = NULL;
	del_neighbor(ndp);
}

void
T(retransmit) (arg)
	void *arg;
{
	struct neighbor *ndp = (struct neighbor *)arg;

	ndp->tp = NULL;
	retransmit(ndp);
}

void
T(do_nud) (arg)
	void *arg;
{
	struct neighbor *ndp = (struct neighbor *)arg;

	ndp->tp = NULL;
	do_nud(ndp);
}

void
T(probe) (arg)
	void *arg;
{
	struct neighbor *ndp = (struct neighbor *)arg;

	ndp->tp = NULL;
	probe(ndp);
}

void
T(setReachableTime) (arg)
	void *arg;
{
	u_int basereach = (u_int)arg;

	randpar = NULL;
	setReachableTime(basereach);
}
