/*
 * Ask for a shortcut: send a neighbor solicitation to the router
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sysexits.h>

#include <net/if.h>
#include <net/if_dl.h>
#ifndef if_type
#include <net/if_var.h>
#endif
#include <net/route.h>

#include <netinet/in_systm.h>
#define INET6
#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <netinet/ip6_opts.h>

#include <netdb.h>
#include <arpa/inet.h>

#define	MAXPACKET	8192	/* max packet size */
#define MAX_MSG_ROUTE	2048

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(int32_t) - 1))) : sizeof(int32_t))
#define ADVANCE(x, n) (x += ROUNDUP(((struct sockaddr *)n)->sa_len))

int sd;
int pid;
int verbose;
struct sockaddr_in6 peer;
struct sockaddr_in6 router;
u_int ifindex;
u_int site;
struct {
       	struct rt_msghdr m_rtm;
        char m_space[MAX_MSG_ROUTE - sizeof(struct rt_msghdr)];
} m_rtmsg;
#define rtm	m_rtmsg.m_rtm
struct nd_neighbor_solicit *nsp;
char packet[MAXPACKET];

int main(int, char *[]);
void argtoindex(const char *);
void argtosite(const char *);
void get_router(void);
void send_ns(void);

int
main(argc, argv)
	int argc;
	char *argv[];
{
	char ch;
	struct hostent *hp;
	int val;
	extern int optind;
	extern char *optarg;

	while ((ch = getopt(argc, argv, "i:r:s:v")) != EOF)
		switch (ch) {
		    case 'i':
			argtoindex(optarg);
			if (ifindex == 0) {
				(void) fprintf(stderr,
					       "shortcut: bad interface %s\n",
					       optarg);
				exit(EX_OSERR);
			}
			break;
		    case 'r':
			hp = getipnodebyname(optarg, AF_INET6, 0, NULL);
			if (hp == NULL) {
				(void) fprintf(stderr,
					       "shortcut: bad router %s\n",
					       optarg);
				exit(EX_NOHOST);
			}
			router.sin6_family = hp->h_addrtype;
			router.sin6_len = sizeof(router);
			bcopy(hp->h_addr, &router.sin6_addr, hp->h_length);
			freehostent(hp);
			break;
		    case 's':
			argtosite(optarg);
			if (site == 0) {
				(void) fprintf(stderr,
					       "shortcut: bad site %s\n",
					       optarg);
				exit(EX_OSERR);
			}
			break;
		    case 'v':
			verbose++;
			break;
		    default:
		    usage:
			(void) fprintf(stderr,
				       "usage: shortcut [-v] [-i ifn] "
				       "[-s site] [-r router] peer\n");
			exit(EX_USAGE);
		}

	argc -= optind;
	argv += optind;

	if (argc != 1)
		goto usage;

	if ((ifindex == 0) || (if_indextoname(ifindex, NULL) == NULL)) {
		(void) fprintf(stderr, "shortcut: no index\n");
		exit(EX_DATAERR);
	}

	hp = getipnodebyname(argv[1], AF_INET6, 0, NULL);
	if (hp == NULL) {
		(void) fprintf(stderr, "shortcut: bad peer %s\n", argv[1]);
		exit(EX_NOHOST);
	}
	peer.sin6_family = hp->h_addrtype;
	peer.sin6_len = sizeof(peer);
	bcopy(hp->h_addr, &peer.sin6_addr, hp->h_length);
	freehostent(hp);
	if (IN6_IS_ADDR_MULTICAST(&peer.sin6_addr)) {
		(void) fprintf(stderr, "shortcut: multicast peer?\n");
		exit(EX_DATAERR);
	} else if (IN6_IS_ADDR_LINKLOCAL(&peer.sin6_addr)) {
		peer.sin6_scope_id = ifindex;
		if (verbose)
			(void) printf("link-local peer?\n");
	} else if (IN6_IS_ADDR_SITELOCAL(&peer.sin6_addr)) {
		peer.sin6_scope_id = site;
		if (site == 0) {
			(void) fprintf(stderr, "shortcut: no site\n");
			exit(EX_DATAERR);
		}
	}

	if (router.sin6_family != AF_INET6) {
		get_router();
		goto got_router;
	}
	if (IN6_IS_ADDR_LINKLOCAL(&router.sin6_addr))
		router.sin6_scope_id = ifindex;
	else if (IN6_IS_ADDR_SITELOCAL(&router.sin6_addr)) {
		router.sin6_scope_id = site;
		if (site == 0) {
			(void) fprintf(stderr, "shortcut: no site\n");
			exit(EX_DATAERR);
		}
	}

    got_router:
	if (IN6_IS_ADDR_MULTICAST(&router.sin6_addr)) {
		(void) fprintf(stderr, "shortcut: multicast router\n");
		exit(EX_DATAERR);
	} else if (!IN6_IS_ADDR_LINKLOCAL(&router.sin6_addr) && verbose)
		(void) printf("router is not link-local?\n");

	if ((sd = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		(void) fprintf(stderr,
			       "shortcut: socket (ICMP): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}

	val = 255;
	if (setsockopt(sd, IPPROTO_IPV6, IPV6_UNICAST_HOPS,
		       &val, sizeof(val)) < 0) {
		(void) fprintf(stderr,
			       "shortcut: setsockopt (UNICAST_HOPS): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}

	(void) shutdown(sd, 0);

	if (verbose) {
		char vp[64], vr[64], s[32];

		(void) inet_ntop(AF_INET6, &peer.sin6_addr, vp, 64);
		(void) inet_ntop(AF_INET6, &router.sin6_addr, vr, 64);
		if (site)
			(void) snprintf(s, 32, " [site %d]", site);
		(void) printf("shortcut on %s#%d%s for %s via %s\n",
			      if_indextoname(ifindex, NULL), ifindex,
			      site ? s : "", vp, vr);
	}

	router.sin6_flowinfo = ND_DEFAULT_CLASS;
	if (connect(sd, (struct sockaddr *)&router, sizeof(router)) < 0) {
		(void) fprintf(stderr,
			       "shortcut: connect (ICMP): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}

	send_ns();
	exit(EX_OK);
}

/*
 * Get an interface index from an argument
 */
void
argtoindex(name)
	const char *name;
{
	if (sscanf(name, "%u\0", &ifindex) == 1)
		return;
	ifindex = if_nametoindex(optarg);
	if (verbose)
		(void) printf("interface %s has index %d\n", name, ifindex);
}

/*
 * Get a site number from an argument
 */
void
argtosite(name)
	const char *name;
{
	struct ifreq ifr;

	if (sscanf(name, "%u\0", &site) == 1)
		return;

	/* the hard way... */

	if ((sd = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		(void) fprintf(stderr,
			       "shortcut: socket (UDP): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}
	(void) bzero(&ifr, sizeof(ifr));
	strnpy(&ifr.ifr_name, name, IFNAMSIZ);
	ifr.ifr_scid6[0] = MADDR6_SCP_SITE;
	if (ioctl(sd, SIOCGIFSCID6, &ifr) < 0) {
		(void) fprintf(stderr,
			       "shortcut: ioctl (SIOCGIFSCID6): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}
	(void) close(sd);
	site = ifr.ifr_scid6[1];
	if (verbose)
		(void) printf("interface %s is in site %d\n", name, site);
}

/*
 * Get router to peer
 */
void
get_router()
{
	char *cp;
	struct sockaddr_dl *sdl;
	int len, i;

	if (verbose) {
		char v[64];

		(void) inet_ntop(AF_INET6, &peer.sin6_addr, v, 64);
		(void) printf("get router for %s on %s#%d\n",
			      v, if_indextoname(ifindex, NULL), ifindex);
	}

	pid = getpid();

	if ((sd = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
		(void) fprintf(stderr,
			       "shortcut: socket (ROUTE): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_GET;
	rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_HOST;
	rtm.rtm_addrs = RTA_DST | RTA_IFP;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = 1;
	rtm.rtm_index = ifindex;

	cp = m_rtmsg.m_space;

	sdl = (struct sockaddr_dl *)cp;
	bcopy(&peer, sdl, sizeof(peer));
	ADVANCE(cp, sdl);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(*sdl);
	sdl->sdl_index = ifindex;
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;
	cp = (char *)&m_rtmsg;

	if (write(sd, cp, len) != len) {
		(void) fprintf(stderr,
			       "shortcut: write (ROUTE): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}

	for (;;) {
		bzero(cp, MAX_MSG_ROUTE);
		if ((len = read(sd, cp, MAX_MSG_ROUTE)) < 0) {
			(void) fprintf(stderr,
				       "shortcut: read (ROUTE): %s\n",
				       strerror(errno));
			exit(EX_OSERR);
		}
		if ((rtm.rtm_seq == 1) && (rtm.rtm_pid == pid))
			break;
	}
	(void) close(sd);
	if (rtm.rtm_version != RTM_VERSION) {
		(void) fprintf(stderr, "shortcut: bad route version\n");
		exit(EX_PROTOCOL);
	}
	if (rtm.rtm_msglen > len) {
		(void) fprintf(stderr, "shortcut: route message too large\n");
		exit(EX_PROTOCOL);
	}
	if (rtm.rtm_errno) {
		(void) fprintf(stderr,
			       "shortcut: get route failed: %s\n",
			       strerror(rtm.rtm_errno));
		exit(EX_OSERR);
	}
	if ((rtm.rtm_flags & RTF_GATEWAY) == 0) {
		(void) fprintf(stderr, "shortcut: get direct route\n");
		exit(EX_PROTOCOL);
	}
	if (rtm.rtm_index != ifindex) {
		(void) fprintf(stderr, "shortcut: get route with bad index\n");
		exit(EX_PROTOCOL);
	}
	if ((rtm.rtm_addrs & RTA_GATEWAY) == 0) {
		(void) fprintf(stderr,
			       "shortcut: no gateway in route message\n");
		exit(EX_PROTOCOL);
	}
	cp = m_rtmsg.m_space;
	for (i = 0; i < RTAX_MAX; i++) {
		if ((rtm.rtm_addrs & (1 << i)) == 0)
			continue;
		sdl = (struct sockaddr_dl *)cp;
		if ((i == RTAX_GATEWAY) &&
		    (sdl->sdl_family == AF_INET6) &&
		    (sdl->sdl_len == sizeof(router)))
			bcopy(sdl, &router, sdl->sdl_len);
		ADVANCE(cp, sdl);
	}

	if (verbose) {
		char v[64];

		(void) inet_ntop(AF_INET6, &router.sin6_addr, v, 64);
		(void) printf("got router %s [%d]\n", v, router.sin6_scope_id);
	}
}

/*
 * Send a neighbor solicitation to the router
 */
void
send_ns()
{
	int len;

	bzero(packet, MAXPACKET);
	len = sizeof(*nsp);
	nsp = (struct nd_neighbor_solicit *)packet;
	nsp->nd_ns_type = ND_NEIGHBOR_SOLICIT;
	bcopy(&peer.sin6_addr, &nsp->nd_ns_target, sizeof(struct in6_addr));

	if (verbose)
		(void) printf("sending\n");

	if (write(sd, packet, len) != len) {
		(void) fprintf(stderr,
			       "shortcut: write (ICMP): %s\n",
			       strerror(errno));
		exit(EX_OSERR);
	}

	if (verbose)
		(void) printf("sent!\n");
}
