/* Neighbor discovery state machine */

#include <ndpd.h>

u_int umaxtries = MAX_UNICAST_SOLICIT;		/* default: 3 */
u_int mmaxtries = MAX_MULTICAST_SOLICIT;	/* default: 3 */

/*
 * States :
 *
 * REJECT: packets are rejected with ICMPv6 unreachables (added)
 *
 * INCOMPLETE: address resolution is being performed (default)
 *
 * REACHABLE: reachable for at max ReachableTime
 *
 * STALE: reachable for at least ReachableTime, wait for a packet
 *
 * DELAY: reachable for at least ReachableTime, wait before probe
 *
 * PROBE: sending probes
 *
 * PENDING: action on route/circuit is pending (added)
 *
 * LOCKED: locked by user reachable state (added)
 */

/*
 * Pending states
 */
void
enter_pending(ndp, nextstate)
	struct neighbor *ndp;
	char nextstate;
{
	if (debug)
		logdebug("enter_pending for %s (%d --> %d)",
			 inet6pr(&ndp->addr), ndp->state, nextstate);

	/* reset timeout */

	if (ndp->tp)
		KILL_TIMEOUT(ndp->tp);
	ndp->tp = NULL;

	/* register next state */

	ndp->pstate = nextstate;
	ndp->state = ND6S_PENDING;

	/* don't wait for more than InactiveTime in pending state */

	if (InactiveTime == 0)
		return;		/* 0 is disabled */
	ndp->tp = NEW_TIMEOUT(MS_TO_SEC(InactiveTime),
			      MS_TO_USEC(InactiveTime),
			      del_neighbor, ndp);
}

/*
 * Leave the pending state for the registered next state
 */
void
leave_pending(ndp)
	struct neighbor *ndp;
{
	ndp->state = ndp->pstate;
	enter_state(ndp, ndp->state);
}

/*
 * Transition table: enter a new state
 */
void
enter_state(ndp, newstate)
	struct neighbor *ndp;
	char newstate;
{
	if (debug)
		logdebug("enter_state for %s (%d -> %d)",
			 inet6pr(&ndp->addr), ndp->state, newstate);

	/* reset timeout */

	if (ndp->tp)
		KILL_TIMEOUT(ndp->tp);
	ndp->tp = NULL;

	/* register next state when pending (see leave_pending()) */

	if (ndp->state == ND6S_PENDING) {
		enter_pending(ndp, newstate);
		return;
	}

	switch (newstate) {
	    case ND6S_REJECT:
	    do_reject:
		/* allways come from INCOMPLETE or PROBE */
		if ((ndp->state != ND6S_INCOMPLETE) &&
		    (ndp->state != ND6S_PROBE))
			goto illegal;
		ndp->state = ND6S_REJECT;
		if (((ndp->flags & ND6F_HAS_ROUTE) == 0) ||
		    !reject_route(ndp)) {
			/* no route */
			del_neighbor(ndp);
			return;
		}
		/* wait HoldDownTime then delete entry */
		ndp->tp = NEW_TIMEOUT(MS_TO_SEC(HoldDownTime),
				      MS_TO_USEC(HoldDownTime),
				      del_neighbor, ndp);
		return;

	    case ND6S_INCOMPLETE:
		/* allways come from INCOMPLETE (== default state) */
		if (ndp->state != ND6S_INCOMPLETE)
			goto illegal;
		/* retransmit up to mmaxtries neighbor solicitations */
		if (ndp->asked < mmaxtries) {
			ndp->asked++;
			(void) send_ndsol(ndp, NS6K_SOL);
			ndp->tp = NEW_TIMEOUT(MS_TO_SEC(RetransTimer),
					      MS_TO_USEC(RetransTimer),
					      retransmit, ndp);
			return;
		} /* else */
		/* failed: reject */
		goto do_reject;

	    case ND6S_REACHABLE:
		/* sanity check */
		if (((ndp->flags & ND6F_HAS_ROUTE) == 0) ||
		    ((ndp->flags & ND6F_HAS_LLADDR) == 0))
			if (verbose)
				logerr("%s become reachable with flags 0x%x",
				       inet6pr(&ndp->addr), ndp->flags);
		ndp->state = ND6S_REACHABLE;
		ndp->asked = 0;
		/* initiate NUD after ReachableTime */
		ndp->tp = NEW_TIMEOUT(MS_TO_SEC(ReachableTime),
				      MS_TO_USEC(ReachableTime),
				      do_nud, ndp);
		return;

	    case ND6S_STALE:
		/* recloned the route on next packet */
		if (ndp->flags & ND6F_HAS_ROUTE)
			(void) del_route(ndp);
		ndp->state = ND6S_STALE;
		/* don't wait for more than InactiveTime */
		if (InactiveTime == 0)
			return;		/* 0 is disabled */
		ndp->tp = NEW_TIMEOUT(MS_TO_SEC(InactiveTime),
				      MS_TO_USEC(InactiveTime),
				      del_neighbor, ndp);
		return;

	    case ND6S_DELAY:
		/* always come from stale */
		if (ndp->state != ND6S_STALE)
			goto illegal;
		ndp->state = ND6S_DELAY;
		ndp->asked = 0;
		/* probe in DelayTimer if still in this state */
		ndp->tp = NEW_TIMEOUT(MS_TO_SEC(DelayTimer),
				      MS_TO_USEC(DelayTimer),
				      probe, ndp);
		return;

	    case ND6S_PROBE:
		/* allways come from itself or delay */
		if ((ndp->state != ND6S_DELAY) && (ndp->state != ND6S_PROBE))
			goto illegal;
		ndp->state = ND6S_PROBE;
		/* retransmit up to umaxtries neighbor solicitations */
		if (ndp->asked < umaxtries) {
			ndp->asked++;
			(void) send_ndsol(ndp, NS6K_DIR);
			ndp->tp = NEW_TIMEOUT(MS_TO_SEC(RetransTimer),
					      MS_TO_USEC(RetransTimer),
					      probe, ndp);
			return;
		} /* else */
		/* failed: reject */
		goto do_reject;

	    case ND6S_PENDING:
		/* can't happen (done before) */
		goto illegal;

	    case ND6S_LOCKED:
		/* more check ZZZZZZZZZ */
		ndp->state = ND6S_LOCKED;
		return;

	    default:
	    illegal:
		/* illegal transitions */
		logerr("illegal transition for %s (%d -> %d)",
		       inet6pr(&ndp->addr), ndp->state, newstate);
		del_neighbor(ndp);
	}
}

/*
 * Retransmit multicast neighbor solicitations
 * (nothing to do in fact)
 */
void
retransmit(ndp)
	struct neighbor *ndp;
{
	enter_state(ndp, ND6S_INCOMPLETE);
}

/*
 * Probe (aka sending unicast neighbor solicitations)
 * (nothing to do again)
 */
void
probe(ndp)
	struct neighbor *ndp;
{
	enter_state(ndp, ND6S_PROBE);
}

/*
 * Initiate neighbor unreachability detection
 * (nothing to do again)
 */
void
do_nud(ndp)
	struct neighbor *ndp;
{
	enter_state(ndp, ND6S_STALE);
}

/*
 * Delete a neighbor
 */
void
del_neighbor(ndp)
	struct neighbor *ndp;
{
	if (verbose)
		logtrace("delete neighbor %s", inet6pr(&ndp->addr));

	/* zap timeout */

	if (ndp->tp)
		KILL_TIMEOUT(ndp->tp);
	ndp->tp = NULL;

	/* zap route */

	if (ndp->flags & ND6F_HAS_ROUTE) {
		switch (ndp->state) {
		    case ND6S_REJECT:
			break;
		    default:
			/* trigger ICMPv6 unreachable */
			(void) reject_route(ndp);
		}
		(void) del_route(ndp);
	}

	/* unlink entry */

	TAILQ_REMOVE(&neighbors, ndp, ngb_list);
	ndp->ngb_list.tqe_next = NULL;

	/* close any circuit */

	if (ndp->circuit && itf.circuit_close)
		itf.circuit_close(ndp);
	ndp->circuit = NULL;

	/* free any link-layer info */

	if (ndp->llinfo)
		free(ndp->llinfo);
	ndp->llinfo = NULL;

	/* free if allocated */

	if ((ndp->flags & ND6F_IS_SELF) == 0) {
		free(ndp);
		return;
	}

	/* cleanup everything */

	bzero(&ndp->sdl, sizeof(struct sockaddr_dl));
	ndp->flags = ND6F_IS_SELF;
	ndp->state = ndp->pstate = ND6S_INCOMPLETE;
	ndp->asked = 0;
}

/*
 * A good neighbor advertisement has been received
 */
void
update(ndp, lp, is_router, is_solicited, is_overwrite)
	struct neighbor *ndp;
	struct nd_opt_hdr *lp;
	int is_router, is_solicited, is_overwrite;
{
	struct nd_opt_hdr *old = ndp->llinfo;
	int staled = 0, same = 1;

	/*
	 * link-layer stuff is not the same if and only if
	 * there is a different link-layer info
	 */

	if (lp && itf.lladdr_cmp(old, lp))
		same = 0;

	/* compute action and next state */

	switch (ndp->state) {
	    case ND6S_REJECT:
		/* come back to genuine INCOMPLETE */
		if (ndp->flags & ND6F_HAS_ROUTE)
			(void) del_route(ndp);
		ndp->state = ND6S_INCOMPLETE;
		ndp->asked = 0;
		is_solicited = is_overwrite = 0;
		/* fall into */

	    case ND6S_INCOMPLETE:
		/* update link-layer stuff (see lladdr_overwrite()) */
		if ((lp != NULL) || (itf.llinfo == NULL))
			(void) lladdr_overwrite(ndp, lp,
						"Nd Adv (incomplete)");
		else {
			if (verbose)
				logtrace("unusable Nd Adv for %s",
					 inet6pr(&ndp->addr));
			return;
		}
		if (is_solicited)
			enter_state(ndp, ND6S_REACHABLE);
		else
			enter_state(ndp, ND6S_STALE);
		return;

	    case ND6S_REACHABLE:
		/* update link-layer stuff if needed */
		if (is_overwrite && !same)
			(void) lladdr_overwrite(ndp, lp,
						"Nd Adv (reachable)");
		/* is it solicited? */
		if (is_solicited) {
			if (same || is_overwrite)
				enter_state(ndp, ND6S_REACHABLE);
			else
				enter_state(ndp, ND6S_STALE);
			return;
		}
		/* not solicited */
	    dumb:
		if (!is_overwrite || same) {
			if (verbose)
				logtrace("dumb Nd Adv for %s",
					 inet6pr(&ndp->addr));
		} else if (ndp->state == ND6S_PENDING)
			enter_pending(ndp, ND6S_STALE);
		else
			enter_state(ndp, ND6S_STALE);
		return;

	    case ND6S_PENDING:
		/* sanity check */
		if ((ndp->pstate != ND6S_STALE) &&
		    (ndp->pstate != ND6S_REACHABLE))
			logerr("Nd Adv for %s in illegal pending state %d",
			       inet6pr(&ndp->addr), ndp->pstate);
	    case ND6S_STALE:
		staled = 1;
	    case ND6S_DELAY:
	    case ND6S_PROBE:
		/* update link-layer stuff if needed */
		if (is_overwrite && !same)
			(void) lladdr_overwrite(ndp, lp,
						"Nd Adv (unreachable)");
		/* is it solicited? */
		if (is_solicited) {
			if (is_overwrite || same) {
				if (staled)
					enter_pending(ndp, ND6S_REACHABLE);
				else
					enter_state(ndp, ND6S_REACHABLE);
			} else {
				if (verbose)
					logtrace("ignored Nd Adv for %s",
						 inet6pr(&ndp->addr));
			}	
			return;
		}
		/* not solicited */
		goto dumb;

	    case ND6S_LOCKED:
		/* is it same link-layer stuff? */
		if (same)
			return;
		/* not same */
		if (verbose)
			logtrace("Nd Adv for locked %s", inet6pr(&ndp->addr));
		return;

	    default:
		/* can't happen */
		logerr("Nd Adv for %s in impossible state %d",
		       inet6pr(&ndp->addr), ndp->state);
	}
}
