/* Neighbor discovery ICMPv6 message reception */

#include <ndpd.h>

char ipacket[MAXPACKET];

/*
 * Receive an ICMPv6/Neighbor Discovery packet
 */
void
recv_icmp()
{
	register struct ip6_hdr *ip;
	register struct icmp6_hdr *icp;
	struct msghdr msg;
	struct iovec iov;
	struct in6_pktinfo pi;
	int cc;

	/* prepare message stuff */

	bzero((caddr_t)&msg, sizeof(struct msghdr));
	bzero((caddr_t)&iov, sizeof(struct iovec));
	msg.msg_iov = &iov;
	msg.msg_iov->iov_base = ipacket;
	msg.msg_iov->iov_len = MAXPACKET;
	msg.msg_iovlen = 1;
	msg.msg_control = (caddr_t)&c_cmsg;
	msg.msg_controllen = sizeof(c_cmsg);

	/* receive message and get packet info */

	if ((cc = recvmsg(icmp_sd, &msg, 0)) < 0) {
		if (verbose)
			logperror("read (ICMPv6)");
		return;
	}
	if (cc == 0) {
		if (verbose)
			logerr("read ICMPv6 return 0");
		return;
	}
	if (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)) {
		if (verbose)
			logerr("read ICMPv6 truncated");
		return;
	}
	bcopy(c_cmsg.c_space, &pi, sizeof(pi));
	if ((msg.msg_controllen == 0) ||
	    (cm.cmsg_len != (sizeof(struct cmsghdr) + sizeof(pi))) ||
	    (cm.cmsg_level != IPPROTO_IPV6) ||
	    (cm.cmsg_type != IPV6_PKTINFO) ||
	    (pi.ipi6_ifindex == 0)) {
		if (verbose)
			logerr("read ICMPv6 doesn't give interface");
		return;
	}

	/* is it for us? */

	if (pi.ipi6_ifindex != itf.index)
		return;

	/* is it a good packet? */

	ip = (struct ip6_hdr *)ipacket;
	if (cc < sizeof(*ip) + sizeof(*icp)) {
		if (verbose)
			logerr("ICMPv6: too short (%d)", cc);
		return;
	}
	if (ip->ip6_hlim != ND_DEFAULT_HOPLIM)
		return;

	/* decode it */

	cc -= sizeof(struct ip6_hdr);
	icp = (struct icmp6_hdr *)(ip + 1);

	switch (icp->icmp6_type) {
	    case ND_ROUTER_SOLICIT:
		recv_rtsol(ip, cc);
		return;
	    case ND_ROUTER_ADVERT:
		recv_rtadv(ip, cc);
		return;
	    case ND_NEIGHBOR_SOLICIT:
		recv_ndsol(ip, cc);
		return;
	    case ND_NEIGHBOR_ADVERT:
		recv_ndadv(ip, cc);
		return;
	    case ND_REDIRECT:
		recv_redirect(ip, cc);
		return;
	    default:
		if (verbose)
			logerr("ICMPv6: unknown type %d", icp->icmp6_type);
	}
}

/*
 * Received a router solicitation
 */
void
recv_rtsol(ip, len)
	struct ip6_hdr *ip;
	int len;
{
	struct nd_router_solicit *rsp;
	struct nd_opt_hdr *op, *lp = NULL;
	struct in6_addr *src, *dst;
	struct neighbor *ndp;

	/* get interesting fields */

	src = &ip->ip6_src;
	dst = &ip->ip6_dst;
	rsp = (struct nd_router_solicit *)(ip + 1);
	op = (struct nd_opt_hdr *)(rsp + 1);
	len -= sizeof(*rsp);

	if (verbose)
		logtrace("Rt Sol from %s to %s len %d",
			 inet6pr(src), inet6pr(dst), len);

	/* is it a valid router solicitation? */

	if (rsp->nd_rs_code != 0) {
		logerr("Rt Sol with code=%d from %s",
		       rsp->nd_rs_code, inet6pr(src));
		return;
	}
	if (IN6_IS_ADDR_MULTICAST(dst) &&
	    (IN6_GET_SCOPE(dst) != MADDR6_SCP_LINK))
		logtrace("Rt Sol from %s for illegal multicast %s len %d",
			 inet6pr(src), inet6pr(dst), len);

	/* decode options: source link-layer address */

	while (len) switch (op->nd_opt_type) {
	    case ND_OPT_SOURCE_LINKADDR:
		if (lp != NULL) {
			logerr("Rt Sol from %s has duplicated src lladdr",
			       inet6pr(src));
			return;
		}
		lp = itf.lladdr_get(op);
		if (lp == NULL)
			return;
		goto next;

	    default:
		if (verbose)
			logtrace("unknown option %d", op->nd_opt_type);

	    next:
		/* go to next option */
		if (op->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       op->nd_opt_type);
			return;
		}
		len -= op->nd_opt_len << 3;
		if (len < 0) {
			if (verbose)
				logerr("Rt Sol underflow");
			return;
		}
		op = (struct nd_opt_hdr *)
			((caddr_t)op + (op->nd_opt_len << 3));
	}

	if (verbose && (lp != NULL))
		logtrace("Rt Sol lladdr=%s", itf.lladdr_pr(lp));

	/* special case for router solicit from unspecified source */

	if (IN6_IS_ADDR_UNSPECIFIED(src)) {
		if (lp != NULL)
			logerr("Rt Sol from :: with lladdr=%s",
			       itf.lladdr_pr(lp));
		return;
	}

	/* get the neighbor entry */

	ndp = ndplookup(src, lp != NULL);
	if (ndp == NULL)
		return;

	/* routers must not send router solicitations */

	if (ndp->flags & ND6F_IS_ROUTER) {
		logtrace("router %s sends a Rt Sol", inet6pr(src));
		ndp->flags &= ~ND6F_IS_ROUTER;
	}
	if (lp == NULL)
		return;

	/* update if needed the link-layer info (unsolicited) */

	if (lladdr_overwrite(ndp, lp, "Rt Sol")) {
		if (ndp->flags & ND6F_HAS_ROUTE)
			enter_state(ndp, ND6S_STALE);
		else
			enter_pending(ndp, ND6S_STALE);
	}
}

/*
 * Received a router advertisement
 */
void
recv_rtadv(ip, len)
	struct ip6_hdr *ip;
	int len;
{
	struct nd_router_advert *rap;
	struct nd_opt_hdr *op, *lp = NULL;
	struct in6_addr *src, *dst;
	struct neighbor *ndp;
	u_int val;

	/* get interesting fields */

	src = &ip->ip6_src;
	dst = &ip->ip6_dst;
	rap = (struct nd_router_advert *)(ip + 1);
	op = (struct nd_opt_hdr *)(rap + 1);
	len -= sizeof(*rap);

	if (verbose)
		logtrace("Rt Adv from %s to %s len %d",
			 inet6pr(src), inet6pr(dst), len);

	/* is it a valid router advertisement? */

	if (rap->nd_ra_code != 0) {
		logerr("Rt Adv with code=%d from %s",
		       rap->nd_ra_code, inet6pr(src));
		return;
	}
	if (!IN6_IS_ADDR_LINKLOCAL(src)) {
		logerr("Rt Adv from %s not link-local", inet6pr(src));
		return;
	}
	if (len < 0) {
		logerr("Rt Adv from %s too short", inet6pr(src));
		return;
	}
	if (IN6_IS_ADDR_MULTICAST(dst) &&
	    (IN6_GET_SCOPE(dst) != MADDR6_SCP_LINK))
		logtrace("Rt Adv from %s for illegal multicast %s len %d",
			 inet6pr(src), inet6pr(dst), len);

	/* extract parameters */

	val = ntohl(rap->nd_ra_reachable);
	if ((val != 0) && (val != BaseReachableTime))
		setReachableTime(val);
	val = ntohl(rap->nd_ra_retransmit);
	if ((val != 0) && (val != RetransTimer))
		setRetransTimer(val);

	/*
	 * decode options:
	 *     source link-layer address
	 *     prefix informations
	 *     mtu
	 */

	while (len) switch (op->nd_opt_type) {
	    case ND_OPT_SOURCE_LINKADDR:
		if (lp != NULL) {
			logerr("Rt Adv from %s has duplicated src lladdr",
			       inet6pr(src));
			return;
		}
		lp = itf.lladdr_get(op);
		if (lp == NULL)
			return;
		goto next;

	    default:
		if (verbose)
			logtrace("unknown option %d", op->nd_opt_type);


	    case ND_OPT_PREFIX_INFORMATION:
	    case ND_OPT_MTU:
	    next:
		/* go to next option */
		if (op->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       op->nd_opt_type);
			return;
		}
		len -= op->nd_opt_len << 3;
		if (len < 0) {
			if (verbose)
				logerr("Rt Adv underflow");
			return;
		}
		op = (struct nd_opt_hdr *)
			((caddr_t)op + (op->nd_opt_len << 3));
	}

	if (verbose && (lp != NULL))
		logtrace("Rt Adv lladdr=%s", itf.lladdr_pr(lp));

	/* get the neighbor entry */

	ndp = ndplookup(src, 1);
	if (ndp == NULL)
		return;

	/* router advertisements are only sent by routers */

	ndp->flags |= ND6F_IS_ROUTER;
	if (lp == NULL)
		return;

	/* update if needed the link-layer info (unsolicited) */

	if (lladdr_overwrite(ndp, lp, "Rt Adv")) {
		if (ndp->flags & ND6F_HAS_ROUTE)
			enter_state(ndp, ND6S_STALE);
		else
			enter_pending(ndp, ND6S_STALE);
	}
}

/*
 * Received a neighbor solicitation
 */
void
recv_ndsol(ip, len)
	struct ip6_hdr *ip;
	int len;
{
	struct nd_neighbor_solicit *nsp;
	struct nd_opt_hdr *op, *lp = NULL;
	struct in6_addr *src, *dst, *tgt;
	struct prefix *pp, *spp, *tpp;
	struct neighbor *ndp;
	int addrconfsol = 0;

	/* get interesting fields */

	src = &ip->ip6_src;
	if (IN6_IS_ADDR_UNSPECIFIED(src))
		addrconfsol = 1;
	dst = &ip->ip6_dst;
	nsp = (struct nd_neighbor_solicit *)(ip + 1);
	tgt = &nsp->nd_ns_target;
	op = (struct nd_opt_hdr *)(nsp + 1);
	len -= sizeof(*nsp);

	if (verbose)
		logtrace("Nd Sol from %s to %s for %s len %d",
			 inet6pr(src), inet6pr(dst), inet6pr(tgt), len);

	/* is it a valid neighbor solicitation? */

	if (nsp->nd_ns_code != 0) {
		logerr("Nd Sol with code=%d from %s for %s",
		       nsp->nd_ns_code, inet6pr(src), inet6pr(tgt));
		return;
	}
	if (len < 0) {
		logerr("Nd Sol from %s too short", inet6pr(src));
		return;
	}
	if (IN6_IS_ADDR_MULTICAST(tgt)) {
		logerr("Nd Sol from %s for multicast %s",
		       inet6pr(src), inet6pr(tgt));
		return;
	}
	if (addrconfsol && !IN6_IS_ADDR_SOLICITED(dst)) {
		logerr("Nd Sol from :: to not solicited %s for %s",
		       inet6pr(dst), inet6pr(tgt));
		return;
	}

	/* decode options: source link-layer address */

	while (len) switch (op->nd_opt_type) {
	    case ND_OPT_SOURCE_LINKADDR:
		if (lp != NULL) {
			logerr("Nd Sol from %s has duplicated src lladdr",
			       inet6pr(src));
			return;
		}
		lp = itf.lladdr_get(op);
		if (lp == NULL)
			return;
		goto next;

	    default:
		if (verbose)
			logtrace("unknown option %d", op->nd_opt_type);
	    next:
		/* go to next option */
		if (op->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       op->nd_opt_type);
			return;
		}
		len -= op->nd_opt_len << 3;
		if (len < 0) {
			if (verbose)
				logerr("Nd Sol underflow");
			return;
		}
		op = (struct nd_opt_hdr *)
			((caddr_t)op + (op->nd_opt_len << 3));
	}

	/* address configuration is without link-layer info */

	if ((lp != NULL) && (addrconfsol == 1)) {
		logerr("Nd Sol from :: for %s with lladdr=%s",
		       inet6pr(tgt), itf.lladdr_pr(lp));
		return;
	}

	/* in other cases there should be a link-layer info */

	if ((lp == NULL) && (addrconfsol == 0) &&
	    (itf.llinfo != NULL) && IN6_IS_ADDR_MULTICAST(dst)) {
		logerr("Nd Sol from %s to %s for %s without lladdr",
		       inet6pr(src), inet6pr(dst), inet6pr(tgt));
		return;
	}

	/* get source and target prefixes */

	spp = tpp = NULL;
	FORALLPP(pp) {
		if (IN6_ARE_ADDR_EQUAL(src, &pp->addr))
			spp = pp;
		if (IN6_ARE_ADDR_EQUAL(tgt, &pp->addr))
			tpp = pp;
	}

	/* anycasts, proxies, ...? */

	if ((spp == NULL) && (tpp == NULL))
		return;

	if (verbose && (lp != NULL))
		logtrace("Nd Sol lladdr=%s", itf.lladdr_pr(lp));

	/* is the solicitation from us? */

	if (spp != NULL) {
		if (addrconfsol) {
			logerr("Got prefix 0x%x for ::", (u_long)spp);
			return;
		}
		if (lp == NULL)
			return;
		/* from me (then ignore it)? */
		if (itf.lladdr_cmp(lp, itf.llinfo) == 0)
			return;
		logerr("Nd Sol from my addr %s with lladdr=%s",
		       inet6pr(src), itf.lladdr_pr(lp));
		return;
	}

	/* from here spp == NULL and tpp != NULL */

	/* deal with duplicate address detection */

	if (addrconfsol) {
#ifdef notyet
		if (tpp->flags & BOOTING)
			return;
#endif
		if (is_router() && is_apaddr(tgt))
			return;
		logerr("duplicate IPv6 addr %s", inet6pr(tgt));
		send_ndadv(&tpp->self, tpp, 0);
		return;
	}

	/* get the neighbor entry */

	ndp = ndplookup(src, 1);
	if (ndp == NULL)
		return;

	/* update if needed the link-layer info (unsolicited) */

	if ((lp != NULL) && lladdr_overwrite(ndp, lp, "Nd Sol")) {
		if (ndp->flags & ND6F_HAS_ROUTE)
			enter_state(ndp, ND6S_STALE);
		else
			enter_pending(ndp, ND6S_STALE);
	}
#ifdef notyet
	if (tpp->flags & BOOTING)
		return;
#endif

	/* anycasts, proxies, ..., are done elsewhere */

	if (is_router() && is_apaddr(tgt))
		return;

	/* send the answer */

	send_ndadv(ndp, tpp, 1);
}

/*
 * Received a neighbor advertisement
 */
void
recv_ndadv(ip, len)
	struct ip6_hdr *ip;
	int len;
{
	struct nd_neighbor_advert *nap;
	struct nd_opt_hdr *op, *lp = NULL;
	struct in6_addr *src, *dst, *tgt;
	struct neighbor *ndp;
	int is_router, is_solicited, is_overwrite;

	/* get interesting fields */

	src = &ip->ip6_src;
	dst = &ip->ip6_dst;
	nap = (struct nd_neighbor_advert *)(ip + 1);
	tgt = &nap->nd_na_target;
	is_router = (nap->nd_na_flags_reserved & ND_NA_FLAG_ROUTER) != 0;
	is_solicited = (nap->nd_na_flags_reserved & ND_NA_FLAG_SOLICITED) != 0;
	is_overwrite = (nap->nd_na_flags_reserved & ND_NA_FLAG_OVERRIDE) != 0;
	op = (struct nd_opt_hdr *)(nap + 1);
	len -= sizeof(*nap);

	if (verbose)
		logtrace("Nd Adv from %s to %s for %s len %d flags '%s%s%s'",
			 inet6pr(src), inet6pr(dst), inet6pr(tgt), len,
			 is_router ? "R" : "",
			 is_solicited ? "S" : "",
			 is_overwrite ? "O" : "");

	/* is it a valid neighbor advertisement? */

	if (nap->nd_na_code != 0) {
		logerr("Nd Adv with code=%d from %s for %s",
		       nap->nd_na_code, inet6pr(src), inet6pr(tgt));
		return;
	}
	if (len < 0) {
		logerr("Nd Adv from %s too short", inet6pr(src));
		return;
	}
	if (IN6_IS_ADDR_MULTICAST(tgt)) {
		logerr("Nd Adv from %s for multicast %s",
		       inet6pr(src), inet6pr(tgt));
		return;
	}
	if (IN6_IS_ADDR_MULTICAST(dst)) {
		if (is_solicited) {
			logerr("solicited Nd Adv from %s to %s for %s",
			       inet6pr(src), inet6pr(dst), inet6pr(tgt));
			return;
		}
		if (IN6_GET_SCOPE(dst) != MADDR6_SCP_LINK)
			logtrace("Nd Adv from %s to illegal multicast %s",
				 inet6pr(src), inet6pr(dst));
	}

	/* decode options: target link-layer address */

	while (len) switch (op->nd_opt_type) {
	    case ND_OPT_TARGET_LINKADDR:
		if (lp != NULL) {
			logerr("Nd Adv from %s has duplicated src lladdr",
			       inet6pr(src));
			return;
		}
		lp = itf.lladdr_get(op);
		if (lp == NULL)
			return;
		goto next;

	    default:
		if (verbose)
			logtrace("unknown option %d", op->nd_opt_type);
	    next:
		/* go to next option */
		if (op->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       op->nd_opt_type);
			return;
		}
		len -= op->nd_opt_len << 3;
		if (len < 0) {
			if (verbose)
				logerr("Nd Adv underflow");
			return;
		}
		op = (struct nd_opt_hdr *)
			((caddr_t)op + (op->nd_opt_len << 3));
	}

	/* multicast advertisement must have link-layer info */

	if ((lp == NULL) && (itf.llinfo != NULL) &&
	    IN6_IS_ADDR_MULTICAST(dst)) {
		logerr("Nd Adv from %s to %s for %s without lladdr",
		       inet6pr(src), inet6pr(dst), inet6pr(tgt));
		return;
	}

	/* check for a duplicate address */

	if (lp != NULL) {
		struct prefix *pp;

		FORALLPP(pp)
			if (IN6_ARE_ADDR_EQUAL(tgt, &pp->addr))
				break;

		if (pp == NULL)
			goto notme;
		/* from me (then ignore it)? */
		if (itf.lladdr_cmp(lp, itf.llinfo) == 0)
			return;
		logerr("Nd Adv for my addr %s with lladdr=%s",
		       inet6pr(tgt), itf.lladdr_pr(lp));
		return;
	}
    notme:

	/* get the neighbor entry */

	ndp = ndplookup(tgt, 0);
	if (ndp == NULL)
		return;

	/* update the neighbor entry */

	update(ndp, lp, is_router, is_solicited, is_overwrite);
}

/*
 * Received a redirect
 */
void
recv_redirect(ip, len)
	struct ip6_hdr *ip;
	int len;
{
	struct nd_redirect *rdp;
	struct nd_opt_hdr *op, *lp = NULL;
	struct in6_addr *src, *dst, *tgt, *rdst;
	struct neighbor *ndp;
	int for_router;

	/* get interesting fields */

	src = &ip->ip6_src;
	dst = &ip->ip6_dst;
	rdp = (struct nd_redirect *)(ip + 1);
	tgt = &rdp->nd_rd_target;
	rdst = &rdp->nd_rd_dst;
	for_router = IN6_ARE_ADDR_EQUAL(tgt, rdst) == 0;
	op = (struct nd_opt_hdr *)(rdp + 1);
	len -= sizeof(*rdp);

	if (verbose)
		logtrace("Rd from %s to %s for %s about %s len %d",
			 inet6pr(src), inet6pr(dst),
			 inet6pr(tgt), inet6pr(rdst), len);

	/* is it a valid redirect? */

	if (rdp->nd_rd_code != 0) {
		logerr("Rd with code=%d from %s",
		       rdp->nd_rd_code, inet6pr(src));
		return;
	}
	if (len < 0) {
		logerr("Rd from %s too short", inet6pr(src));
		return;
	}
	if (!IN6_IS_ADDR_LINKLOCAL(src)) {
		logerr("Rd from illegal src %s", inet6pr(src));
		return;
	}
	if (IN6_IS_ADDR_MULTICAST(dst)) {
		logerr("Rd from %s to multicast %s",
		       inet6pr(src), inet6pr(dst));
		return;
	}
	if (for_router && !IN6_IS_ADDR_LINKLOCAL(tgt)) {
		logerr("router Rd from %s for illegal %s",
		       inet6pr(src), inet6pr(tgt));
		return;
	} else if (IN6_IS_ADDR_MULTICAST(tgt)) {
		logerr("Rd from %s for multicast %s",
		       inet6pr(src), inet6pr(tgt));
		return;
	}
	if (IN6_IS_ADDR_MULTICAST(rdst)) {
		logerr("Rd from %s about multicast %s",
		       inet6pr(src), inet6pr(rdst));
		return;
	}

	/*
	 * decode options:
	 *     target link-layer address
	 *     redirected header
	 */

	while (len) switch (op->nd_opt_type) {
	    case ND_OPT_TARGET_LINKADDR:
		if (lp != NULL) {
			logerr("Rd from %s has duplicated src lladdr",
			       inet6pr(src));
			return;
		}
		lp = itf.lladdr_get(op);
		if (lp == NULL)
			return;
		goto next;

	    default:
		if (verbose)
			logtrace("unknown option %d", op->nd_opt_type);

	    case ND_OPT_REDIRECTED_HEADER:
	    next:
		/* go to next option */
		if (op->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       op->nd_opt_type);
			return;
		}
		len -= op->nd_opt_len << 3;
		if (len < 0) {
			if (verbose)
				logerr("Rd underflow");
			return;
		}
		op = (struct nd_opt_hdr *)
			((caddr_t)op + (op->nd_opt_len << 3));
	}

	/* get the neighbor entry */

	ndp = ndplookup(tgt, 1);
	if (ndp == NULL)
		return;

	/* a router is changed for a router */

	if (for_router)
		ndp->flags |= ND6F_IS_ROUTER;
	if (lp == NULL)
		return;

	/* update if needed the link-layer info (unsolicited) */
	/* note that external redirect will be stuck in the pending state */
	/* ZZZZZZZZZZZZZ => TODO */

	if (lladdr_overwrite(ndp, lp, "Rd")) {
		if (ndp->flags & ND6F_HAS_ROUTE)
			enter_state(ndp, ND6S_STALE);
		else
			enter_pending(ndp, ND6S_STALE);
	}
}

/*
 * Lookup or enter a new neighbor in the ndp table
 */
struct neighbor *
ndplookup(addr, create)
	struct in6_addr *addr;
	int create;
{
	struct neighbor *ndp;

	if (verbose)
		logtrace("lookup/%d for %s", create, inet6pr(addr));

	FORALLND(ndp)
		if (IN6_ARE_ADDR_EQUAL(addr, &ndp->addr))
			return ndp;

	/* not found */

	if (!create)
		return NULL;
	if (Calloc(ndp, neighbor) == NULL) {
		if (verbose)
			logerr("out of memory");
		return NULL;
	}
	bcopy(addr, &ndp->addr, sizeof(*addr));
	TAILQ_INSERT_HEAD(&neighbors, ndp, ngb_list);
	return ndp;
}

/*
 * Overwrite link-layer information
 */
int
lladdr_overwrite(ndp, lp, name)
	struct neighbor *ndp;
	struct nd_opt_hdr *lp;
	char *name;
{
	struct nd_opt_hdr *old = ndp->llinfo;
	int size = lp->nd_opt_len << 3;

	/* special case for link without addresses ??? */

	if (lp == NULL)
		goto updated;

	/* if equal there is nothing to do */

	if (itf.lladdr_cmp(old, lp) == 0)
		return 0;

	/* don't overwrite locked entries */

	if (ndp->state == ND6S_LOCKED) {
		if (verbose)
			logtrace("%s tried to overwrite locked entry for %s",
				 name, inet6pr(&ndp->addr));
		return 0;
	}

	/* update link-layer info */

	if (old)
		free(old);
	old = (struct nd_opt_hdr *)calloc(1, size);
	if (old == NULL) {
		logerr("out of memory");
		abort();
	}
	bcopy(lp, old, size);
	ndp->llinfo = old;

    updated:

	/* update is done */

	if (verbose)
		logtrace("%s overwrite %s lladdr", name, inet6pr(&ndp->addr));

	/* do link-layer dependent stuff */

	itf.lladdr_updated(ndp);

	/* update the route if possible */

	if ((ndp->flags & ND6F_HAS_LLADDR) && (ndp->flags & ND6F_HAS_ROUTE))
		change_route(ndp);

	/* return success */

	return 1;
}
