/* Neighbor discovery kernel stuff and route message reception */

#include <ndpd.h>

static int mib[] = { CTL_NET, PF_ROUTE, 0, AF_INET6, NET_RT_IFLIST, 1 };
int rtm_seq = 0;
TIMEOUT *randpar = NULL;

/* timers */

u_int BaseReachableTime;
u_int ReachableTime;
u_int InactiveTime;
u_int HoldDownTime;
u_int RetransTimer;
u_int DelayTimer;

/*
 * Get interface information and addresses
 */
int
init(name)
	char *name;
{
	struct if_msghdr *ifm = NULL;
	struct prefix *pp;
	struct sockaddr_in6 *sin;
	int i;
	char *buf, *cp, *ap;
	unsigned bufsize;

	/* global inits */

	TAILQ_INIT(&neighbors);
	TAILQ_INIT(&prefixes);
	bzero(&itf, sizeof(itf));

	/* get the interface index from the given name */

	itf.index = if_nametoindex(name);

	if (itf.index <= 0)
		return 0;
	mib[5] = itf.index;

	/* get interface informations (ifinfo) */

	if (sysctl(mib, 6, NULL, &bufsize, NULL, 0) < 0) {
		logperror("sysctl0");
		exit(4);
	}
	buf = (char *)malloc(bufsize);
	if (buf == NULL) {
		logerr("out of memory");
		exit(4);
	}
	if (sysctl(mib, 6, buf, &bufsize, NULL, 0) < 0) {
		logperror("sysctl");
		exit(4);
	}
	ifm = (struct if_msghdr *)buf;

	/* sanity check */

	if ((ifm->ifm_type != RTM_IFINFO) || (ifm->ifm_index != itf.index)) {
		logerr("can't get ifinfo for interface %s", name);
		exit(4);
	}

	/* decode ifinfo and interface addresses */

	for (cp = buf; cp < buf + bufsize; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;

		/* first case: ifinfo */

		if (ifm->ifm_type == RTM_IFINFO) {
			if (cp != buf) {
				logerr("got extra ifinfo ?");
				exit(4);
			}
			/* decode ifinfo / fill itf structure */
			if (!NEW_IF(ifm, 1))
				return 0;
			continue;
		}

		/* second case: newaddr */

		if (ifm->ifm_type != RTM_NEWADDR) {
			if (verbose)
				logerr("bad message type (%d)",
				       ifm->ifm_type);
			continue;
		}

		/* sanity check (there must be an interface address) */

		if ((ifm->ifm_addrs & RTA_IFA) == 0) {
			if (verbose)
				logerr("no ifa for %s", name);
			continue;
		}

		/* sanity check (point-to-point needs a destination address) */
		if ((ifm->ifm_flags & IFF_POINTOPOINT) &&
		    ((ifm->ifm_addrs & RTA_BRD) == 0)) {
			if (verbose)
				logerr("no dest for p2p %s", name);
			continue;
		}
		if (verbose)
			logtrace("get NEWADDR");

		/* allocate a prefix structure */

		if (Calloc(pp, prefix) == NULL) {
			if (verbose)
				logerr("out of memory");
			continue;
		}

		/*
		 * decode parameters:
		 *     netmask (prefix length)
		 *     interface address
		 *     broadcast address (destination address)
		 */

		ap = cp + sizeof(struct ifa_msghdr);
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifm->ifm_addrs & (1 << i)) == 0)
				continue;
			sin = (struct sockaddr_in6 *)ap;
			ADVANCE(ap, sin);
			switch (i) {
			    case RTAX_NETMASK:
				/* netmask */
				if (sin->sin6_family != AF_UNSPEC)
					goto skip;
				pp->plen = get_prefixlen(sin);
				bcopy(&sin->sin6_addr, &pp->amask,
				      sizeof(struct in6_addr));
				break;

			    case RTAX_IFA:
				/* interface address */
				if (sin->sin6_family != AF_INET6)
					goto skip;
				bcopy(&sin->sin6_addr, &pp->addr,
				      sizeof(struct in6_addr));
				break;

			    case RTAX_BRD:
				/* broadcast address */
				if ((ifm->ifm_flags & IFF_POINTOPOINT) == 0)
					break;
				if (sin->sin6_family != AF_INET6)
					goto skip;
				bcopy(&sin->sin6_addr, &pp->dest,
				      sizeof(struct in6_addr));
				break;
			}
		}

		/* sanity checks */

		if ((pp->plen < 3) || IN6_IS_ADDR_UNSPECIFIED(&pp->addr))
			goto skip;
		if ((ifm->ifm_flags & IFF_POINTOPOINT) &&
		    IN6_IS_ADDR_UNSPECIFIED(&pp->dest))
			goto skip;

		if (verbose)
			logtrace("interface %s got addr %s plen %d addrs %x",
				 name, inet6pr(&pp->addr),
				 pp->plen, ifm->ifm_addrs);
		if (verbose && (ifm->ifm_flags & IFF_POINTOPOINT))
			logtrace("interface %s got dest %s",
				 inet6pr(&pp->dest));

		/* keep first link-local address */

		if (IN6_IS_ADDR_LINKLOCAL(&pp->addr) &&
		    (IN6_IS_ADDR_UNSPECIFIED(&itf.llocal)))
			bcopy(&pp->addr, &itf.llocal, sizeof(struct in6_addr));

		/*
		 * ignore strange prefixes:
		 *     reserved first byte (0 & 0xff)
		 *     very short (< 16 bits)
		 *     too long (inside interface ID)
		 */

		if ((pp->addr.s6_addr[0] == 0) ||
		    (pp->addr.s6_addr[0] == 0xff) ||
		    (pp->plen < 16) ||
		    (pp->plen != 128 &&
		     pp->plen > (128 - 8 * itf.ltoken))) {
			if (verbose)
				logerr("ignored interface address %s",
				       inet6pr(&pp->addr));
			goto skip;
		}

		/* build and apply masks */

		bcopy(&pp->addr, &pp->pref, sizeof(struct in6_addr));
		mk_pref(pp);
		if (pp->plen != 128)
			bcopy(&pp->pmask, &pp->amask, sizeof(struct in6_addr));
		else
			mk_mask(&pp->amask, 128);

		/* build the local neighbor entry for this prefix */

		mk_self(pp);

		/* link all this stuff */

		TAILQ_INSERT_TAIL(&prefixes, pp, prf_list);
		TAILQ_INSERT_TAIL(&neighbors, &pp->self, ngb_list);

		/* get next prefix */

		continue;
    skip:
		/* failed/ignored */

		free(pp);
	}

	/* done: release buffer */

	free(buf);

	/* last check: there must be a link-local address */

	if (IN6_IS_ADDR_UNSPECIFIED(&itf.llocal)) {
		logerr("ignoring interface %s without LL", name);
		return 0;
	}
	return 1;
}

/*
 * Initialize parameters
 */
void
init_param()
{
	BaseReachableTime = REACHABLE_TIME;
	setReachableTime(0);
	setRetransTimer(RETRANS_TIMER);
}

/*
 * Recompute Time parameters (Reachable)
 */
void
setReachableTime(basereach)
	u_int basereach;
{
	/* something to do ? */

	if (basereach == BaseReachableTime)
		return;
	if (basereach != 0)
		BaseReachableTime = basereach;

	/* reset timeout */

	if (randpar)
		KILL_TIMEOUT(randpar);
	randpar = NULL;

	/* randomize */

	ReachableTime = RANDOMIZE(BaseReachableTime);

	/* check bounds */

	if (ReachableTime >= MAX_REACHABLE_TIME)
		ReachableTime = MAX_REACHABLE_TIME - 1;
	if (ReachableTime <= MIN_REACHABLE_TIME)
		ReachableTime = MIN_REACHABLE_TIME + 1;

	/* deduct other parameters */

	InactiveTime = INACTIVE_TIME;
	if (InactiveTime <= BaseReachableTime)
		InactiveTime = BaseReachableTime;
	HoldDownTime = BaseReachableTime / 10;
	if (HoldDownTime >= MAX_HOLD_DOWN_TIME)
		HoldDownTime = MAX_HOLD_DOWN_TIME - 1;

	/* call it again in one hours ? */

	if (randpar == NULL)
		randpar = NEW_TIMEOUT(MS_TO_SEC((u_long)MAX_REACHABLE_TIME),
				      0, setReachableTime, 0);
}

/*
 * Recompute Timer parameters (Retransmit)
 */
void
setRetransTimer(retrans)
	u_int retrans;
{
	RetransTimer = retrans;

	/* check bounds */

	if (RetransTimer >= BaseReachableTime / 5)
		RetransTimer = (BaseReachableTime / 5) - 1;
	if (RetransTimer <= MIN_RETRANS_TIMER)
		RetransTimer = MIN_RETRANS_TIMER + 1;

	/* deal with delay timer */

	DelayTimer = DELAY_FIRST_PROBE_TIME;
}

/*
 * Receive a route from route socket
 */
void
recv_route()
{
	struct ifa_msghdr *ifam;
	char *cp;
	int cc;

	/* read the route message */

	cp = (caddr_t)&m_rtmsg;
	if (debug)
		bzero(cp, MAX_MSG_ROUTE);
	cc = read(route_sd, cp, MAX_MSG_ROUTE);
	if (cc < 0) {
		if (verbose)
			logperror("read (route message)");
		return;
	} else if (cc == 0) {
		if (verbose)
			logerr("read (route message) return 0");
		return;
	}

	/* sanity check */

	if (rtm.rtm_version != RTM_VERSION) {
		if (verbose)
			logerr("bad message version %d", rtm.rtm_version);
		return;
	}

	if (verbose)
		logtrace("receive route (%d,%d)",
			 rtm.rtm_msglen, rtm.rtm_type);
	DUMPR("received");

	/*
	 * decode the route message:
	 *     interface address addition
	 *     interface address removal
	 *     resolution request
	 */

	switch (rtm.rtm_type) {
	    case RTM_NEWADDR:
		/* interface address addition */
		ifam = (struct ifa_msghdr *)cp;
		if (ifam->ifam_index != itf.index)
			return;
		if (verbose)
		    	logtrace("receive route newaddr");

		recv_newaddr(ifam);
		break;

	    case RTM_DELADDR:
		/* interface address removal */
		ifam = (struct ifa_msghdr *)cp;
		if (ifam->ifam_index != itf.index)
			return;
		if (verbose)
			logtrace("receive route deladdr");

		recv_deladdr(ifam);
		break;

	    case RTM_RESOLVE:
		/* resolution request */
		if (rtm.rtm_index != itf.index)
			return;
		if (verbose)
			logtrace("receive route resolve");
		recv_resolve(&rtm);
	}
}

/*
 * Receive an interface address addition route message
 * (allocate a new prefix)
 */
void
recv_newaddr(ifam)
	struct ifa_msghdr *ifam;
{
	struct sockaddr_in6 *sa, *sin = NULL, *dsin = NULL;
	struct prefix *pp;
	char *cp;
	int i, plen = 0;

	DUMPR("recv_newaddr");

	/*
	 * decode parameters:
	 *     netmask (prefix length)
	 *     interface address
	 *     broadcast address (destination address)
	 */

	cp = (caddr_t)ifam;
	cp += sizeof(struct ifa_msghdr);
	for (i = 0; i < RTAX_MAX; i++) {
		if ((ifam->ifam_addrs & (1 << i)) == 0)
			continue;
		sa = (struct sockaddr_in6 *)cp;
		ADVANCE(cp, sa);
		switch (i) {
		    case RTAX_NETMASK:
			/* netmask */
			if (sa->sin6_family != AF_UNSPEC)
				return;
			plen = get_prefixlen(sa);
			if (plen < 3)
				return;
			break;

		    case RTAX_IFA:
			/* interface address */
			if (sa->sin6_family != AF_INET6)
				return;
			sin = sa;
			break;

		    case RTAX_BRD:
			/* broadcast address */
			if ((itf.flags & IFF_POINTOPOINT) == 0)
				break;
			if (sa->sin6_family != AF_INET6)
				return;
			dsin = sa;
			break;
		}
	}

	/* sanity checks */

	if ((sin == NULL) ||
	    (sin->sin6_addr.s6_addr[0] == 0) ||
	    (sin->sin6_addr.s6_addr[0] == 0xff))
		return;
	if ((dsin != NULL) &&
	    ((dsin->sin6_addr.s6_addr[0] == 0) ||
	     (dsin->sin6_addr.s6_addr[0] == 0xff)))
		return;

	if (verbose)
		logtrace("get address %s/%d",
			 inet6pr(&sin->sin6_addr), plen);
	if (verbose && (dsin != NULL))
		logtrace("get destination %s", inet6pr(&dsin->sin6_addr));

	/* special case for link-local addresses */

	if (IN6_IS_ADDR_LINKLOCAL(&sin->sin6_addr) &&
	    bcmp(&sin->sin6_addr, &itf.llocal, sizeof(struct in6_addr))) {
		logerr("changing lladdr from %s to %s",
		       inet6pr(&itf.llocal), inet6pr(&sin->sin6_addr));
		bcopy(&sin->sin6_addr, &itf.llocal, sizeof(struct in6_addr));
	}

	/* try to find this prefix in the current list */

	FORALLPP(pp)
		if (IN6_ARE_ADDR_EQUAL(&sin->sin6_addr, &pp->addr) &&
		    (plen == pp->plen))
			break;

	/* is this prefix already known? */

	if (pp != NULL) {
		if (verbose)
			logtrace("reget? address %s/%d",
				 inet6pr(&sin->sin6_addr), plen);
		/* ZZZZZZZZ support p2p prefixes with same local address */
		if (verbose && (dsin != NULL) &&
		    !IN6_ARE_ADDR_EQUAL(&dsin->sin6_addr, &pp->dest))
			logtrace("reget with different dest %s != %s",
				 inet6pr(&pp->dest),
				 inet6pr(&dsin->sin6_addr));
		return;
	}

	/* reject prefixes with a strange length */

	if ((plen > 128 - 8 * itf.ltoken) || (plen < 16)) {
		logerr("bad length for prefix %s/%d",
		       inet6pr(&sin->sin6_addr), plen);
		return;
	}

	/* allocate the prefix structure */

	if (Calloc(pp, prefix) == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}

	/* fill the structure */

	pp->plen = plen;
	bcopy(&sin->sin6_addr, &pp->addr, sizeof(struct in6_addr));
	bcopy(&sin->sin6_addr, &pp->pref, sizeof(struct in6_addr));
	mk_pref(pp);
	if (dsin != NULL)
		bcopy(&dsin->sin6_addr, &pp->dest, sizeof(struct in6_addr));

	/* build the local neighbor entry for this prefix */

	mk_self(pp);

	/* link all this stuff */

	TAILQ_INSERT_TAIL(&prefixes, pp, prf_list);
	TAILQ_INSERT_TAIL(&neighbors, &pp->self, ngb_list);

	logtrace("add new address %s/%d", inet6pr(&sin->sin6_addr), plen);
	if (dsin != NULL)
		logtrace("with destination %s", inet6pr(&dsin->sin6_addr));
}

/*
 * Receive an interface address removal route message
 * (release a prefix)
 */
void
recv_deladdr(ifam)
	struct ifa_msghdr *ifam;
{
	struct sockaddr_in6 *sa, *sin = NULL;
	struct prefix *pp;
	char *cp;
	int i, plen = 0;

	DUMPR("recv_deladdr");

	/*
	 * decode parameters:
	 *     netmask (prefix length)
	 *     interface address
	 */

	cp = (caddr_t)ifam;
	cp += sizeof(struct ifa_msghdr);
	for (i = 0; i < RTAX_MAX; i++) {
		if ((ifam->ifam_addrs & (1 << i)) == 0)
			continue;
		sa = (struct sockaddr_in6 *)cp;
		ADVANCE(cp, sa);
		switch (i) {
		    case RTAX_NETMASK:
			/* netmask */
			if (sa->sin6_family != AF_UNSPEC)
				return;
			plen = get_prefixlen(sa);
			if (plen < 3)
				return;
			break;

		    case RTAX_IFA:
			/* interface address */
			if (sa->sin6_family != AF_INET6)
				return;
			sin = sa;
			break;
		}
	}

	/* sanity checks */

	if ((sin == NULL) ||
	    (sin->sin6_addr.s6_addr[0] == 0) ||
	    (sin->sin6_addr.s6_addr[0] == 0xff))
		return;

	if (verbose)
		logtrace("del address %s/%d",
			 inet6pr(&sin->sin6_addr), plen);

	/* try to find the prefix in the current list */

	FORALLPP(pp)
		if (IN6_ARE_ADDR_EQUAL(&sin->sin6_addr, &pp->addr) &&
		    (plen == pp->plen))
			break;

	/* is the prefix already known? */

	if (pp == NULL) {
		if (verbose)
			logtrace("redel? address %s/%d",
				 inet6pr(&sin->sin6_addr), plen);
		return;
	}

	/* unlink (should we remove some routes?) */

	TAILQ_REMOVE(&neighbors, &pp->self, ngb_list);
	TAILQ_REMOVE(&prefixes, pp, prf_list);

	/* release the prefix structure */

	free(pp);

	logtrace("del old address %s/%d", inet6pr(&sin->sin6_addr), plen);

	/* special case for link-local prefixes */

	if (IN6_IS_ADDR_LINKLOCAL(&sin->sin6_addr)) {
		/* is this prefix the link-local one? */
		if (bcmp(&sin->sin6_addr, &itf.llocal,
			 sizeof(struct in6_addr)) != 0)
			return;
		/* try to find another link-local prefix */
		FORALLPP(pp)
			if (IN6_IS_ADDR_LINKLOCAL(&pp->addr))
				break;
		/* no other link-local prefix? */
		if (pp == NULL) {
			logerr("lost last link-local address %s",
			       inet6pr(&itf.llocal));
			return;
		}
		/* change the link-local prefix */
		logerr("changing lladdr from %s to %s",
		       inet6pr(&itf.llocal), inet6pr(&pp->addr));
		bcopy(&pp->addr, &itf.llocal, sizeof(struct in6_addr));
	}
}

#undef rtm

/*
 * Receive a resolution request route message
 */
void
recv_resolve(rtm)
	struct rt_msghdr *rtm;
{
	struct sockaddr_in6 *sa, *sin = NULL;
	struct neighbor *ndp;
	char *cp;
	int i;

	DUMPR("recv_resolve");

	/* decode parameters: destination (neighbor address) */

	cp = (caddr_t)rtm;
	cp += sizeof(struct rt_msghdr);
	for (i = 0; i < RTAX_MAX; i++) {
		if ((rtm->rtm_addrs & (1 << i)) == 0)
			continue;
		sa = (struct sockaddr_in6 *)cp;
		ADVANCE(cp, sa);
		switch (i) {
		    case RTAX_DST:
			/* destination */
			if (sa->sin6_family != AF_INET6)
				return;
			sin = sa;
			break;
		}
	}

	/* sanity check */

	if (sin == NULL) {
		logerr("get resolve without destination");
		return;
	}

	/* multicast for ATM uses XRESOLVE too */

	if (IN6_IS_ADDR_MULTICAST(&sin->sin6_addr))
		return;

	if (verbose)
		logtrace("get resolve request for %s",
			 inet6pr(&sin->sin6_addr));

	/* get the neighbor entry */

	ndp = ndplookup(&sin->sin6_addr, 1);
	if (ndp == NULL) {
		/* failed??? zap the route! */
		struct neighbor tmp;

		logerr("can't create entry for resolving %s",
		       inet6pr(&sin->sin6_addr));
		bzero(&tmp, sizeof(tmp));
		bcopy(&sin->sin6_addr, &tmp.addr, sizeof(struct in6_addr));
		tmp.flags |= ND6F_HAS_ROUTE;
		(void) reject_route(&tmp);
		(void) del_route(&tmp);
		/* do more ? ZZZZZZZ */
		return;
	}

	/* now there is a route for this neighbor */

	ndp->flags |= ND6F_HAS_ROUTE;

	/* compute/execute an action according to the neighbor state */

	switch (ndp->state) {
	    case ND6S_INCOMPLETE:
		/* new neighbor */
		enter_state(ndp, ND6S_INCOMPLETE);
		return;

	    case ND6S_STALE:
		/* new packet to send */
		if ((ndp->flags & ND6F_HAS_LLADDR) == 0) {
			logerr("%s is stale without lladdr?");
			(void) del_route(ndp);
			del_neighbor(ndp);
			return;
		}
		/* [re]activate the route */
		change_route(ndp);
		/* delay probing */
		enter_state(ndp, ND6S_DELAY);
		return;

	    case ND6S_PENDING:
		/* waiting for something */
		if (ndp->flags & ND6F_HAS_LLADDR) {
			/* wanted the route */
			change_route(ndp);
			leave_pending(ndp);
			return;
		}
		/* the route is not enough */
		return;

	    default:
		/* should have a route in other states */
		logerr("recv_resolve in state %d?", ndp->state);
		del_neighbor(ndp);
	}
}

/*
 * Get sysctl net.inet6.ipv6.forwarding flag value
 * (this can't change after IPv6 setup)
 */
int
is_router()
{
	static int rmib[] =
		{ CTL_NET, PF_INET6, IPPROTO_IP, IP6CTL_FORWARDING };
	static int val = -1;
	int len = sizeof(int);

	if (val >= 0)
		return val;
	if (sysctl(rmib, sizeof(rmib) / sizeof(int),
		   &val, &len, NULL, NULL) < 0) {
		if (verbose)
			logperror("can't get is_route");
		val = 0;
	}
	return val;
}

/*
 * ZZZZZZ to be done
 */
int
is_apaddr(addr)
	struct in6_addr *addr;
{
	return 0;
}
