/* Neighbor discovery specific routines for LANs */

#include <ndpd.h>

/*
 * Hardware dependent interface initialisations
 */
int
new_lan_if(ifm, first)
	struct if_msghdr *ifm;
	int first;
{
	struct sockaddr_dl *sdl;
	struct nd_opt_hdr *ll;
	int hasphy = 0, ltoken = 0, lmac = 0;

	/* first time? */

	if (first != 1)
		return;

	/* addresses should be interface only */

	if (ifm->ifm_addrs != RTA_IFP) {
		logerr("bad ifinfo (no ifp#%d)", ifm->ifm_index);
		return 0;
	}

	/* get and check the data-link address */

	sdl = (struct sockaddr_dl *)((char *)ifm + sizeof(struct if_msghdr));
	if ((sdl->sdl_family != AF_LINK) || (sdl->sdl_nlen == 0)) {
		logerr("bad sdl address");
		return 0;
	}

	/* verify flags and MTU */

#define IFF_LAN		(IFF_BROADCAST | IFF_MULTICAST)
	if ((ifm->ifm_flags & IFF_LOOPBACK) ||
	    (ifm->ifm_flags & IFF_POINTOPOINT) ||
	    ((ifm->ifm_flags & IFF_LAN) != IFF_LAN) ||
	    (ifm->ifm_data.ifi_mtu < IP6_MMTU)) {
		logerr("incompatible flags 0x%x for interface %.*s",
		       (unsigned)ifm->ifm_flags,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* everything has 48 bit MAC address... */

	ltoken = MACTOKEN;
	lmac = MACMAC;
	if (ifm->ifm_data.ifi_addrlen != lmac) {
		logerr("bad MAC address length (%d != 6) for interface %.*s",
		       ifm->ifm_data.ifi_addrlen,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	if (sdl->sdl_alen != lmac) {
		logerr("unsupported type for interface %.*s",
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* fill the interface structure */

	itf.mtu = ifm->ifm_data.ifi_mtu;
	itf.flags = ifm->ifm_flags;
	itf.ltoken = ltoken;
	itf.lmac = lmac;
	bcopy(sdl->sdl_data, itf.name, sdl->sdl_nlen);
#ifdef notyet
	itf.ntosend = MAX_RTR_SOLICITATIONS;
#endif

	/* allocate and fill the link-layer info */

	if (Calloc(ll, nd_opt_hdr) == NULL) {
		logerr("out of memory");
		return 0;
	}
	ll->nd_opt_len = 1;
	bcopy(LLADDR(sdl), (caddr_t)ll + 2, lmac);
	itf.llinfo = ll;

	/* fill methods */

	itf.lladdr_get = lan_addr_get;
	itf.lladdr_cmp = lan_addr_cmp;
	itf.lladdr_updated = lan_addr_updated;
	itf.lladdr_pr = lan_addr_pr;
	itf.sdl_pr = lan_sdl_pr;

	/* success */

	if (verbose)
		logtrace("got interface %s#%d", itf.name, itf.index);
	return 1;
}

/*
 * Transform a LAN link-layer info in an usable stuff
 */
struct nd_opt_hdr *
lan_addr_get(lp)
	struct nd_opt_hdr *lp;
{
	u_char *cp = (u_char *)lp;

	/* checks */

	if (lp->nd_opt_len != 1) {
		if (verbose)
			logerr("LAN llinfo with length %d > 1",
			       lp->nd_opt_len);
		return NULL;
	}
	if (cp[2] & 0x01) {
		if (verbose)
			logerr("group MAC address %s", lanpr(cp + 2));
		return NULL;
	}

	/* done */

	if (debug)
		logdebug("get MAC address %s", lanpr(cp + 2));
	lp->nd_opt_type = 0;
	return lp;
}

/*
 * Compare two LAN link-layer infos (specialized bcmp, return 0 if equal)
 */
int
lan_addr_cmp(l1, l2)
	struct nd_opt_hdr *l1, *l2;
{
	u_char *cp1, *cp2;

	/* both null: equal */

	if ((l1 == NULL) && (l2 == NULL))
		return 0;

	/* only one is null: different */

	if ((l1 == NULL) || (l2 == NULL))
		return 1;

	/* binary comparison */

	cp1 = (u_char *)l1;
	cp2 = (u_char *)l2;
	return (bcmp(cp1 + 2, cp2 + 2, 6));
}

/*
 * Update a data-link address from a LAN link-layer info
 */
void
lan_addr_updated(ndp)
	struct neighbor *ndp;
{
	u_char *cp;

	/* copy the address inside the link-layer info */

	cp = (u_char *)ndp->llinfo;
	bcopy(cp + 2, LLADDR(&ndp->sdl), 6);
	ndp->sdl.sdl_alen = 6;

	/* success */

	ndp->flags |= ND6F_HAS_LLADDR;
}

/*
 * CLose a circuit (none for LANs)
 */
void
lan_circuit_close(ndp)
	struct neighbor *ndp;
{
	if (debug && (ndp->circuit != NULL))
		logerr("%s as a circuit 0x%x on a LAN",
		       inet6pr(&ndp->addr), (u_long)ndp->circuit);
}

/*
 * Print a LAN link-layer info
 */
char *
lan_addr_pr(lp)
	struct nd_opt_hdr *lp;
{
	u_char *cp = (u_char *)lp;

	if (lp == NULL)
		return (char *)"[NULL]";
	return lanpr(cp + 2);
}

/*
 * Print a LAN data-link address
 */
char *
lan_sdl_pr(sdl)
	struct sockaddr_dl *sdl;
{
	static char s[sizeof("[bad-lenXXX]")];

	if (sdl == NULL)
		return (char *)"[NULL]";
	if (sdl->sdl_alen != 6) {
		(void) snprintf(s, sizeof(s), "[bad-len%d]", sdl->sdl_alen);
		return s;
	}
	return lanpr(LLADDR(sdl));
}

#define LAN_ADDR_SIZE	sizeof("[11:22:33:44:55:66]")

/*
 * Raw print of a LAN address
 */
char *
lanpr(cp)
	u_char *cp;
{
	static char s[8][LAN_ADDR_SIZE];
	static int i = 0;
	int j;

	j = i;
	(void) snprintf(s[j], LAN_ADDR_SIZE,
			"[%02x:%02x:%02x:%02x:%02x:%02x]",
			cp[0], cp[1], cp[2], cp[3], cp[4], cp[5]);
	i += 1;
	i &= 7;
	return s[j];
}
