/* Neighbor discovry main loop, inits and signal handlers */

#include <ndpd.h>

struct timeval now;
int icmp_sd, route_sd;			/* Socket file descriptors */
int pid;

#ifdef USE_NDPD_MAIN

int verbose = 0;			/* logging flags */
int debug = 0;
int timestamps = 0;

static char usage[] = "Usage: ndpd-%s [-d] [-v] [-t] ifname\n";

int main __P((int, char *[]));

/*
 * Entry point
 */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	char ch;
	u_int val;
	int nfds, cc;
	fd_set rfds;
	struct timeval tv;
	TIMEOUT *tp;
	extern int optind;
	extern char *optarg;

#ifdef DO_TIMEOUTS_INIT
	TAILQ_INIT(&timeouts);
#endif
	setlinebuf(stdout);

	/* decode arguments */

	while ((ch = getopt(argc, argv, "dvt")) != EOF)
		switch (ch) {
		    case 'd':
			debug = 1;
			break;
		    case 'v':
			verbose++;
			break;
		    case 't':
			timestamps = 1;
			break;
		    default:
			(void) fprintf(stderr, usage, NAME);
			exit(1);
		}

	argc -= optind;
	argv += optind;

	/* last argument is the mandatory interface name */

	if (argc == 0) {
		(void) fprintf(stderr, "Missing interface name\n");
		(void) fprintf(stderr, usage, NAME);
		exit(1);
	} else if (argc > 1) {
		(void) fprintf(stderr, "Extra parameters\n");
		(void) fprintf(stderr, usage);
		exit(1);
	}

	/* inits */

	ndpd_init();

	/* get interface parameters */

	if (!init(argv[1])) {
		logerr("No IPV6 interface %s!!", argv[1]);
		exit(1);
	}

	/* parameter inits */

	init_param();

	/* main loop: run timeouts, get ICMPv6 and route messages */

	for (;;) {
		RUN_TIMEOUTS();

		/* prepare select() */

		FD_ZERO(&rfds);
		FD_SET(icmp_sd, &rfds);
		nfds = icmp_sd;
		FD_SET(route_sd, &rfds);
		nfds = route_sd > nfds ? route_sd : nfds;

		GET_NEXTTIME(&tv);

		/* main loop select() */

		cc = select(nfds + 1, &rfds, NULL, NULL, &tv);
		if (cc < 0) {
			if (errno != EINTR && verbose)
				logperror("select");
			continue;
		}
		if (cc == 0)
			continue;

		/* get messages */

		(void) gettimeofday(&now, NULL);
		if (FD_ISSET(route_sd, &rfds))
			recv_route();
		if (FD_ISSET(icmp_sd, &rfds))
			recv_icmp();
		if (verbose)
			logerr("end of main loop reached");
	}
	/*NOTREACHED*/
}

#endif

/*
 * First inits (sockets, misc)
 */
void
ndpd_init()
{
	u_int val;
	struct icmp6_filter filter;

	/* misc */

	pid = getpid();
	(void) setpriority(PRIO_PROCESS, 0, -15);
	srandom((int)gethostid());

	/* get the file descriptor for ICMPv6 */

	if ((icmp_sd = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		logperror("socket (ICMP)");
		exit(2);
	}

	/* manage the IPv6 header ourself ??? */

	val = 1;
	if (setsockopt(icmp_sd, IPPROTO_IPV6, IP_HDRINCL,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IP_HDRINCL)");
		exit(3);
	}

	/* get packet information on any recvmsg() call */

	val = 1;
	if (setsockopt(icmp_sd, IPPROTO_IPV6, IPV6_RECVPKTINFO,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IPV6_RECVPKTINFO)");
		exit(3);
	}

	/* no (not yet) need for unicast hops */

	/* set hop limit for multicast packets */

	val = ND_DEFAULT_HOPLIM;
	if (setsockopt(icmp_sd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_HOPS)");
		exit(3);
	}

	/* don't reflect multicast packets */

	val = 0;
	if (setsockopt(icmp_sd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_LOOP)");
		exit(3);
	}

	/* get only interesting ICMPv6 packets */

	ICMP6_FILTER_SETBLOCKALL(&filter);
	ICMP6_FILTER_SETPASS(ND_ROUTER_SOLICIT, &filter);
	ICMP6_FILTER_SETPASS(ND_ROUTER_ADVERT, &filter);
	ICMP6_FILTER_SETPASS(ND_NEIGHBOR_SOLICIT, &filter);
	ICMP6_FILTER_SETPASS(ND_NEIGHBOR_ADVERT, &filter);
	ICMP6_FILTER_SETPASS(ND_REDIRECT, &filter);
	if (setsockopt(icmp_sd, IPPROTO_ICMPV6, ICMP6_FILTER,
		       &filter, sizeof(filter)) < 0) {
		logperror("setsockopt (ICMP6_FILTER)");
		exit(3);
	}

#ifdef notyet
	/* protect the ICMPv6 socket against mobility stuff */

	disable_binding(icmp_sd);
	disable_homeaddr(icmp_sd);
#endif

	/* get the socket descriptor for route message */

	if ((route_sd = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
		logperror("socket (ROUTE)");
		exit(2);
	}

	/* misc */

	(void) signal(SIGUSR1, set_verbose);
	(void) signal(SIGUSR2, set_verbose);
	(void) signal(SIGINT, dump);
	(void) signal(SIGTERM, finish);
	(void) signal(SIGQUIT, finish);

	(void) gettimeofday(&now, NULL);
}

/*
 * Set verbosity (signal handler):
 *     USR1: increase
 *     USR2: reset
 */
void
set_verbose(sig)
	int sig;
{
	if (sig == SIGUSR1)
		verbose++;
	else
		verbose = 0;
}

/*
 * Termination routine (signal handler)
 */
void
finish(s)
	int s;
{
	struct neighbor *ndp;

	if (verbose)
		logtrace("finished!");

	/* delete all neighbor entries */

	FORALLND(ndp) {
		/* avoid real reject by race */
		ndp->state = ND6S_REJECT;
		del_neighbor(ndp);
	}

	/* close every circuit (special NULL argument) */

	if (itf.circuit_close)
		itf.circuit_close(NULL);

	/* not in error a priori */

	exit(0);
}

/* names for dates */

static const char wday_name[7][3] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};
static const char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

/*
 * Dump everything (signal handler)
 */
void dump(s)
	int s;
{
	TIMEOUT *tp;
	struct prefix *pp;
	struct neighbor *ndp;
	struct tm *tmnow;

	/* get and print current time */

	(void) gettimeofday(&now, NULL);
	tmnow = localtime(&now.tv_sec);

	logdebug("DUMP at %.3s %.3s%3d %02d:%02d:%02d.%06d %d",
		 wday_name[tmnow->tm_wday],
		 mon_name[tmnow->tm_mon],
		 tmnow->tm_mday, tmnow->tm_hour,
		 tmnow->tm_min, tmnow->tm_sec,
		 now.tv_usec,
		 1900 + tmnow->tm_year);

	/* timeouts */

	logdebug("\nTIMEOUT LIST:");
	FORALLTP(tp)
		DUMP_TIMEOUT(tp);

	/* the interface */

	logdebug("\nINTERFACE:");
	dump_interface(&itf);

	/* prefixes */

	logdebug("\nPREFIX LIST:");
	FORALLPP(pp)
		dump_prefix(pp);

	/* neighbor entries */

	logdebug("\nNEIGHBOR LIST:");
	FORALLND(ndp)
		dump_neighbor(ndp);
}

/*
 * Dump the interface structure
 */
void
dump_interface(ifp)
	struct interface *ifp;
{
	logdebug("\tname=%s(if#%d)", itf.name, itf.index);
	logdebug("\tlink-local address=%s", inet6pr(&itf.llocal));
	logdebug("\tlgtoken %d lgmac %d", itf.ltoken, itf.lmac);
	logdebug("\tllinfo=%s",
		 itf.llinfo ? itf.lladdr_pr(itf.llinfo) : "NONE");
	logdebug("\tMTU=%lu\tflags=%x", itf.mtu, itf.flags);
}

/*
 * Dump a prefix structure
 */
void
dump_prefix(pp)
	struct prefix *pp;
{
	logdebug("prefix (%lx):", (u_long)pp);
	logdebug("\tprefix=%s with len=%d and mask=%s",
		 inet6pr(&pp->pref), pp->plen, inet6pr(&pp->pmask));
	logdebug("\tif address=%s with mask=%s",
		 inet6pr(&pp->addr), inet6pr(&pp->amask));
	if (!IN6_IS_ADDR_UNSPECIFIED(&pp->dest))
		logdebug("\tdest address=%s", inet6pr(&pp->dest));
	logdebug("\tself=%lx", (u_long)&pp->self);
}

/*
 * Dump a router entry
 */
void
dump_neighbor(ndp)
	struct neighbor *ndp;
{
	logdebug("neighbor (%lx):", (u_long)ndp);
	logdebug("\taddress=%s", inet6pr(&ndp->addr));
	logdebug("\tflags=%x state=%d pstate=%d asked=%d",
		 ndp->flags, ndp->state, ndp->pstate, ndp->asked);
	logdebug("\tsdl=%s", itf.sdl_pr(&ndp->sdl));
	logdebug("\tllinfo=%s",
		 ndp->llinfo ? itf.lladdr_pr(ndp->llinfo) : "NONE");
	logdebug("\ttimeout=%lx circuit=%lx",
		 (u_long)ndp->tp, (u_long)ndp->circuit);
}
