/* Neighbor discovery ICMPv6 message emission */

#include <ndpd.h>

char opacket[MAXPACKET];

/*
 * Send a Router Solicitation packet
 */
int
send_rtsol()
{
	struct ip6_hdr *ip;
	struct nd_router_solicit *rsp;
	struct nd_opt_hdr *op, *lp = itf.llinfo;
	struct sockaddr_in6 dst;
	int len, tlen, cc;

	/* prepare the destination address */

	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	dst.sin6_addr.s6_addr[0] = 0xff;
	dst.sin6_addr.s6_addr[1] = MADDR6_SCP_LINK;
	dst.sin6_addr.s6_addr[15] = MADDR6_ALLROUTERS;
	dst.sin6_scope_id = itf.index;

	/* compute lengths */

	len = sizeof(*rsp);
	if (lp != NULL)
		len += lp->nd_opt_len << 3;
	tlen = len + sizeof(*ip);
	bzero(opacket, tlen);

	/* fill the IPv6 header */

	ip = (struct ip6_hdr *)opacket;
	bcopy(&itf.llocal, &ip->ip6_src, sizeof(struct in6_addr));
	bcopy(&dst.sin6_addr, &ip->ip6_dst, sizeof(struct in6_addr));
	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len = htons(len);

	/* fill the ICMPv6/RtSol header */

	rsp = (struct nd_router_solicit *)(ip + 1);
	rsp->nd_rs_type = ND_ROUTER_SOLICIT;

	/* other fields are 0 */

	/* fill the source link-layer address option if needed */

	if (lp != NULL) {
		op = (struct nd_opt_hdr *)(rsp + 1);
		bcopy(lp, op, lp->nd_opt_len << 3);
		op->nd_opt_type = ND_OPT_SOURCE_LINKADDR;
	}

	/* compute the ICMPv6 checksum here */

	rsp->nd_rs_cksum = in_cksum((u_short *)ip, tlen);

	/* finish the IPv6 header */

	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	if (verbose)
		logtrace("Sending Rt Sol to %s len %d",
			 inet6pr(&dst.sin6_addr), len - sizeof(*rsp));

	/* send the packet */

	cc = sendto(icmp_sd, opacket, tlen, 0,
		    (struct sockaddr *)&dst, sizeof(dst));

	/* failed? */

	if ((cc < 0) || (cc != tlen))  {
		logerr("wrote %d chars, ret=%d", tlen, cc);
		if (cc < 0)
			logperror("sendto");
		return -1;
	}

	/* success */

	return 0;
}

/*
 * Send a Neighbor Solicitation packet
 */
int
send_ndsol(ndp, kind)
	struct neighbor *ndp;
	int kind;
{
	struct ip6_hdr *ip;
	struct nd_neighbor_solicit *nsp;
	struct nd_opt_hdr *op, *lp = itf.llinfo;
	struct sockaddr_in6 dst;
	int len, tlen, cc;

	/* prepare the destination address */

	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	dst.sin6_scope_id = itf.index;

	/* finish the destination address according to the kind */

	switch (kind) {
	    case NS6K_DAD:
		/* duplicate address detection: ff02::1 */
		dst.sin6_addr.s6_addr[0] = 0xff;
		dst.sin6_addr.s6_addr[1] = MADDR6_SCP_LINK;
		dst.sin6_addr.s6_addr[15] = MADDR6_ALLNODES;
		break;

	    case NS6K_SOL:
		/* solicited multicast address: ff02::1:ffxxyyzz */
		dst.sin6_addr.s6_addr[0] = 0xff;
		dst.sin6_addr.s6_addr[1] = MADDR6_SCP_LINK;
		dst.sin6_addr.s6_addr[11] = 0x01;
		dst.sin6_addr.s6_addr[12] = 0xff;
		dst.sin6_addr.s6_addr[13] = ndp->addr.s6_addr[13];
		dst.sin6_addr.s6_addr[14] = ndp->addr.s6_addr[14];
		dst.sin6_addr.s6_addr[15] = ndp->addr.s6_addr[15];
		break;

	    case NS6K_DIR:
		/* direct: neighbor address */
		bcopy(&ndp->addr, &dst.sin6_addr, sizeof(struct in6_addr));
		break;

	    default:
		/* should not happen */
		logerr("Send Nd Sol unkown kind %d", kind);
		return -1;
	}

	/* compute lengths */

	len = sizeof(*nsp);
	if ((lp != NULL) &&
	    ((kind == NS6K_SOL) ||
	     ((kind == NS6K_DIR) && (ndp->asked > 0))))
		len += lp->nd_opt_len << 3;
	else
		lp = NULL;
	tlen = len + sizeof(*ip);
	bzero(opacket, tlen);

	/* fill the IPv6 header */

	ip = (struct ip6_hdr *)opacket;
	bcopy(&itf.llocal, &ip->ip6_src, sizeof(struct in6_addr));
	bcopy(&dst.sin6_addr, &ip->ip6_dst, sizeof(struct in6_addr));
	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len = htons(len);

	/* fill the ICMPv6/NdSol header */

	nsp = (struct nd_neighbor_solicit *)(ip + 1);
	nsp->nd_ns_type = ND_NEIGHBOR_SOLICIT;
	bcopy(&ndp->addr, &nsp->nd_ns_target, sizeof(struct in6_addr));

	/* other fields are 0 */

	/* fill the source link-layer address option if needed */

	if (lp != NULL) {
		op = (struct nd_opt_hdr *)(nsp + 1);
		bcopy(lp, op, lp->nd_opt_len << 3);
		op->nd_opt_type = ND_OPT_SOURCE_LINKADDR;
	}

	/* compute the ICMPv6 checksum here */

	nsp->nd_ns_cksum = in_cksum((u_short *)ip, tlen);

	/* finish the IPv6 header */

	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	if (verbose)
		logtrace("Sending%s Nd Sol to %s for %s len %d",
			 kind == NS6K_DAD ? " DAD" : "",
			 inet6pr(&dst.sin6_addr),
			 inet6pr(&nsp->nd_ns_target),
			 len - sizeof(*nsp));

	/* send the packet */

	cc = sendto(icmp_sd, opacket, tlen, 0,
		    (struct sockaddr *)&dst, sizeof(dst));

	/* failed? */

	if ((cc < 0) || (cc != tlen)) {
		logerr("wrote %d chars, ret=%d", tlen, cc);
		if (cc < 0)
			logperror("sendto");
		return -1;
	}

	/* success */

	return 0;
}

/*
 * Send a Neighbor Advertisement packet
 */
int
send_ndadv(ndp, pp, is_solicited)
	struct neighbor *ndp;
	struct prefix *pp;
	int is_solicited;
{
	struct ip6_hdr *ip;
	struct nd_neighbor_advert *nap;
	struct nd_opt_hdr *op, *lp = itf.llinfo;
	struct sockaddr_in6 dst;
	int len, tlen, cc, for_router = is_router();

	/* prepare the destination address */

	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	bcopy(&ndp->addr, &dst.sin6_addr, sizeof(struct in6_addr));
	dst.sin6_scope_id = itf.index;

	/* compute lengths */

	len = sizeof(*nap);
	if (lp != NULL)
		len += lp->nd_opt_len << 3;
	tlen = len + sizeof(*ip);
	bzero(opacket, tlen);

	/* fill the IPv6 header */

	ip = (struct ip6_hdr *)opacket;
	bcopy(&itf.llocal, &ip->ip6_src, sizeof(struct in6_addr));
	bcopy(&dst.sin6_addr, &ip->ip6_dst, sizeof(struct in6_addr));
	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len = htons(len);

	/* fill the ICMPv6/NdAdv header */

	nap = (struct nd_neighbor_advert *)(ip + 1);
	nap->nd_na_type = ND_NEIGHBOR_ADVERT;
	nap->nd_na_flags_reserved = ND_NA_FLAG_OVERRIDE;
	if (is_solicited)
		nap->nd_na_flags_reserved |= ND_NA_FLAG_SOLICITED;
	if (for_router)
		nap->nd_na_flags_reserved |= ND_NA_FLAG_ROUTER;
	bcopy(&pp->addr, &nap->nd_na_target, sizeof(struct in6_addr));

	/* other fields are 0 */

	if (lp != NULL) {
		op = (struct nd_opt_hdr *)(nap + 1);
		bcopy(lp, op, lp->nd_opt_len << 3);
		op->nd_opt_type = ND_OPT_TARGET_LINKADDR;
	}

	/* compute the ICMPv6 checksum here */

	nap->nd_na_cksum = in_cksum((u_short *)ip, tlen);

	/* finish the IPv6 header */

	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	if (verbose)
		logtrace("Sending Nd Adv to %s for %s len %d flags %s%sO",
			 inet6pr(&dst.sin6_addr),
			 inet6pr(&nap->nd_na_target),
			 len - sizeof(*nap),
			 for_router ? "R" : "",
			 is_solicited ? "S" : "");

	/* send the packet */

	cc = sendto(icmp_sd, opacket, tlen, 0,
		    (struct sockaddr *)&dst, sizeof(dst));

	/* failed? */

	if ((cc < 0) || (cc != tlen)) {
		logerr("wrote %d chars, ret=%d", tlen, cc);
		if (cc < 0)
			logperror("sendto");
		return -1;
	}

	/* success */

	return 0;
}
