/* Neighbor discovery utility routines */

#include <ndpd.h>

#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

static void logdate(FILE *);

void
apply_mask(addr, mask)
	register struct in6_addr *addr, *mask;
{
	register int i;

	for (i = 0; i < 16; i++)
		addr->s6_addr[i] &= mask->s6_addr[i];
}

void
mk_pref(pp)
	register struct prefix *pp;
{
	mk_mask(&pp->pmask, pp->plen);
	apply_mask(&pp->pref, &pp->pmask);
}

void
mk_mask(addr, len)
	register struct in6_addr *addr;
	register int len;
{
	register u_char *cp;

	CLR_ADDR6(*addr);
	cp = (u_char *)addr;
	for (cp = (u_char *)addr; len > 7; len -= 8)
		*cp++ = 0xff;
	if (len)
		*cp = 0xff << (8 - len);
}

/*
 * Build a neighbor which is me for this prefix
 */
void
mk_self(pp)
	struct prefix *pp;
{
	struct neighbor *ndp = &pp->self;
	struct nd_opt_hdr *lli = itf.llinfo;

	bcopy(&pp->addr, &ndp->addr, sizeof(struct in6_addr));
	if (is_router())
		ndp->flags = ND6F_IS_ROUTER;
	ndp->state = ND6S_LOCKED;
	ndp->flags |= ND6F_IS_SELF;
	if (lli != NULL) {
		struct nd_opt_hdr *lln;
		int size = lli->nd_opt_len << 3;

		lln = (struct nd_opt_hdr *)calloc(1, size);
		if (lln == NULL) {
			logerr("out of memory");
			abort();
		}
		bcopy(lli, lln, size);
		ndp->llinfo = lln;
	}
	itf.lladdr_updated(ndp);
	(void) local_route(ndp);
}

/* return 0 on error */
u_char
get_prefixlen(sin)
	struct sockaddr_in6 *sin; /* family=AF_UNSPEC, may be truncated */
{
	register u_char *cp;
	register int byte, bit, len;

#define FBEGIN	((int)&(((struct sockaddr_in6 *)0)->sin6_addr))
#define FEND	(FBEGIN + sizeof(struct in6_addr))
	if (sin->sin6_len < FBEGIN ||
	    sin->sin6_len > sizeof(struct sockaddr_in6))
		return(0);

	cp = (u_char *)sin + FBEGIN;

	len = 0;
	for (byte = FBEGIN;
	     (byte < sin->sin6_len) && (byte < FEND);
	     byte++, cp++, len += 8)
		if (*cp != 0xff)
			break;
	if ((byte == sin->sin6_len) || (byte == FEND))
		return((u_char)len);

	for (bit = 7; bit > 0; bit--, len++)
		if (!(*cp & (1 << bit)))
			break;
	for (; bit > 0; bit--)
		if (*cp & (1 << bit))
			return(0);

	byte++; cp++;
	for (; (byte < sin->sin6_len) && (byte < FEND); byte++, cp++)
		if (*cp)
			return(0);
	return((u_char)len);
}

/* Internet checksum */
int
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	u_short odd_byte = 0;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&odd_byte) = *(u_char *)w;
		sum += odd_byte;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

static char p_buf[NPR6BUF][INET6_ADDRSTRLEN];
static p_pos = 0;
#define PR_NXTBUF	p_buf[p_pos = (p_pos+1) % NPR6BUF]
#define PR_CURBUF	p_buf[p_pos]

const char *
inet6pr(addr)
	register struct in6_addr *addr;
{
	return inet_ntop(AF_INET6, addr, PR_NXTBUF, INET6_ADDRSTRLEN);
}

#undef PR_NXTBUF
#undef PR_CURBUF

static logging = 0;

void
initlog()
{
	logging++;
	openlog("ndpd", LOG_PID | LOG_CONS, LOG_DAEMON);
}

static void
logdate(f)
	FILE *f;
{
	char *timestamp;

	if (!timestamps)
		return;
	(void)gettimeofday(&now, NULL);
	timestamp = ctime(&now.tv_sec);
	timestamp[19] = '\0';
	(void) fprintf(f, "%s.%06ld: ", timestamp + 11, now.tv_usec);
}

void
#if __STDC__
logerr(char *fmt, ...)
#else
logerr(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	va_list ap;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (logging)
		vsyslog(LOG_ERR, fmt, ap);
	else {
		logdate(stderr);
		(void) vfprintf(stderr, fmt, ap);
		(void) fprintf(stderr, "\n");
	}
}

void
#if __STDC__
logtrace(char *fmt, ...)
#else
logtrace(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	va_list ap;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (logging)
		vsyslog(LOG_INFO, fmt, ap);
	else {
		logdate(stdout);
		(void) vfprintf(stdout, fmt, ap);
		printf("\n");
	}
}

void
#if __STDC__
logdebug(char *fmt, ...)
#else
logdebug(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	va_list ap;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (logging)
		vsyslog(LOG_DEBUG, fmt, ap);
	else {
		logdate(stdout);
		(void) vfprintf(stdout, fmt, ap);
		printf("\n");
	}
}

void
logperror(str)
	char *str;
{
	if (logging)
		syslog(LOG_ERR, "%s: %m", str);
	else {
		logdate(stderr);
		(void) fprintf(stderr, "%s: %s\n", str, strerror(errno));
	}
}

#ifdef TRACE

#define RND(l)	{ if ((l) == 0) (l) = 1; (l) = ((l) + 3) / 4; }
void
DUMPS(p)
	u_int32_t *p;
{
	struct sockaddr *s = (struct sockaddr *)p;
	int l = s->sa_len;

	RND(l);
	if (l-- > 0)
		printf("%08x", (int)htonl(*p++));
	while (l-- > 0)
		printf(" %08x", (int)htonl(*p++));
	printf("\n");
}

static const char *rtan[] = {
	"DST", "GATEWAY", "NETMASK", "GENMASK", "IFP",
	"IFA", "AUTHOR", "BRD", "DOWNSTREAM",
};

void
DUMPSP(addrs, pa, len)
	int addrs, len;
	u_int32_t *pa;
{
	int b, l, i;
	u_int32_t *p;
	char vb[80];
#define P(t)	((struct t *)p)

	printf("addrs");
	for (b = 0, l = len, p = pa; b < RTAX_MAX; b++) {
		if ((addrs & (1 << b)) == 0)
			continue;
		addrs &= ~(1 << b);
		printf(" %s", rtan[b]);
		if (l <= 0)
			continue;
		switch (P(sockaddr)->sa_family) {
		case AF_INET:
			inet_ntop(AF_INET,
				  &P(sockaddr_in)->sin_addr,
				  vb, sizeof vb);
			printf("(%s)", vb);
			break;
		case AF_INET6:
			inet_ntop(AF_INET6,
				  &P(sockaddr_in6)->sin6_addr,
				  vb, sizeof vb);
			printf("(%s", vb);
			if (P(sockaddr_in6)->sin6_scope_id)
				printf("#%d", P(sockaddr_in6)->sin6_scope_id);
			printf(")");
			break;
		case AF_LINK:
			printf("(%s#%d)",
				link_ntoa(P(sockaddr_dl)),
				P(sockaddr_dl)->sdl_index);
		}
		i = P(sockaddr)->sa_len;
		RND(i);
		p += i;
		l -= 4 * i;
	}
	if (addrs)
		printf(" 0x%x(???)", addrs);
	printf("\n");

	for (l = len, p = pa; l > 0; ) {
		DUMPS(p);
		i = P(sockaddr)->sa_len;
		RND(i);
		p += i;
		l -= 4 * i;
	}
#undef P
}

void
DUMPRT(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct rt_msghdr *r = (struct rt_msghdr *)ad;
	u_int32_t *p;
	int len = l - sizeof(*r);

	printf("route(%s) ty=%d l=%d(%d) idx=%d fl=%x "
	       "add=%x pid=%d/%d err=%d us=%d mtx=%lx\n",
		nom, r->rtm_type, r->rtm_msglen, l, r->rtm_index, r->rtm_flags,
		r->rtm_addrs,
		r->rtm_pid, r->rtm_seq, r->rtm_errno,
		r->rtm_use, r->rtm_inits);
	p = (u_int32_t *)&(r->rtm_rmx);
	printf("%08x %08x %08x %08x %08x\n%08x %08x %08x %08x %08x\n",
	       *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++);
	p = (u_int32_t *)(ad + sizeof(*r));
	DUMPSP(r->rtm_addrs, p, len);
}

void
DUMPA(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct ifa_msghdr *r = (struct ifa_msghdr *)ad;
	u_int32_t *p;
	int len = l - sizeof(*r);

	printf("ifa(%s) ty=%d l=%d(%d) idx=%d fl=%x add=%x mtx=%x\n",
		nom, r->ifam_type, r->ifam_msglen, l, r->ifam_index,
		r->ifam_flags, r->ifam_addrs, r->ifam_metric);
	p = (u_int32_t *)(ad + sizeof(*r));
	DUMPSP(r->ifam_addrs, p, len);
}

void
DUMPI(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct if_msghdr *r = (struct if_msghdr *)ad;
	u_int32_t *p;
	int len = l - sizeof(*r);

	printf("if(%s) ty=%d l=%d(%d) idx=%d fl=%x add=%x\n",
		nom, r->ifm_type, r->ifm_msglen, l, r->ifm_index, r->ifm_flags,
		r->ifm_addrs);
	p = (u_int32_t *)&(r->ifm_data);
	printf("%08x %08x %08x %08x %08x\n%08x %08x %08x %08x %08x\n",
	       *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++);
	printf("%08x %08x %08x %08x %08x\n%08x %08x %08x %08x %08x\n",
	       *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++, *p++);
	p = (u_int32_t *)(ad + sizeof(*r));
	DUMPSP(r->ifm_addrs, p, len);
}

void
DUMPX(p, l)
	u_int32_t *p;
	int l;
{
	int i = 0;

	while (l > 0) {
		printf("%08x ", *p++);
		l--;
		i++;
		if (i == 8) {
			i = 0;
			printf("\n");
		}
	}
	if (i)
		printf("\n");
}

void
DUMP(nom, ad, l)
	const char *nom;
	void *ad;
	int l;
{
	struct rt_msghdr *r = (struct rt_msghdr *)ad;

	if (r->rtm_type == RTM_IFINFO)
		DUMPI(nom, ad, l);
	else if (r->rtm_type == RTM_NEWADDR || r->rtm_type == RTM_DELADDR)
		DUMPA(nom, ad, l);
	else DUMPRT(nom, ad, l);
}

#endif /* TRACE */
