#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>
#include <sys/uio.h>
#include <sys/resource.h>

#include <net/if.h>
#ifndef LLADDR
#include <net/if_dl.h>
#endif
#include <net/if_types.h>
#ifndef IF_QFULL
#include <net/if_var.h>
#endif
#include <net/route.h>

#include <netinet/in_systm.h>
#define INET6
#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <netinet/ip6_opts.h>

#include <netdb.h>
#include <arpa/inet.h>

#ifdef TRACE
void DUMP(const char *, void *, int);
#define DUMPR(x) DUMP(x, &rtm, rtm.rtm_msglen)
#else
#define DUMPR(x)
#endif

#define NPR6BUF	8	/* max printing computed strings in parallel */

#ifndef TAILQ_INSERT_BEFORE
#define	TAILQ_INSERT_BEFORE(listelm, elm, field) {			\
	(elm)->field.tqe_prev = (listelm)->field.tqe_prev;		\
	(elm)->field.tqe_next = (listelm);				\
	*(listelm)->field.tqe_prev = (elm);				\
	(listelm)->field.tqe_prev = &(elm)->field.tqe_next;		\
}
#endif

#define IN6_GET_SCOPE(a)	((a)->s6_addr[1] & 0x0f)

#define Calloc(p, t)	(p = (struct t *)calloc(1,sizeof(struct t)))

/* for some sanity checks */
#define MAXHOP	255
#define MINHOP	1

#define MAXTOKEN	12	/* That is a max */
#define MACTOKEN	(8)	/* eth, fddi, ...*/
#define MACMAC		6
#define PPPTOKEN	8	/* ppp */
#define SITTOKEN	4	/* sit point to point */

/* millisecond to timeval */
#define M_SECOND			1000	/* ms per sec */
#define MU_SECOND			1000	/* microsec per M_SECOND */
#define MS_TO_SEC(x)			((x)/M_SECOND)
#define MS_TO_USEC(x)			(((x)%M_SECOND)*MU_SECOND)

/*
   from RFC 2461
	MAX_INITIAL_RTR_ADVERT_INTERVAL  16 seconds
	MAX_INITIAL_RTR_ADVERTISEMENTS    3 transmissions
	MAX_FINAL_RTR_ADVERTISEMENTS      3 transmissions
	MIN_DELAY_BETWEEN_RAS             3 seconds
	MAX_RA_DELAY_TIME                 .5 seconds

	MAX_RTR_SOLICITATION_DELAY        1 second
	RTR_SOLICITATION_INTERVAL         4 seconds
	MAX_RTR_SOLICITATION              3 transmissions

	MAX_MULTICAST_SOLICIT             3 transmissions
        MAX_UNICAST_SOLICIT               3 transmissions
	MAX_ANYCAST_DELAY_TIME            1 second
	MAX_NEIGHBOR_ADVERTISEMENT        3 transmissions
	REACHABLE_TIME                30000 milliseconds
	RETRANS_TIMER                  1000 milliseconds
	DELAY_FIRST_PROBE_TIME            5 seconds
	MIN_RANDOM_FACTOR                 .5
	MAX_RANDOM_FACTOR                1.5

	AdvSendAdvertisements             FALSE
	MaxRtrAdvInterval        Default: 600 seconds
	MaxRtrAdvInterval        Maximum: 1800 seconds
	MaxRtrAdvInterval        Minimum: 4 seconds
	MinRtrAdvInterval        Default: 0.33 * MaxRtrAdvInterval
	MinRtrAdvInterval        Maximum: 0.75 * MaxRtrAdvInterval
	MinRtrAdvInterval        Minimum: 3 seconds
	AdvManagedFlag           Default: FALSE
	AdvOtherConfigFlag       Default: FALSE
	AdvLinkMTU               Default: 0 (unspecified)
	AdvReachableTime         Default: 0 (unspecified)
	AdvReachableTime         Maximum: 3600000 milliseconds (1 hour)
	AdvRetransTimer          Default: 0 (unspecified)
	AdvMaxHopLimit           Default: value in "Assigned Numbers" RFC
	AdvDefaultLifetime       Default: 3 * MaxRtrAdvInterval
	AdvDefaultLifetime       Maximum: 9000 seconds
	AdvDefaultLifetime       Minimum: 0 (not used as a default router)
	AdvPrefixList       Each prefix has an associated:
	       AdvValidLifetime           Default: 2592000 seconds (30 days)
	       AdvValidLifetime           Maximum: 0xffffffff (infinity)
	       AdvOnLinkFlag              Default: TRUE
	       AdvPreferredLifetime       Default: 604800 seconds (7 days)
	       AdvPreferredLifetime       Maximum: 0xffffffff (infinity)
	       AdvAutonomousFlag          Default: TRUE

   from mobility draft 7
	MAX_ADVERT_REXMIT		  3 transmissions

   from implementation
	INACTIVE_TIME                     600 seconds (10 minutes)
	INACTIVE_TIME            Maximum: 3600 seconds (1 hour)
	HOLD_DOWN_TIME                    AdvReachableTime / 10
	HOLD_DOWN_TIME           Maximum: 5 seconds
	AdvReachableTime         Minimum: 2000 milliseconds
	AdvRetransTimer          Maximum: AdvReachableTime / 5
	AdvRetransTimer          Minimum: 20 milliseconds
*/

#define MAX_INITIAL_RTR_ADVERT_INTERVAL	(16*M_SECOND)
#define MAX_INITIAL_RTR_ADVERTISEMENTS	3
#define MAX_FINAL_RTR_ADVERTISEMENTS	3
#define MIN_DELAY_BETWEEN_RAS		(3*M_SECOND)
#define MAX_RA_DELAY_TIME		(M_SECOND/2)

#define MAX_RTR_SOLICITATION_DELAY	1.0
#define RTR_SOLICITATION_INTERVAL 	4.0
#define MAX_RTR_SOLICITATIONS 		3

#define MAX_MULTICAST_SOLICIT		3
#define MAX_UNICAST_SOLICIT		3
#define MAX_ANYCAST_DELAY_TIME		(M_SECOND)
#define MAX_NEIGHBOR_ADVERTISEMENT	3
#define REACHABLE_TIME			30000
#define MAX_REACHABLE_TIME		3600000
#define MIN_REACHABLE_TIME		2000
#define RETRANS_TIMER			1000
#define MIN_RETRANS_TIMER		20
#define INACTIVE_TIME			600000
#define MAX_INACTIVE_TIME		3600000
#define MAX_HOLD_DOWN_TIME		5000
#define DELAY_FIRST_PROBE_TIME		5000
#define MIN_RANDOM_FACTOR		0.5
#define MAX_RANDOM_FACTOR		1.5

#define DEF_RTR_ADV_INTERVAL		(600*M_SECOND)
#define MAX_RTR_ADV_INTERVAL		(1800*M_SECOND)
#define MIN_RTR_ADV_INTERVAL		(4*M_SECOND)
#define MAX_RT_LIFE			0xffff
#define MAX_ADVERT_REXMIT		3

/* randomize between 1.5 - 0.5, include bound */
#define RANDOMIZE(x)			((((x)+1)/2) + random() % (1+(x)))

/* Router Adv flags */
#define RTADV_FLG_M	0x80
#define RTADV_FLG_O	0x40
#define RTADV_FLG	(RTADV_FLG_M|RTADV_FLG_O)

/* icmp socket */
#define	MAXPACKET	8192	/* max packet size */
struct {
	struct cmsghdr c_cm;
	u_char c_space[64];
} c_cmsg;
#define cm	c_cmsg.c_cm
extern char packet[];

/* route socket */
#define MAX_MSG_ROUTE	2048
struct {
       	struct rt_msghdr m_rtm;
        char m_space[MAX_MSG_ROUTE - sizeof(struct rt_msghdr)];
} m_rtmsg;
#define rtm	m_rtmsg.m_rtm
extern int rtm_seq;
extern int pid;

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(int32_t) - 1))) : sizeof(int32_t))
#define ADVANCE(x, n) (x += ROUNDUP(((struct sockaddr *)n)->sa_len))

extern struct timeval now;

struct prefix;
int in_cksum(u_short *, int);
void apply_mask(struct in6_addr *, struct in6_addr *);
void mk_mask(struct in6_addr *, int);
void mk_pref(struct prefix *);
void mk_self(struct prefix *);
u_char get_prefixlen(struct sockaddr_in6 *);
void initlog(void);
void logerr(char *, ...), logtrace(char *, ...), logdebug(char *, ...);
void logperror(char *);
const char *inet6pr(struct in6_addr *);
